/*******************************************************************************
+
+  AGD 1.2
+
+
+  AugmentationModule.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        AugmentationModule.h
//
//
//    author: Carsten Gutwenger
//    date:   july 1997
//-------------------------------------------------------------------


#ifndef _AGD_AUGMENTATION_MODULE_H
#define _AGD_AUGMENTATION_MODULE_H


#include <AGD/AGDModule.h>


/*{\Manpage {AugmentationModule} {} {Augmentation Modules} {A} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5.5}{AGDModule}
\arhook{5.5}{2}{6.5}
\hdclass{0.3}{0}{6.7}{AugmentationModule}
\end{hierarchy}
}*/

class __exportAGD AugmentationModule : public AGDModule {

/*{\Mdefinition
The class |\Mname| is the base class for augmentation modules. An
augmentation module transforms an input graph $G$ into an output
graph $G'$ by adding edges, such that $G'$ satisfies a certain
condition, e.g., biconnected. This condition is expressed by
the postcondition of the module.

\headerline{Input and Output Parameters}

\begin{tabular}{lcc}
input parameter: & graph & |in_graph=1| \\
output parameter: & supergraph & |out_supergraph=1|
\end{tabular}

The integer constants |in_graph| and |out_supergraph| denote the
positions in the input and output parameter lists. The values can be used to
reference the desired parameter when setting the pre- and postcondition.  

\headerline{Implementation of Augmentation Algorithms}

An implementation of an augmentation module must override
the protected method do\_call($G$,$L$), which gets as
input a graph reference $G$. The augmented edges are added
to $G$, and the list of added edges is returned in $L$.
}*/

public:

	// the input and output parameters of an augmentation module
	static const int in_graph;
	static const int out_supergraph;

	/*{\Minitialization A 5}*/

	AugmentationModule () : AGDModule(1,1) { }

	/*{\Mcreate initializes an augmentation module.}*/

	virtual ~AugmentationModule () { }


	/*{\Moperations}*/

	/*{\Mtext
	\headerline{Standard Interface} }*/

	bool check (const leda::graph& G, AgdKey &p) const;
	/*{\Mop returns |true| if $G$ satisfies the precondition of |\Mvar|.
	Otherwise, |false| is returned and |p| contains a property
	that is not satisfied.}*/

	void call (leda::graph& G);
	/*{\Mop calls the augmentation module for graph $G$.}*/

	void call (leda::graph& G, leda::list<leda::edge>& L);
	/*{\Mop calls the augmentation module for graph $G$. Returns the
	list of added edges in |L|.}*/
	
	int num_added_edges() const {
		return _num_added_edges;
	}
	/*{\Mop returns the number of added edges.}*/

protected:

	/*{\Mtext
	\bigskip
	\headerline{Protected Overridables to Implement Functionality} }*/

	virtual void do_call (leda::graph& G, leda::list<leda::edge>& L)
	/*{\Moppure implements the augmentation algorithm for graph |G|.
	Returns the list of added edges in |L|.}*/
	= 0;

	/*{\Mtext
	\bigskip
	\headerline{Protected Auxiliary Methods} }*/

	static void hide_self_loops (leda::graph& G, leda::list<leda::edge>& L);
	/*{\Mstatic hides all self loops and appends them to |L|.}*/

	static void hide_multiple_edges (leda::graph& G, leda::list<leda::edge>& L);
	/*{\Mstatic hides all but one of each set of parallel edges and
	appends all hidden edges to |L|.}*/

private:
	int _num_added_edges;
};


#endif
