/*******************************************************************************
+
+  AGD 1.2
+
+
+  Cluster.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         Cluster.h
//
//    author: Dirk Luetke-Huettmann
//    date:   july 1999
//-------------------------------------------------------------------

#ifndef AGD_CLUSTER_H
#define AGD_CLUSTER_H 1

#include <LEDA/graph/graph.h>
#include <AGD/basic.h>


/*{\Manpage {Cluster} {} {Clusters} {C} }*/

/*{\Mbaseclasses
\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5}{Cluster}
\end{hierarchy}
}*/

class __exportAGD Cluster {

/*{\Mdefinition
An instance |\Mvar| of type |\Mname| offers operations for the member nodes
of |\Mvar| and its subclusters.
}*/

public:

/*{\Mcreation C 5.5}*/

  Cluster();
  /*{\Mcreates creates an empty cluster with no member nodes and no parent cluster.}*/

  //  Cluster(string name);
  ~Cluster();

/*{\Moperations}*/

  Cluster* get_supercluster() const {
	  return supercluster;
  }
  /*{\Mop returns the parent cluster of |\Mvar|.}*/

  void set_supercluster(Cluster* cl) {
	  supercluster = cl;
  }
  /*{\Mop makes |cl| the parent cluster of |\Mvar|. {\bf Attention:} this 
    function does not add |\Mvar| to the children list of |cl|.}*/

  const leda::list<Cluster*> &get_subclusters() const {
	  return subclusters;
  }
  /*{\Mop returns a list of subclusters of |\Mvar|.}*/

  void add_subcluster(Cluster* cl) {
	  subclusters.append(cl);
  }
  /*{\Mop makes |cl| a subcluster of |\Mvar|.}*/

  void remove_subcluster(Cluster* cl);
  /*{\Mop removes |cl| from the list of subcluster of |\Mvar|. {\bf
   Attention:} |cl| is unchanged.}*/

  const leda::list<leda::node> &get_nodes() const {
	  return clusternodes;
  }
  /*{\Mop returns a list of member nodes of |\Mvar|.}*/

  void add_node(leda::node v) {
    clusternodes.append(v);
  };
  /*{\Mop makes node |v| a member of |\Mvar|.}*/

  bool remove_node(leda::node v);
  /*{\Mop removes node |v| from the list of member nodes.}*/
  
  void set_edge_order(leda::list<leda::edge> cl_edge_order) {
    clear_edge_order();
    leda::edge e;
    forall(e, cl_edge_order) append_to_edge_order(e);
  }
  /*{\Mop sets the order of the cluster incident edges around |\Mvar|
    to the order of the edges in |cl_edge_order|.}*/

  const leda::list<leda::edge> &get_edge_order() const {
    return cl_embedded_edges;
  }
  /*{\Mop returns a ordered list of cluster incident edges around
    cluster |\Mvar|.}*/

  void append_to_edge_order(leda::edge e) {
    cl_embedded_edges.append(e);
  }

  void clear_edge_order() {
    cl_embedded_edges.clear();
  }
  

  void clear_subclusters() {
	  subclusters.clear();
  }

  void clear_clusternodes() {
	  clusternodes.clear();
  }
  
private:
  // reference to the super-cluster
  Cluster* supercluster;

  // list of references to the child cluster
  leda::list<Cluster*> subclusters;

  // list of references to the nodes contained inside the cluster
  leda::list<leda::node> clusternodes;

  // ordered list of cluster incidential edges after embedding
  leda::list<leda::edge> cl_embedded_edges;

};

#endif
