/*******************************************************************************
+
+  AGD 1.2
+
+
+  ConvexDrawLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                      ConvexDrawLayout.h
//
//    author: Carsten Gutwenger
//    date:   february 1997
//-------------------------------------------------------------------


#ifndef _AGD_CONVEX_DRAW_LAYOUT_H
#define _AGD_CONVEX_DRAW_LAYOUT_H


#include <AGD/GridLayoutModuleImpl.h>
#include <AGD/PlaneGraphCopyGridConcept.h>


class CanonicalOrder;


/*{\Manpage {ConvexDrawLayout} {} {The ConvexDraw Algorithm} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{9.5}{5.5}
\hdclass{0}{4.5}{5}{AGDModule}
\arhook{5}{5}{6}
\hdclass{1.5}{3}{5}{LayoutModule}
\arhook{6.5}{3.5}{7.5}
\hdclass{2}{1.5}{6}{GridLayoutModule}
\arhook{8}{2}{9}
\hdclass{3.3}{0}{6.2}{ConvexDrawLayout}
\end{hierarchy}
\hfill\agdeps{convex_draw.eps}{3.8cm}
}*/

class __exportAGD ConvexDrawLayout : public GridLayoutModuleImpl<PlaneGraphCopyGridConcept> {

/*{\Mdefinition
The class |\Mname| represents the layout algorithm ConvexDraw by
Marek Chrobak and Goos Kant~\cite{CK97}.

This algorithm draws a planar, triconnected graph |G| straight-line
without crossings. |G| must not contain self-loops or multiple edges.
All faces are drawn convex, the grid layout
size is at most $(n-1) \times (n-1)$ for a graph with $n$ nodes ($n \geq 3$).

The algorithm runs in two phases. In the first phase, 
a so-called shelling order (also called canonical ordering)
for triangulated planar graphs
is computed. Then, the vertices are placed incrementally
according to the shelling order.

\horizline
\commoninfo{ConvexDraw}{ConvexDraw (Chrobak/Kant)}{M. Chrobak, G. Kant}
{C. Gutwenger}{March 1997}{2.0}

\begin{preposttab}
precondition & $\{\, \emph{planar, triconnected, simple, no self loops} \,\}$ \\
postcondition(|PRE|) & $\{\, \emph{straight-line, no crossings, convex} \,\}$
\end{preposttab}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |bool| {\bf size\_opt} = |true|\\
a boolean value determining if the grid size of the
layout is optimized compared to the original ConvexDraw algorithm. This
option reduces the grid size especially for graphs with few edges.
If the graph is (almost) triangulated, the grid size will be (almost)
the same.
\item |bool| {\bf side\_opt} = |false|\\
a boolean value determining if it is allowed to change
the shape of the outerface (from a triangle) in order to improve the
grid size of the layout.
\item |double| {\bf base\_ratio} $= 0.33$\\
a double value determining, how many nodes are drawn on the base line.
If |base_ratio| is $r$ and the outerface is $F$, the number of nodes
on the base line is $\max\left(2,\min\left(\ell, r\Labs{F}\right)\right)$,
where $\ell = \max\{k \mid v_1,\ldots,v_k$ is a path on $F_{out}$ with
$(v_i,v_j)\notin E$ for all $j\neq i+1 \}$.
\end{options}\horizline
}*/

public:

	/*{\Mcreation L 5}*/

	ConvexDrawLayout ();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	virtual ~ConvexDrawLayout() { };


	/*{\Moperations 1.5 2.5}*/

	leda::string name () const {
		return "ConvexDraw";
	}

	leda::string long_name () const {
		return "ConvexDraw (Chrobak/Kant)";
	}

	leda::string author () const {
		return "M. Chrobak, G. Kant";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "2.0";
	}

	leda::string impl_date () const {
		return "March 1997";
	}


	AGDModule *clone () const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base classes (|LayoutModule|, |GridLayoutModule|).}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	virtual void call (const graph& G, face f, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, face f, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	IRect bounding_box ();
	*/
	/*{\Mop}*/

	/* inherited
	void set_mapper (const GridCoordinateMapper& gcm);
	*/
	/*{\Mop}*/
	
	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	bool size_opt() const {
		return the_size_opt;
	}
	/*{\Mop}*/

	void size_opt (bool so) {
		the_size_opt = so;
	}
	/*{\Mop}*/

	bool side_opt() const {
		return the_side_opt;
	}
	/*{\Mop}*/

	void side_opt (bool so) {
		the_side_opt = so;
	}
	/*{\Mop}*/

	double base_ratio() const {
		return the_base_ratio;
	}
	/*{\Mop}*/
	
	void base_ratio (double br) {
		the_base_ratio = br;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/


protected:
	void get_node_grid_size (PlaneGraphCopyGridConcept &, const LayoutInterface &) { }
	void get_node_grid_size (PlaneGraphCopyGridConcept &C, const GridLayout &) {
		C.set_zero_size();
	}

	void do_call (PlaneGraphCopyGridConcept &C);

private:
	bool the_size_opt;
	bool the_side_opt;
	double the_base_ratio;

	void compute_coordinates (PlaneGraphCopyGridConcept &C,
		CanonicalOrder &c_o);
};

/*{\Mimplementation
The computation of the layout takes time $\BigO(n)$ where,
$n$ is the number of nodes of the input graph.}*/


#endif
