/*******************************************************************************
+
+  AGD 1.2
+
+
+  FixedGridMapper.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       FixedGridMapper.h
//
//    author: Carsten Gutwenger
//    date:   january 1999
//-------------------------------------------------------------------


#ifndef _AGD_FIX_GRID_MAPPER_H
#define _AGD_FIX_GRID_MAPPER_H


#include <AGD/GridCoordinateMapper.h>


/*{\Manpage {FixedGridMapper} {} {Mapping with Fixed Grid Width} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{0}{1.5}{7.5}{GridCoordinateMapper}
\arhook{7.5}{2}{8.5}
\hdclass{3}{0}{6}{FixedGridMapper}
\end{hierarchy}
}*/

class __exportAGD FixedGridMapper : public GridCoordinateMapper {

/*{\Mdefinition
|\Mname| provides a mapping with predefined grid width |width_x| in
x-direction and |width_y| in y-direction.

The input grid sizes of nodes ({\sl get\_node\_gride\_size}() function) are
\begin{eqnarray*}
|grid_width|(v) & = & \max\left( 0,\left\lceil \frac{|width|(v)+|sep_x|}
{|width_x|}-1\right\rceil\right) \\
|grid_height|(v) & = & \max\left( 0,\left\lceil \frac{|height|(v)+|sep_y|}
{|width_y|}-1\right\rceil\right)
\end{eqnarray*}

The nodes are enlarged, such that a minimal overhang is guaranteed
\begin{eqnarray*}
|width|'(v) & = & \max(|width|(v),2|min_overhang_x|+|grid_width(v)|*|width_x|) \\
|height|'(v) & = & \max(|height|(v),2|min_hang_y|+|grid_height(v)|*|width_y|)
\end{eqnarray*}
|\Mname| realizes the following mapping functions:
\begin{eqnarray*}
\varphi_x(|column|) & = & (|column| - |xmin|) * |width_x| + |origin.x| \\
\varphi_y(|row|) & = & (|row| - |ymin|) * |width_y| + |origin.y|
\end{eqnarray*}
The positions of nodes and the polylines of edges are given by the
default implementations of |GridCoordinateMapper| using the mapping
functions defined above.

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |double| {\bf width\_x} = 40\\
determines the predefined grid width in x-direction.
\item |double| {\bf width\_y} = 40\\
determines the predefined grid width in y-direction.
\item |double| {\bf min\_overhang\_x} = 2\\
determines the minimal allowed overhang of each node in x-direction.
\item |double| {\bf min\_overhang\_y} = 2\\
determines the minimal allowed overhang of each node in y-direction.
\item |double| {\bf sep\_x} = 20\\
determines the minimal distance (separation) between nodes in x-direction.
\item |double| {\bf sep\_y} = 20\\
determines the minimal distance between nodes in y-direction.
\item |DPoint| {\bf origin} = $(0,0)$\\
determines the origin of the coordinate system, that is the point to which
the lower left grid point $(|xmin|,|ymin|)$ is mapped.
\end{options}\horizline
}*/

public:

	/*{\Mcreation}*/

	FixedGridMapper (double width_x = 40, double width_y = 40,
		double min_overhang_x = 2, double min_overhang_y = 2,
		double sep_x = 20, double sep_y = 20,
		DPoint origin = DPoint(0,0)) :
		_wx(width_x), _wy(width_y), _min_over_x(min_overhang_x), _min_over_y(min_overhang_y),
		_sep_x(sep_x), _sep_y(sep_y), _origin(origin) { }
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|, and sets
	the optional parameters |width_x|, |width_y|, |min_overhang_x|,
	|min_overhang_y|, |sep_x|, |sep_y| and |origin|.}*/

	virtual ~FixedGridMapper() { }


	/*{\Moperations}*/

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|GridCoordinateMapper|).
	}*/

	/* inherited
	void get_node_grid_size(const leda::graph& G, GridLayout &gl, const LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void get_node_grid_size(const GraphCopy& GC, GridLayout &gl, const LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const leda::graph& G, const GridLayout &gl, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const GraphCopy& GC, const GridLayout &gl, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	double width_x() const {
		return _wx;
	}
	/*{\Mop}*/

	void width_x (double wx) {
		_wx = wx;
	}
	/*{\Mop}*/

	double width_y() const {
		return _wy;
	}
	/*{\Mop}*/

	void width_y (double wy) {
		_wy = wy;
	}
	/*{\Mop}*/

	double min_overhang_x() const {
		return _min_over_x;
	}
	/*{\Mop}*/

	void min_overhang_x (double x) {
		_min_over_x = x;
	}
	/*{\Mop}*/

	double min_overhang_y() const {
		return _min_over_y;
	}
	/*{\Mop}*/

	void min_overhang_y (double y) {
		_min_over_y = y;
	}
	/*{\Mop}*/

	double sep_x() const {
		return _sep_x;
	}
	/*{\Mop}*/

	void sep_x (double x) {
		_sep_x = x;
	}
	/*{\Mop}*/

	double sep_y() const {
		return _sep_y;
	}
	/*{\Mop}*/

	void sep_y (double y) {
		_sep_y = y;
	}
	/*{\Mop}*/

	DPoint origin() const {
		return _origin;
	}
	/*{\Mop}*/

	void origin (DPoint p) {
		_origin = p;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/


	AGDModule *clone() const {
		return new FixedGridMapper (_wx,_wy,_min_over_x,_min_over_y,
			_sep_x,_sep_y,_origin);
	}


protected:
	void do_get_node_grid_size(const LayoutInterface &A);

	void init (LayoutInterface &A);

	double map_x (int column) const {
		return _wx * (column-xmin) + _origin.xcoord();
	}

	double map_y (int row) const {
		return _wy * (row-ymin) + _origin.ycoord();
	}

	double xmin, ymin;
	double _wx, _wy, _min_over_x, _min_over_y, _sep_x, _sep_y;
	DPoint _origin;
};

/*{\Mimplementation
The get\_node\_size() function takes time $\BigO(n)$ and
the call() function takes time $\BigO(n+m+b)$, where $n$ is
the number of nodes and $m$ is the number of edges of $G$, and $b$ is the
total number of bend points.
}*/


#endif
