/*******************************************************************************
+
+  AGD 1.2
+
+
+  GraphCopy.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         GraphCopy.h
//
//    author: Carsten Gutwenger
//    date:   january 2000
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_GRAPH_COPY_H
#define _AGD_GRAPH_COPY_H


#include <LEDA/graph/graph.h>
#include <AGD/basic.h>
//#include <AGD/impl/node_map.h>
//#include <AGD/impl/edge_map.h>


/*{\Manpage {GraphCopy} {} {Graph Copies} {GC} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5}{graph}
\arhook{5}{2}{6}
\hdclass{1.5}{0}{5}{GraphCopy}
\end{hierarchy}
}*/

class __exportAGD GraphCopy : public leda::graph {

/*{\Mdefinition
An instance |\Mvar| of type |\Mname| is a copy of an original graph |G|.
For each node |v'| of |\Mvar|, |\Mname| stores the corresponding node
|original(v')| in the original graph |G|, for each node $v$ in |G| the
newly created copy |copy(v)| in |\Mvar|, for each edge |e'| in |\Mvar|
the corresponding edge |original(e')| in |G|, and for each edge $e$ in
$G$ a list of dummy edges |chain(e)| in |\Mvar|.

For nodes or edges which are created afterwards with new\_node or new\_edge,
the corresponding node or edge is set to |nil|.}*/

public:

	/*{\Mcreation GC 5.5}*/

	GraphCopy() : leda::graph() {
		g_orig = nil;
	}
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to an empty graph without original graph.}*/

	GraphCopy(const leda::graph& G);
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to a copy of |G|.}*/

	GraphCopy(const GraphCopy& GC);
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to a copy of $|original|(GC)$.}*/

	GraphCopy(const GraphCopy& GC, leda::node_array<leda::node> &v_copy,
		leda::edge_array<leda::edge> &e_copy);
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to a copy of $|original|(GC)$. Returns in |v_copy| the newly created
	nodes, i.e., $|v_copy|[v]$ is the node created for node $v\in|GC|$,
	and in |e_copy| the newly created edges, i.e., $|e_copy|[e]$ is the edge
	created for edge $e\in|GC|$.}*/

	virtual ~GraphCopy() { }


	/*{\Moperations}*/

	GraphCopy& operator=(const GraphCopy& GC);

	const leda::graph &original() const {
		return *g_orig;
	}
	/*{\Mop returns the original graph |G|.}*/

	leda::node original (leda::node v) const {
		return v_orig[v];
	}
	/*{\Mop returns the node of the original graph |G| that corresponds
	to |v|, |nil| if |v| was not in |G|, e.g. |v| was created afterwards.}*/

	leda::edge original (leda::edge e) const {
		return e_orig[e];
	}
	/*{\Mop returns the edge of the original graph |G| that corresponds
	to |e|, |nil| if |e| was not in |G|, e.g. |e| was created afterwards.}*/

	leda::node copy (leda::node v) const {
		return _copy[v];
	}
	/*{\Mop returns |copy(v)|. \precond $v\in G$.}*/

	const leda::list<leda::edge>& chain (leda::edge e) const {
		return _chain[e];
	}
	/*{\Mop returns the list of dummy edges for edge $e$.
	\precond $e$ is an edge in $G$.}*/

	bool is_dummy (leda::node v) const {
		return original(v) == nil;
	}
	/*{\Mop returns |true| iff $v$ has a corresponding node in the
	original graph.}*/
	
	bool is_dummy (leda::edge e) const {
		return original(e) == nil;
	}
	/*{\Mop returns |true| iff $e$ has a corresponding edge in the
	original graph.}*/

	bool is_reversed (leda::edge e) const {
		return source(e) != original(source(_chain[e].head()));
	}
	/*{\Mop returns |true| iff $e\in G$ has been reversed in |\Mvar|.}*/

	virtual leda::edge split (leda::edge e);
	/*{\Mop splits edge $e=(v,w)$ into edges $(v,u)$, $(u,w)$ by introducing
	a new node $u$. Returns the edge $(u,w)$.}*/

	virtual bool is_map() const {
		return false;
	}
	/*{\Mop virtual method returning |false|.}*/

	// add a new node with predefined original node
	leda::node add_node (leda::node orig) {
		leda::node v = new_node();
		v_orig[_copy[orig] = v] = orig;
		return v;
	}

protected:

	void map_to_copy (leda::node_array<leda::node>& v_copy,
		leda::edge_array<leda::edge>& e_copy, const leda::graph& G);
	leda::edge _first_adjacent (leda::node v);
	leda::edge _succ_adjacent (leda::edge e, leda::node v);

	leda::graph const *g_orig;

	leda::node_array<leda::node> _copy;
	leda::edge_map<leda::edge> e_orig;
	leda::node_map<leda::node> v_orig;

	leda::edge_map<leda::list<leda::edge> > _chain;
	leda::edge_map<leda::list_item> _in_chain;

	void _init_GC (const GraphCopy& GC);
	void _init_GC (const GraphCopy& GC, leda::node_array<leda::node> &v_copy,
		leda::edge_array<leda::edge> &e_copy);
};

/*{\Mimplementation
All operations for graphs are available for instances of
type |\Mname|, too. All additional operations take constant time.}*/


#endif
