/*******************************************************************************
+
+  AGD 1.2
+
+
+  GreedyInsertHeuristic.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                      GreedyInsertHeuristic.h
//
//
//    author: Michael Krger
//    date:   march 1998
//-------------------------------------------------------------------


#ifndef _AGD_GREEDY_INSERT_HEURISTIC_H
#define _AGD_GREEDY_INSERT_HEURISTIC_H


#include <AGD/TwoLayerCrossMin.h>
#include <LEDA/graph/graph.h>


class CrossingsMatrix;


/*{\Manpage {GreedyInsertHeuristic} {} {Greedy Insert} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{0.9}{1.5}{6.1}{TwoLayerCrossMin}
\arhook{7}{2}{8}
\hdclass{1.5}{0}{7}{GreedyInsertHeuristic}
\end{hierarchy}
}*/

class __exportAGD GreedyInsertHeuristic : public TwoLayerCrossMin {

/*{\Mdefinition
The module |\Mname| proceeds from left to right by successively
choosing the next vertex $v$ to be the one which minimizes the number of
crossings that edges adjacent to $v$ make with edges adjacent to 
vertices to the left of $v$ (see Eades and Kelly~\cite{EK86}). 

\horizline
\commoninfo{Greedy-Insert}{Greedy-Insert (Eades,Kelly)}{P. Eades, D. Kelly}{M. Kr\"uger, Th. Ziegler}
{March 1998}{1.0}

\begin{preposttab}
precondition & $\emptyset$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\horizline
}*/

public:

	/*{\Mcreation M 5}*/

	GreedyInsertHeuristic() { }
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|.}*/

	~GreedyInsertHeuristic() { }


	/*{\Moperations}*/

	leda::string name () const {
		return "Greedy-Insert";
	}

	leda::string long_name () const {
		return "Greedy-Insert (Eades,Kelly)";
	}

	leda::string author () const {
		return "P. Eades, D. Kelly";
	}

	leda::string impl_author () const {
		return "M. Kr\201ger, Th. Ziegler";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "March 1998";
	}

	AGDModule *clone() const {
		return new GreedyInsertHeuristic();
	}

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|TwoLayerCrossMin|).
	}*/

	/* inherited
	bool check (const leda::graph& G, const Layer& L, AgdKey &p) const;
	*/
	/*{\Mop}*/

	void init (const leda::graph& G, const Hierarchy& H);
	/*{\Mop}*/

	void call (const leda::graph& G, Layer& L);
	/*{\Mop}*/

	void cleanup ();
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

private:
	CrossingsMatrix *cp;
	leda::node_array<double> weight;
};


/*{\Mimplementation
Let $n_L$ denote the number of nodes on the changeable layer
and $m_L$ the number of edges between the fixed and the changeable layer.
The algorithm computes in a preprocessing step a so-called
crossing matrix, which takes time $\BigO(n_L m_L)$.
The actual crossing reduction algorithm takes time
$\BigO(n_L^2)$. Both steps are implemented in the
\texttt{call()}-method. The \texttt{init()}-method takes time
$\BigO(n)$, where $n$ is the number of nodes in the input graph.
}*/


#endif
