/*******************************************************************************
+
+  AGD 1.2
+
+
+  GreedySwitchHeuristic.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                      GreedySwitchHeuristic.h
//
//
//    author: Michael Krger
//    date:   march 1998
//-------------------------------------------------------------------


#ifndef _AGD_GREEDY_SWITCH_HEURISTIC_H
#define _AGD_GREEDY_SWITCH_HEURISTIC_H


#include <AGD/TwoLayerCrossMin.h>

class CrossingsMatrix;


/*{\Manpage {GreedySwitchHeuristic} {} {Greedy Switch} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{0.9}{1.5}{6.1}{TwoLayerCrossMin}
\arhook{7}{2}{8}
\hdclass{1.3}{0}{7.2}{GreedySwitchHeuristic}
\end{hierarchy}
}*/

class __exportAGD GreedySwitchHeuristic : public TwoLayerCrossMin {

/*{\Mdefinition
The greedy switch heuristic passes over all consecutive pairs of vertices and
switches them if this decreases the number of crossings (see Eades and Kelly~\cite{EK86}).

\horizline
\commoninfo{Greedy-Switch}{Greedy-Switch (Eades,Kelly)}{P. Eades, D. Kelly}{M. Kr\"uger, Th. Ziegler}
{March 1998}{1.0}

\begin{preposttab}
precondition & $\emptyset$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\horizline
}*/

public:

	/*{\Mcreation M 5}*/

	GreedySwitchHeuristic() { }
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|.}*/

	~GreedySwitchHeuristic() { }


	/*{\Moperations}*/

	leda::string name () const {
		return "Greedy-Switch";
	}

	leda::string long_name () const {
		return "Greedy-Switch (Eades,Kelly)";
	}

	leda::string author () const {
		return "P. Eades, D. Kelly";
	}

	leda::string impl_author () const {
		return "M. Kr\201ger, Th. Ziegler";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "March 1998";
	}


	AGDModule *clone() const {
		return new GreedySwitchHeuristic();
	}

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|TwoLayerCrossMin|).
	}*/

	/* inherited
	bool check (const leda::graph& G, const Layer& L, AgdKey &p) const;
	*/
	/*{\Mop}*/

	void init (const leda::graph& G, const Hierarchy& H);
	/*{\Mop}*/

	void call (const leda::graph& G, Layer& L);
	/*{\Mop}*/

	void cleanup ();
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

private:
	CrossingsMatrix *cp;
};


/*{\Mimplementation
Let $n_L$ denote the number of nodes on the changeable layer
and $m_L$ the number of edges between the fixed and the changeable layer.
The algorithm computes in a preprocessing step a so-called
crossing matrix, which takes time $\BigO(n_L m_L)$.
For the running time of the actual crossing reduction algorithm,
we quote the paper: ``The time complexity of this algorithm is
hard to compute, because it depends on the number of switchings
required to reach a stable ordering. However, since the total
number of crossings is at most $m^2$, we can at least
guarantee that the algorithm is $\BigO(n_L m_L^2)$.''
}*/


#endif
