/*******************************************************************************
+
+  AGD 1.2
+
+
+  GridLayoutModuleImpl.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                      GridLayoutModuleImpl.h
//
//    author: Carsten Gutwenger
//    date:   january 1999
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_GLM_IMPL_H
#define _AGD_GLM_IMPL_H


#include <AGD/GridLayoutModule.h>


/*{\Manpage {GridLayoutModuleImpl} {GRID_CONCEPT} {Implementation of Grid Layout Algorithms} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{11}{5.5}
\hdclass{4}{4.5}{5}{AGDModule}
\arhook{9}{5}{10}
\hdclass{5.5}{3}{5}{LayoutModule}
\arhook{10.5}{3.5}{11.5}
\hdclass{6}{1.5}{6}{GridLayoutModule}
\arhook{12}{2}{13}
\hdclass{0}{0}{13.5}{GridLayoutModuleImpl{\tt<}GRID\_CONCEPT{\tt>}}
\end{hierarchy}
}*/

template <class GRID_CONCEPT>
class __exportAGD GridLayoutModuleImpl : public GridLayoutModule {

/*{\Mdefinition
The class |\Mname| represents a generic method for implementing grid
layout algorithms. The template argument |GRID_CONCEPT| is the class
which represents the used implementation concept. It is responsible
for allocating the data structures used by the grid layout algorithm
and for translating the grid layout computed by the algorithm into
a layout or grid layout for the input graph. A class used as template
argument |GRID_CONCEPT| must provide the following constructors:
\begin{quote}
\begin{tabular}{p{0.5cm}p{12cm}}
\multicolumn{2}{l}{|GRID_CONCEPT|(const |graph| \&|G|,|LayoutInterface| \&|A|)} \\
& allocate data structures for graph $G$ with layout interface $A$.\\[2mm]
\multicolumn{2}{l}{|GRID_CONCEPT|(const |graph| \&|G|,|face| |f|,|LayoutInterface| \&|A|)} \\
& allocate data structures for planar map $G$ with exterior face $f$ and layout interface $A$.\\[2mm]
\multicolumn{2}{l}{|GRID_CONCEPT|(|PlaneGraphCopy| \&|PG|,|face| |f|,|LayoutInterface| \&|A|)} \\
& allocate data structures for plane graph copy |PG| with exterior face $f$ and layout interface $A$.\\[5mm]
\multicolumn{2}{l}{|GRID_CONCEPT|(const |graph| \&|G|,|GridLayout| \&|gl|)} \\
& allocate data structures for graph $G$ with grid layout |gl|.\\[2mm]
\multicolumn{2}{l}{|GRID_CONCEPT|(const |graph| \&|G|,|face| |f|,|GridLayout| \&|gl|)} \\
& allocate data structures for planar map $G$ with exterior face $f$ and grid layout |gl|.\\[2mm]
\multicolumn{2}{l}{|GRID_CONCEPT|(|PlaneGraphCopy| \&|PG|,|face| |f|,|GridLayout| \&|gl|)} \\
& allocate data structures for plane graph copy |PG| with exterior face $f$ and grid layout |gl|.
\end{tabular}
\end{quote}
These constructors correspond to call() and call\_grid() methods inherited
by |LayoutModule| and |GridLayoutModule|. If an exterior face $f$ is
passed, it may be |nil| indicating that the algorithm chooses an
appropriate exterior face. Moreover, the following operations are required:
\begin{quote}
\begin{tabular}{p{0.5cm}p{12cm}}
\multicolumn{2}{l}{void get\_node\_grid\_size(|GridCoordinateMapper| \&|gcm|,const |LayoutInterface| \&|A|)}\\
& determines initial grid size of nodes, where |gcm| is the current grid
coordinate mapper.\\[2mm]
\multicolumn{2}{l}{void get\_node\_grid\_size(const |GridLayout| \&|gl|)}\\
& determines inital grid size of nodes. The input of the user is given
by the node grid sizes in |gl|.\\[5mm]
\multicolumn{2}{l}{|IRect| get\_bbox()}\\
& returns the bounding box of the computed grid layout.\\[5mm]
\multicolumn{2}{l}{void update(|GridCoordinateMapper| \&|gcm|,|LayoutInterface| \&|A|)}\\
& transforms the computed grid layout into a layout for the input graph,
where |gcm| is the current grid coordinate mapper and |A| is a layout
interface for the input graph.\\[2mm]
\multicolumn{2}{l}{void update(|GridLayout| \&|gl|)}\\
& transforms the computed grid layout into a grid layout |gl| for the
input graph.\\
\multicolumn{2}{l}{static void set\_postcondition(|list<AgdKey>| \&|post|)} \\
& assigns the general postcondition to |post|, i.e., adds |key::planar_map_callable|
to post if calls for planar maps (second and fifth constructor) are supported, and adds
|key::PlaneGraphCopy_callable| to |post| if calls for |PlaneGraphCopy|'s (third
and sixth constructor) are supported.
\end{tabular}
\end{quote}
|\Mname| implements all call() and call\_grid() methods inherited by |LayoutModule|
and |GridLayoutModule|. The computation of a grid layout (call\_grid() methods)
works as follows:
\begin{enumerate}
\item An object |C| of type |GRID_CONCEPT| is constructed using the
constructor corresponding to the call\_grid() parameter list.
\item The initial grid node sizes are determined by calling
virtual method get\_node\_grid\_size$(C,|gl|)$.
\item The actual computation of the grid layout is performed by
calling pure virtual method do\_call$(C)$.
\item The bounding box of the grid layout is determined by calling
get\_bbox() of $C$.
\item The computed grid layout is transformed into a grid layout for
the input graph by calling update$(|gl|)$ of $C$.
\end{enumerate}
The following steps are preformed in order to compute a layout
(call() methods):
\begin{enumerate}
\item An object |C| of type |GRID_CONCEPT| is constructed using the
constructor corresponding to the call() parameter list.
\item The initial grid node sizes are determined by calling
virtual method get\_node\_grid\_size$(C,A)$.
\item The actual computation of the grid layout is performed by
calling pure virtual method do\_call$(C)$.
\item The bounding box of the grid layout is determined by calling
get\_bbox() of $C$.
\item The computed grid layout is transformed into a layout for the
input graph by calling update$(|gcm|,A)$ of $C$.
\end{enumerate}
Thus, it is sufficient to implement do\_call$(C)$. All necessary
transformations are performed by the grid implementation concept class.
The methods get\_node\_grid\_size()
must be overriden if the standard implementation of |GRID_CONCEPT|
is not suitable, e.g., if the algorithms assumes that each node
is just one grid point (all nodes have width and height 0).
}*/

public:
	
	GridLayoutModuleImpl() {
		leda::list<AgdKey> post;
		GRID_CONCEPT::set_postcondition(post);
		AgdKey p;
		forall(p,post) add_post_rule(p,0);
	}

	~GridLayoutModuleImpl() { }

	/*{\Moperations 1.5 4}*/
	/*{\Mtext |\Mname| implements the call() and call\_grid() methods
	of its base classes. Furthermore, there are the following protected
	overridables:}*/

	void call (const leda::graph& G, LayoutInterface& A) {
		GRID_CONCEPT C(G,A);
		get_node_grid_size(C,A);
		do_call(C);
		finish_call(C,A);
	}

	void call (const leda::graph& G, leda::face f, LayoutInterface& A) {
		if (!postcondition(0).member(key::planar_map_callable))
			leda::error_handler(1, "GridLayoutModuleImpl: algorithm not callable for planar maps!");
		GRID_CONCEPT C(G,f,A);
		get_node_grid_size(C,A);
		do_call(C);
		finish_call(C,A);
	}

	void call (PlaneGraphCopy& PG, leda::face f, LayoutInterface& A) {
		if (!postcondition(0).member(key::planar_map_callable))
			leda::error_handler(1, "GridLayoutModuleImpl: algorithm not callable for PlaneGraphCopy's!");
		GRID_CONCEPT C(PG,f,A);
		get_node_grid_size(C,A);
		do_call(C);
		finish_call(C,A);
	}

	void call_grid (const leda::graph& G, GridLayout &gl) {
		GRID_CONCEPT C(G,gl);
		get_node_grid_size(C,gl);
		do_call(C);
		finish_call_grid(C,gl);
	}

	void call_grid (const leda::graph& G, leda::face f, GridLayout &gl) {
		if (!postcondition(0).member(key::planar_map_callable))
			leda::error_handler(1, "GridLayoutModuleImpl: algorithm not callable for planar maps!");
		GRID_CONCEPT C(G,f,gl);
		get_node_grid_size(C,gl);
		do_call(C);
		finish_call_grid(C,gl);
	}

	void call_grid (PlaneGraphCopy& PG, leda::face f, GridLayout &gl) {
		if (!postcondition(0).member(key::planar_map_callable))
			leda::error_handler(1, "GridLayoutModuleImpl: algorithm not callable for PlaneGraphCopy's!");
		GRID_CONCEPT C(PG,f,gl);
		get_node_grid_size(C,gl);
		do_call(C);
		finish_call_grid(C,gl);
	}

protected:

	virtual void do_call (GRID_CONCEPT &C) = 0;
	/*{\Moppure implements the algorithm.}*/

	virtual void get_node_grid_size (GRID_CONCEPT &C, const LayoutInterface &A) {
		C.get_node_grid_size(mapper.get(),A);
	}
	/*{\Mop determines initial grid sizes of nodes. The default implementation
	calls C.get\_node\_grid\_size() with the current grid coordinate mapper.}*/

	virtual void get_node_grid_size (GRID_CONCEPT &C, const GridLayout &gl) {
		C.get_node_grid_size(gl);
	}
	/*{\Mop determines initial grid sizes of nodes. The grid sizes in |gl|
	represent the user's input. The default implementation
	calls C.get\_node\_grid\_size() with |gl|.}*/


	void finish_call(GRID_CONCEPT &C, LayoutInterface &A) {
		_grid_bbox = C.get_bbox();
		C.update(mapper.get(),A);
	}

	void finish_call_grid(GRID_CONCEPT &C, GridLayout &gl) {
		_grid_bbox = C.get_bbox();
		C.update(gl);
	}

};

/*{\Mexample
The following example presents a simplified version of the
PlanarStraight algorithm. It requires as input a planar map
and adds eventually new edges in order to make the graph
biconnected. The output is a layout on the grid, where we
assume that each node is represented by a single grid point.
Therefore, we chose |PlaneGraphCopyGridConcept| as implementation
concept class.
\small\tt
\begin{verbatim}
#include <AGD/GridLayoutModuleImpl.h>
#include <AGD/PlaneGraphCopyGridConcept.h>
#include <AGD/AGDModuleOption.h>
#include <AGD/AugmentationModule.h>

class PlanarStraightLayout : public GridLayoutModuleImpl<PlaneGraphCopyGridConcept> {
public:
  PlanarStraightLayout ();
  ~PlanarStraightLayout() { }

  bool size_opt() const {
    return _size_opt;
  }

  void size_opt (bool so) {
    _size_opt = so;
  }

  AGDModule *clone () const;
  AGD_DECLARE_MODULE (AugmentationModule, augmenter)

protected:
  void get_node_grid_size (PlaneGraphCopyGridConcept &C, const LayoutInterface &A) { }
  void get_node_grid_size (PlaneGraphCopyGridConcept &C, const GridLayout &gl) {
    C.set_zero_size();
  }

  void do_call (PlaneGraphCopyGridConcept &C);

private:
  bool _size_opt;
};
\end{verbatim}\normalsize\rm
|PlanarStraightLayout| declares one optional parameter |size_opt|
and one module option |augmenter|. The get\_node\_grid\_size()
methods are overridden such that the size of each node is initially
a single grid point. The constructor declares the pre- and postcondition
and initializes |size_opt| and |augmenter|:
\small\tt
\begin{verbatim}
#include <AGD/PlanarStraightLayout.h>
#include <AGD/PlaneGraphCopy.h>
#include <AGD/PlanAug.h>

PlanarStraightLayout::PlanarStraightLayout () : _size_opt(true)
{
  add_precondition (key::planar);
  add_precondition (key::simple);
  add_precondition (key::no_self_loops);

  add_post_rule (key::straight_line);
  add_post_rule (key::no_crossings);

  PreCond  PRE;
  PostCond POST;
  PRE [1] << key::planar << key::simple << key::no_self_loops
    << key::directed;
  POST[1] << key::planar << key::biconnected << key::simple
    << key::no_self_loops;
  augmenter.init (PRE,POST,PlanAug());
}
\end{verbatim}\normalsize\rm
The clone() method returns a new object of type |PlanarStraightLayout|
with identical options:
\small\tt
\begin{verbatim}
AGDModule *PlanarStraightLayout::clone() const
{
  PlanarStraightLayout *M = new PlanarStraightLayout;

  M->size_opt (size_opt());

  M->set_augmenter (augmenter.get());
  M->set_mapper (mapper.get());

  return M;
}
\end{verbatim}\normalsize\rm
The actual computation is done by the do\_call() method. Of course,
we omit here the complete code of |PlanarStraightLayout|::do\_call().
The example simply demonstrates, how the graph and the grid layout
are accessed via |PlaneGraphCopyGridConcept|.
\small\tt
\begin{verbatim}
void PlanarStraightLayout::do_call (PlaneGraphCopyGridConcept &C)
{
  PlaneGraphCopy &G  = C.get_graph ();
  GridLayout     &gl = C.get_layout();

  ...
}
\end{verbatim}\normalsize\rm
}*/


#endif
