/*******************************************************************************
+
+  AGD 1.2
+
+
+  LayoutModule.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         LayoutModule.h
//
//
//    author: Carsten Gutwenger
//    date:   january 2000
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_LAYOUT_MODULE_H
#define _AGD_LAYOUT_MODULE_H


#include <AGD/AGDModule.h>


class LayoutInterface;
class PlaneGraphCopy;


/*{\Manpage {LayoutModule} {} {Layout Algorithms} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5}{AGDModule}
\arhook{5}{2}{6}
\hdclass{1.5}{0}{5}{LayoutModule}
\end{hierarchy}
}*/

class __exportAGD LayoutModule : public AGDModule {

/*{\Mdefinition
The class |\Mname| defines the standard interface for layout algorithms.

A layout algorithm |\Mvar| gets as input a graph $G$ and for each node $v$
of $G$ the size of the rectangle surrounding the graphics of $v$. |\Mvar|
computes for each node $v$ a point, which will be the center of the
graphics of $v$, and for each edge a list of bend points with a
source and target anchor point. Layout algorithms are also allowed to change
the size of the graphics of nodes.

Layout algorithms can be implemented by directly deriving from |\Mname|
and overriding all pure virtual methods of the standard interface. More
convenient is to derive from |LayoutModuleImpl| with a suitable implementation
concept class. This is explained in detail in Chapter~\ref{Chapter Implementation of Layout Algorithms}.
In the latter case, only one method (\texttt{do\_call()}) has to be overridden.
If the layout algorithm places vertices and bend points on integer grid
coordinates, the algorithm should be implemented as grid layout algorithm
(see |GridLayoutModule|, Sect.~\ref{Grid Layout Algorithms}).

\headerline{Input and Output Parameters}

\begin{tabular}{lcc}
input parameter: & graph & |in_graph=1| \\
output parameter: & layout & |out_layout=1|
\end{tabular}

The integer constants |in_graph| and |out_layout| denote the
positions in the input and output parameter lists. The values can be used to
reference the desired parameter when setting the pre- and postcondition.  
}*/

public:

	// the input and output parameters of a layout algorithm module
	static const int in_graph;
	static const int out_layout;

	/*{\Minitialization}*/

	LayoutModule ()  : AGDModule(1,1) { }
	/*{\Mcreate initializes a layout algorithm.}*/

	virtual ~LayoutModule () { }


	/*{\Moperations 1.2 3.5}*/

	/*{\Mtext
	\headerline{Standard Interface} }*/

	virtual bool check (const leda::graph& G, AgdKey &p) const;
	/*{\Mop returns |true| if $G$ satisfies the preconditions of |\Mvar|.
	Otherwise, |false| is returned and |p| contains a property that
	is not satisfied.}*/

	virtual void call (const leda::graph& G, LayoutInterface& A) = 0;
	/*{\Moppure calls |\Mvar| for graph $G$ with layout
	interface $A$. \precond $A$ is a layout interface for $G$.}*/


	/*{\Mtext\smallskip
	The following call interfaces are not supported by all layout algorithms.
	The general postcondition (|\Mvar|.postcondition(0)) declares, which
	call interfaces are supported.}*/

	virtual void call (const leda::graph& G, leda::face f, LayoutInterface& A) = 0;
	/*{\Moppure calls |\Mvar| for graph $G$ with layout interface
	|A|. \precond $|key::planar_map_callable|\in |post|(0)$,
	$G$ represents a planar map, $f$ is |nil|
	or a face in $G$, and $A$ is a layout interface for $G$.}*/

	virtual void call (PlaneGraphCopy& PG, leda::face f, LayoutInterface& A) = 0;
	/*{\Moppure calls |\Mvar| for plane graph copy $PG$ with layout
	interface |A|. \precond $|key::PlaneGraphCopy_callable|\in|post|(0)$,
	$f$ is |nil| or a face in $PG$, and $A$ is a
	layout interface for |PG|.}*/
};



#endif
