/*******************************************************************************
+
+  AGD 1.2
+
+
+  NaiveMaximalPlanarSubgraph.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         NaiveMaximalPlanarSubgraph.h
//
//
// author: Thomas Ziegler
//        
// date:   july 2000
//-------------------------------------------------------------------


#ifndef _AGD_NAIVE_MAXIMAL_PLANAR_SUBGRAPH_H
#define _AGD_NAIVE_MAXIMAL_PLANAR_SUBGRAPH_H


#include <AGD/SubgraphModule.h>

/*{\Manpage {NaiveMaximalPlanarSubgraph} {} {Naive Maximal Planar Subgraphs} {S} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{1}{1.5}{6}{SubgraphModule}
\arhook{7}{2}{8}
\hdclass{2.5}{0}{10}{NaiveMaximalPlanarSubgraph}
\end{hierarchy}
}*/

class __exportAGD NaiveMaximalPlanarSubgraph : public SubgraphModule {

/*{\Mdefinition
The module |\Mname| computes a maximal planar subgraph of an
input graph $G$ by the following simple strategy: 
We start with the empty graph and construct
a maximal planar subgraph by trying to add the edges of $G$ 
one after the other in an arbitrary order. 
For each edge we use a planarity testing algorithm to check if
the current subgraph becomes non-planar after adding the edge.
If the resulting graph is still planar, the edge is added, otherwise,
the edge is discarded.

\horizline
\commoninfo{Naive Maximal Planar Subgraph}{Naive Maximal Planar Subgraph}{}{T. Ziegler}
{July 2000}{}

\begin{preposttab}
precondition & $\emptyset$ \\
postcondition(|PRE|) & $\{\, \emph{planar} \,\}
\cup \Big( \{\, \emph{simple, no self loops, no reversals} \,\} \cap |PRE| \Big)$
\end{preposttab}
\horizline

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |int| {\bf runs} $= 3$\\
It is not difficult to see that the result of the algorithm
depends on the order in which the edges are considered. The parameter
determines, how many different insertion orders are tested. The insertion
orders are obtained by randomly permuting the list of all edges of the 
given graph. The result is the largest computed maximal planar subgraph. 

\end{options}
\horizline
}*/

public:
	/*{\Mcreation S 4.5}*/

	NaiveMaximalPlanarSubgraph();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	virtual ~NaiveMaximalPlanarSubgraph() { }

	/*{\Moperations}*/

	leda::string name () const {
		return leda::string ("Naive Maximal Planar Subgraph");
	}

	leda::string author () const {
		return "";
	}

	leda::string impl_author () const {
		return leda::string ("Thomas Ziegler");
	}

	leda::string impl_date () const {
		return "July 2000";
	}


	AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|SubgraphModule|).
	}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	bool call (const graph& G, list<edge>& L);
	*/
	/*{\Mop}*/

	/* inherited
	bool call (graph& G);
	*/
	/*{\Mop}*/

	/* inherited
	bool call (GraphCopy& G, list<edge>& L);
	*/
	/*{\Mop}*/

	/* inherited
	int num_del_edges() const;
	*/
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	int runs() const {
		return _runs;
	}
	/*{\Mop}*/

	void runs(int n) {
		if (n >= 1) _runs = n;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/


protected:
	void do_call (const leda::graph& G, leda::list<leda::edge>& L);

	int _runs;
};

/*{\Mimplementation
The algorithm has running time $\BigO(m(n+m))$, where $n$ is the
number of nodes and $m$ is the number of edges of the input graph $G$.
}*/


#endif
