/*******************************************************************************
+
+  AGD 1.2
+
+
+  OrthogonalLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//#line 654 "Lweb/OrthoLayout.lw"
#ifndef _AGD_ORTHO_H 
#define _AGD_ORTHO_H

#include <AGD/Ovid.h>

/*{\Moptions
outfile=OrthogonalLayout.man
}*/

/*{\Manpage {OrthogonalLayout} {} {The Orthogonal Drawing Algorithm} {O} }*/

/*{\Mbaseclasses
\begin{hierarchy}{9.5}{7}
\hdclass{0}{6}{5}{AGDModule} \arhook{5}{6.5}{6}
\hdclass{1.5}{4.5}{5}{LayoutModule} \arhook{6.5}{5}{7.5}
\hdclass{2}{3}{6}{GridLayoutModule} \arhook{8}{3.5}{9}
\hdclass{6.5}{1.5}{3}{Ovid} \arhook{9.5}{2}{10.5}
\hdclass{6}{0}{8}{OrthogonalLayout}
\end{hierarchy}
\hfill\agdeps{OrthogonalLayout.eps}{5.5cm} 
}*/

class __exportAGD OrthogonalLayout : public Ovid {

/*{\Mdefinition 
  The class |\Mname| is an implementation of an algorithm extending
  Tamassia's bend minimizing algorithm to the class of planar graphs with
  arbitrary degree~\cite{KM98,Klau97}. It differs from the implementation 
  of the class |QuasiOrthogonalLayout| only in its output style.

  In an orthogonal drawing, vertices of degree greater than four are 
  drawn as boxes whose corner points are mapped to grid points. The internal
  computations are the same as in the classes |QuasiOrthogonalLayout| and 
  |PureOrthogonalLayout|.

  \horizline

  \commoninfo{Orthogonal}{Orthogonal Layout}{G.~W.~Klau, P.~Mutzel}{G.~W.~Klau}{April 1999}{1.1}

  \headerline{Pre- and Postcondition}
  
  \begin{tabular}{lcl}
  precondition & $=$ & \{\,$\emph{planar, simple, no self loops}$\,\} \\
  postcondition(|PRE|) & $=$ & $\{\,$\emph{no crossings, orthogonal}$\,\}$
  \end{tabular}\\

  \headerline{Exchangeable Modules}
  Instances of type |\Mname| provide the following module options:
  \begin{options}
  \item {\bf compactor}\\
  a |CompactionModule| option used for the compaction phase.
  
  \begin{tabular}{ll}
  guaranteed precondition: & \emph{no crossings, orthogonal}\\
  required postcondition: &  \emph{no crossings, orthogonal}\\
  initial module: & |FlowCompaction|
  \end{tabular}
  \end{options}

  \headerline{Optional Parameters}
  Instances of |\Mname| provide the following optional parameters:
  \begin{options}
  \item |int| {\bf output\_level} $= 0$\\
  The |output_level| determines the amount of output to |stdout| that
  is produced by the algorithm. The default value, zero, causes |\Mname| to
  run in silent mode, values $\geq 10$ cause the maximal amount of output.
  
  \item |bool| {\bf verify} = |true|\\
  If this flag is set, the checking routines are activated. Especially the orthogonal
  representation is verified, 
  certificating the correctness of the computation.
  \end{options}
  \horizline
  }*/
               
public: 

    /*{\Mcreation O 5}*/
                
    OrthogonalLayout(); 
    /*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/
               
    virtual ~OrthogonalLayout() { }
               
    /*{\Moperations 1.5 2.5}*/

    leda::string name() const { return leda::string("OrthogonalLayout"); }
               
    leda::string long_name() const { 
      return leda::string("Orthogonal Drawing Algorithm"); 
    }
  
    leda::string author() const { return leda::string("G.~W.~Klau, P.~Mutzel"); }
  
    virtual leda::string impl_author() const { return leda::string("G.~W.~Klau"); }
  
    virtual leda::string impl_version() const { return leda::string("1.1"); }
  
    virtual leda::string impl_date() const { return leda::string("April 1999"); }

    AGDModule* clone() const;

    /*{\Moptions warnings=no}*/
    
    /*{\Mtext       
      \headerline{Standard Interface (Inherited Methods)}
      The detailed description of these methods can be found in the
      manual entries of the base classes (|LayoutModule|, |GridLayoutModule|).
      \stdstart{1.5cm}}*/
    
    /* inherited
       bool check(const graph& G, string& message);
    */
    /*{\Mop}*/
    
    /* inherited
       void call (const graph& G, LayoutInterface& A);
    */
    /*{\Mop}*/
    
    /* inherited
       virtual void call (const graph& G, face f, LayoutInterface& A);
    */
    /*{\Mop}*/
    
    /* inherited
       virtual void call (PlaneGraphCopy& PG, face f, LayoutInterface& A);
    */
    /*{\Mop}*/
    
    /* inherited
       void call_grid (const graph& G, GridLayout &gl);
    */
    /*{\Mop}*/
    
    /* inherited
       void call_grid (const graph& G, face f, GridLayout &gl);
    */
    /*{\Mop}*/
    
    /* inherited
       void call_grid (PlaneGraphCopy& PG, face f, GridLayout &gl);
    */
    /*{\Mop}*/
    
    /* inherited
       IRect bounding_box ();
    */
    /*{\Mop}*/
    
    /* inherited
       void set_mapper (const GridCoordinateMapper& gcm);
    */
    /*{\Mop}*/
    
    /*{\Mtext \stdend}*/
        
protected:

    void preprocessing(OrthogonalRepresentation &H);
    void postprocessing(OrthogonalRepresentation &H, GridLayout &L);
};

/*{\Mimplementation The time to compute an orthogonal drawing is
  $O((n \log n)^2)$ where $n$ is the number of vertices in the graph to be drawn. It is thus
  asymptotically equal to the time to compute a [[PureOrthogonalLayout]].
}*/

#endif

