/*******************************************************************************
+
+  AGD 1.2
+
+
+  PlanAug.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         PlanAug.h
//
//    author: Sergey Fialko, (AGD-Anpassung: Carsten Gutwenger)
//    date:   august 1998
//-------------------------------------------------------------------


#ifndef _AGD_PLAN_AUG_H
#define _AGD_PLAN_AUG_H


#include <AGD/AugmentationModule.h>


/*{\Manpage {PlanAug} {} {Planar Biconnected} {A} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{0.3}{1.5}{6.7}{AugmentationModule}
\arhook{7}{2}{8}
\hdclass{3.5}{0}{5}{PlanAug}
\end{hierarchy}
}*/

class __exportAGD PlanAug : public AugmentationModule {

/*{\Mdefinition
Augmenting a planar graph to a planar biconnected graph by adding 
the minimum number of edges is an NP-hard optimization problem.
Fialko and Mutzel~\cite{FM98} have suggested a $5/3$-approximative polynomial 
time algorithm for the planar (biconnectivity) augmentation problem. 
An implementation of the slightly modified algorithm is presented 
in module |\Mname|.

The algorithm first constructs a block tree $T$ from the given graph and
then iteratively adds edges between blocks of degree one in $T$ (in a clever 
way). After each step an update is necessary.
Although the approximation factor of $5/3$ is not a good one, our experiments
on a benchmark set of graphs have shown that in around $96\%$ of all the
cases our algorithm has found the optimal solution to the planar augmentation problem.

\horizline
\commoninfo{Planar Biconnected}{Planar Biconnected}{S.Fialko, P. Mutzel}{S. Fialko}
{October 1997}{}

\begin{preposttab}
precondition & $\{\, \emph{planar} \,\}$ \\
postcondition(|PRE|) & $\{\, \emph{planar, biconnected} \,\}
\cup \Big( \{\, \emph{simple, no self loops} \,\} \cap |PRE| \Big)$
\end{preposttab}

\headerline{Optional Parameters}
Instances of |\Mname| provide two parameters, which vary the
optimization process. For a better solution, you should try
all four possible combinations perhaps many times with permutation
of the input graph. On small graphs with less than 100 nodes,
the results are in most cases identical.
\begin{options}
\item |bool| {\bf size\_strategy} = |false|\\
Setting of |size_strategy| to |false| gives statistical
better results, but differences for our test graphs were
minimal (less than 1\%).
\item |bool| {\bf use\_blabels} = |true|\\
Setting |use_blables| to |true| has better approximation
ratio (but not necessarily a better solution) and is a
little slower.
\end{options}\horizline
}*/

public:

	/*{\Mcreation}*/

	PlanAug();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	virtual ~PlanAug() { }


	/*{\Moperations}*/

	leda::string name () const {
		return "Planar Biconnected";
	}

	leda::string author () const {
		return "S.Fialko, P. Mutzel";
	}

	leda::string impl_author () const {
		return "S. Fialko";
	}

	leda::string impl_date () const {
		return "August 1998";
	}


	void parameters (bool use_blabels, bool size_strategy) {
	  _use_blabels = use_blabels;
	  _size_strategy = size_strategy;
	}

	int nPlanTests () {
	  return _nTests;
	}

	AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|AugmentationModule|).
	}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	bool call (graph& G);
	*/
	/*{\Mop}*/

	/* inherited
	int num_added_edges() const;
	*/
	/*{\Mop}*/

	/* inherited
	bool call (graph& G, list<edge>& L);
	*/
	/*{\Mop}*/
	
	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	bool size_strategy () const {
		return _size_strategy;
	}
	/*{\Mop}*/

	void size_strategy (bool b) {
		_size_strategy = b;
	}
	/*{\Mop}*/

	bool use_blabels () const {
	  return _use_blabels;
	}
	/*{\Mop}*/

	void use_blabels (bool b) {
	  _use_blabels = b;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

protected:
	void do_call (leda::graph& G, leda::list<leda::edge>& L);

private:
	bool _use_blabels;
	bool _size_strategy;
	int _nTests;
};


/*{\Mimplementation
The algorithm has running time $\BigO(n^3)$, where $n$ is the number of nodes
of the input graph.}*/


#endif
