/*******************************************************************************
+
+  AGD 1.2
+
+
+  PlanarSPQRTree.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                          PlanarSPQRTree.h
//
//
//    author: Carsten Gutwenger
//    date:   August 1999
//-------------------------------------------------------------------


#ifndef _AGD_PL_SPQRTREE_H
#define _AGD_PL_SPQRTREE_H

#include <AGD/SPQRTree.h>
#include <LEDA/numbers/integer.h>
#include <LEDA/core/array.h>
#include <LEDA/core/tuple.h>


/*{\Manpage {PlanarSPQRTree} {} {SPQR-Trees for Planar Graphs} {T} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5}{SPQRTree}
\arhook{5}{2}{6}
\hdclass{1}{0}{5.5}{PlanarSPQRTree}
\end{hierarchy}
}*/

class __exportAGD PlanarSPQRTree : public SPQRTree {

/*{\Mdefinition
The class |\Mname| maintains the triconnected components of a
planar biconnected graph $G$ and represents all possible embeddings
of $G$. Each skeleton graph is embedded, i.e., it is a planar map.

The current embeddings of the skeletons define an embedding of $G$.
There are two basic operations for obtaining another embedding
of $G$: reverse$(v)$, which flips the skeleton of an R-node $v$
around its poles, and swap$(v,e_1,e_2)$, which exchanges the
positions of the edges $e_1$ and $e_2$ in the skeleton of a
P-node $v$.
}*/

public:

	/*{\Mcreation}*/

	PlanarSPQRTree(const leda::graph &G, bool is_planar_map = false) :
		SPQRTree(G,is_planar_map)
	{
		init(G,is_planar_map);
	}
	/*{\Mcreate creates an SPQR tree ${\calT}$ for planar graph $G$ rooted at
	the first edge of $G$. If |is_planar_map| is set to |true|, $G$ must be a
	planar map, i.e., for each edge $e$ the reversal information |rev(e)|
	is set and the counter-clockwise order of the outgoing edges of each
	vertex defines an embedding. In this case, $e$ and |rev(e)| are considered as a single
	edge. \precond $G$ is planar and biconnected and contains at least 3 nodes,
	or |G| has exactly 2 nodes and at least 3 (resp.\ 6 if |is_planar_map|
	is true) edges.}*/

	PlanarSPQRTree(const leda::graph &G, leda::edge e, bool is_planar_map = false) :
		SPQRTree(G,e,is_planar_map)
	{
		init(G,is_planar_map);
	}
	/*{\Mcreate creates an SPQR tree ${\calT}$ for planar graph $G$ rooted at edge $e$.
	If |is_planar_map| is set to |true|, $G$ must be a
	planar map, i.e., for each edge $e$ the reversal information |rev(e)|
	is set and the counter-clockwise order of the outgoing edges of each
	vertex defines an embedding. In this case, $e$ and |rev(e)| are considered as a single
	edge. \precond $G$ is planar and biconnected and contains at least 3 nodes,
	or |G| has exactly 2 nodes and at least 3 (resp.\ 6 if |is_planar_map|
	is true) edges.}*/

	~PlanarSPQRTree() { }


	/*{\Moperations}*/

	leda::integer number_of_embeddings() const {
		return number_of_embeddings(_root_node);
	}
	/*{\Mop returns the number of possible embeddings of $G$.}*/

	leda::integer number_of_embeddings(leda::node v) const;
	/*{\Mop returns the number of possible embeddings of the pertinent
	graph of node $v$. \precond $v\in{\calT}$}*/

	void reverse(leda::node v);
	/*{\Mop flips the skeleton $S$ of $v$ around its poles by reversing
	the order of outgoing edges of each vertex in $S$. \precond $v$
	is an R-node in $\calT$.}*/

	void swap(leda::node v, leda::edge e1, leda::edge e2);
	/*{\Mop exchanges the positions of |e1| and |e2| in skeleton of $v$.
	\precond $v$ is a P-node in $\calT$ and $|e1|,|e2|\in|skeleton|(v)$.}*/

	void embed(leda::graph &G);
	/*{\Mop embeds $G$ according to the current embeddings of the skeletons
	of $\calT$. \precond $G$ is the graph passed to the constructor of $\calT$
	and $G$ is a map.}*/

	void random_embed();
	/*{\Mop embeds all skeleton graphs randomly.}*/

	void random_embed(leda::graph &G) {
		random_embed();
		embed(G);
	}
	/*{\Mop embeds all skeleton graphs randomly, and embeds $G$ according to
	the embeddings of the skeletons. \precond $G$ is the graph passed to
	the constructor of $\calT$ and $G$ is a map.}*/

protected:

	void init(const leda::graph &G, bool is_planar_map);
	void rec_init_embed(leda::node vT, 
                leda::list<leda::two_tuple<leda::edge,leda::node> > &edges,
		leda::node_array<leda::list<leda::list_item> > &adj,
		const leda::edge_array<leda::list_item> &pos_in_adj);

	void swap_out(leda::graph &M, leda::edge e1, leda::edge e2);

	void expand_virtual_embed(leda::edge eT, leda::node v_orig, leda::list<leda::edge> &edges);
	void create_inner_vertices_embed(leda::node vT, leda::list<leda::edge> &edges);

	//static leda::random_source rs;
};


/*{\Mimplementation
Let $n$ be the number of vertices and $m$ be the number of edges in
$G$. The construction of an SPQR-tree $\calT$ for graph $G$
takes time $\BigO(n+m)$. Operations embed() and random\_embed()
take time $\BigO(n+m)$, reverse($v$) takes time
$\BigO(\Labs{|skeleton|(v)})$, and swap() takes constant time.
Operation number\_of\_embeddings() takes time $\BigO(n+m)$ and
number\_of\_embeddings(|vT|) takes time $\BigO(\Labs{P})$,
where $P$ is the pertinent graph of |vT|.
The data structure uses $\BigO(n+m)$ space.
}*/


#endif
