/*******************************************************************************
+
+  AGD 1.2
+
+
+  PlanarSubgraph.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         PlanarSubgraph.h
//
//
// author: Carsten Gutwenger, Thomas Ziegler (pre- and postprocessing)
//         Sebastian Leipert (actual algorithm)
// date:   novembre 1997
//-------------------------------------------------------------------


#ifndef _AGD_PLANAR_SUBGRAPH_H
#define _AGD_PLANAR_SUBGRAPH_H


#include <AGD/SubgraphModule.h>


/*{\Manpage {PlanarSubgraph} {} {Planar Subgraphs} {S} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{1}{1.5}{6}{SubgraphModule}
\arhook{7}{2}{8}
\hdclass{2.5}{0}{6}{PlanarSubgraph}
\end{hierarchy}
}*/

class __exportAGD PlanarSubgraph : public SubgraphModule {

/*{\Mdefinition
The module |\Mname| computes a planar subgraph of an
input graph $G$ by deleting (in general, a small number
of) edges using PQ-trees~\cite{Lei95,JLM96}.
The algorithm is similar to the one by Jayakumar et al.~\cite{JTS89}.
Notice that planarization algorithms using PQ-trees cannot guarantee 
to delete a minimum nor a minimal number of edges~\cite{JLM98b,JLM97}.
However, among the classical heuristics for the maximum planar subgraph 
problem, this is one of the best~\cite{Mut94}.

The idea of the algorithm is essentially to proceed as in the 
planarity testing algorithm based on PQ-trees. First, an st-numbering 
is constructed and then vertices are subsequently added to the empty graph. 
Together with the newly constructed graph,
a so-called bush form $T$ (represented as a PQ-tree) is updated. 
In order to add a new vertex $v$ without introducing crossings, 
 all the virtual edges labeled $v$ in $T$ need to be brought together.
If this fails, some of the edges corresponding to those virtual edges
are deleted.

\horizline
\commoninfo{Planar Subgraph}{Planar Subgraph}{R. Jayakumar et al., S. Leipert}{S. Leipert}
{November 1997}{}

\begin{preposttab}
precondition & $\emptyset$ \\
postcondition(|PRE|) & $\{\, \emph{planar} \,\}
\cup \Big( \{\, \emph{simple, no self loops, no reversals} \,\} \cap |PRE| \Big)$
\end{preposttab}
\horizline

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |int| {\bf runs} $= 3$\\
determines, how many times the actual planar subgraph algorithm is called.
Each repetition (except for the first) starts with randomly
permuted node lists. Deterministic behaviour can be achieved
by setting |runs| to $1$.
\end{options}
\horizline
}*/

public:

	/*{\Mcreation S 4.5}*/

	PlanarSubgraph();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	virtual ~PlanarSubgraph() { }


	/*{\Moperations}*/

	leda::string name () const {
		return leda::string ("Planar Subgraph");
	}

	leda::string author () const {
		return "R. Jayakumar et al., S. Leipert";
	}

	leda::string impl_author () const {
		return leda::string ("Sebastian Leipert");
	}

	leda::string impl_date () const {
		return "November 1997";
	}


	AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|SubgraphModule|).
	}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	bool call (const graph& G, list<edge>& L);
	*/
	/*{\Mop}*/

	/* inherited
	bool call (graph& G);
	*/
	/*{\Mop}*/

	/* inherited
	bool call (GraphCopy& G, list<edge>& L);
	*/
	/*{\Mop}*/

	/* inherited
	int num_del_edges() const;
	*/
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	int runs() const {
		return _runs;
	}
	/*{\Mop}*/

	void runs(int n) {
		if (n >= 1) _runs = n;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

	
protected:
	void do_call (const leda::graph& G, leda::list<leda::edge>& L);

	int _runs;
};

/*{\Mimplementation
The algorithm has running time $\BigO(n+m)$, where $n$ is the
number of nodes and $m$ is the number of edges of the input graph $G$.
}*/


#endif
