/*******************************************************************************
+
+  AGD 1.2
+
+
+  PlanarizationLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                      PlanarizationLayout.h
//
//    author: Carsten Gutwenger
//    date:   august 1997
//-------------------------------------------------------------------


#ifndef _AGD_PLANARIZATION_LAYOUT_H
#define _AGD_PLANARIZATION_LAYOUT_H


#include <AGD/LayoutModuleImpl.h>
#include <AGD/ConstLedaGraphConcept.h>
#include <AGD/PlanarizerModule.h>
#include <AGD/AGDModuleOption.h>


/*{\Manpage {PlanarizationLayout} {} {The Planarization Method} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{8}
\hdclass{0}{7}{5}{AGDModule}
\arhook{5}{7.5}{6}
\hdclass{1}{5.5}{5.5}{LayoutModule}
\arhook{6.5}{6}{7.5}
\hdclass{1.5}{4}{6.5}{PlanarizationLayout}
\end{hierarchy}
\hfill
\agdeps{planarization.epsi}{5cm}
}*/

class __exportAGD PlanarizationLayout : public LayoutModuleImpl<ConstLedaGraphConcept> {

/*{\Mdefinition
The class |\Mname| represents a general framework for drawing graphs with
planarization. It applies a planarization module in order to transform
a general graph into a planar graph by introducing dummy vertices at edge
crossings, and draws the resulting planar graph using a layout algorithm
for planar graphs. The planarization and planar layout algorithms are
exchangeable modules allowing to use |\Mname| with different methods for
planarization and planar drawing.

\horizline
\commoninfo{Planarization}{Planarization}{}
{C. Gutwenger}{December 1997}{1.0}

\begin{preposttab}
precondition & $\{\, \emph{no self loops}\,\}\cup(\emph{Pre}_L \cap \{
\emph{simple, four-graph, connected}\}$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}\\[0.5ex]
where $\emph{Pre}_L$ is the precondition of the |planar_layouter| module.

\headerline{Exchangeable Modules}
Instances of type |\Mname| provide the following module options:
\begin{options}
\item {\bf planarizer}\\
a |PlanarizerModule| option used to compute the planarization of the input graph.

\begin{tabular}{ll}
guaranteed precondition: & $\{\, \emph{no self loops, directed} \,\}$\\
required postcondition: & $\{\, \emph{planar, no self loops} \,\}$\\
initial module: & |PlanarSubgraph|
\end{tabular}

\item {\bf planar\_layouter}\\
a |LayoutModule| option used to draw the planarized graph.

\begin{tabular}{ll}
guaranteed precondition: & $\{\, \emph{planar, simple, no self loops, connected,
four-graph} \,\}$\\
required postcondition: & $(\emph{PlaneGraphCopy-callable}, \{\, \emph{no crossings} \,\})$\\
initial module: & |QuasiOrthogonalLayout|
\end{tabular}
\end{options}
\horizline
}*/

AGD_DECLARE_MODULE(PlanarizerModule,planarizer)

protected:
	AGDModuleOption<LayoutModule> planar_layouter;
public:
	bool set_planar_layouter (const LayoutModule& M);

public:

	/*{\Mcreation L 4.5}*/

	PlanarizationLayout();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	~PlanarizationLayout() { }

	/*{\Moperations 1.5 4.5}*/

	leda::string name () const {
		return "Planarization";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "decembre 1997";
	}

	AGDModule *clone () const;

	int number_of_crossings() const {
		return _no_crossings;
	}
	/*{\Mop returns the number of crossings in the layout produced
	by a preceding call.}*/

	/*{\Moptions warnings=no}*/

	/*{\Mtext\medskip\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|LayoutModule|).
	}*/

	/* inherited
	virtual bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/
	
	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	/*
	bool set_planar_layouter (const LayoutModule& M);
	*/
	/*{\Mop}*/

	/*
	bool set_planarizer (const PlanarizerModule& M);
	*/
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

protected:
	int _no_crossings;

	void do_call (ConstLedaGraphConcept &C);
};


#endif
