/*******************************************************************************
+
+  AGD 1.2
+
+
+  SPQRTree.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                           SPQRTree.h
//
//
//    author: Carsten Gutwenger
//    date:   january 2000
//-------------------------------------------------------------------


#ifndef _AGD_SPQRTREE_H
#define _AGD_SPQRTREE_H

#include <AGD/Skeleton.h>
#include <AGD/PertinentGraph.h>
#include <LEDA/core/slist.h>


/*{\Manpage {SPQRTree} {} {SPQR-Trees} {T} }*/

class __exportAGD SPQRTree {

/*{\Mdefinition
The class |\Mname| maintains the arrangement of the triconnected
components of a biconnected multi-graph $G$ \cite{HT73} as a so-called
SPQR tree ${\calT}$ \cite{BT96b,BT96a}.
We call $G$ the original graph of ${\calT}$.
The class |\Mname| supports only the statical construction
of an SPQR-tree for a given graph $G$, dynamic updates are
not supported.

Each node of the tree has an associated type (represented by
|\Mname|::|NodeType|), which is either |SNode|, |PNode|, or
|RNode|, and a skeleton (represented by the class |Skeleton|).
The skeletons of the nodes of ${\calT}$ are in one-to-one
correspondence to the triconnected components of $G$, i.e.,
S-nodes correspond to polygons, P-nodes to bonds, and
R-nodes to triconnected graphs.

In our representation of SPQR-trees, Q-nodes are omitted. Instead,
the skeleton $S$ of a node $v\in{\calT}$ contains two types of edges:
real edges, which correspond to edges in $G$, and virtual edges, which
correspond to edges in $\calT$ having $v$ as an endpoint.
There is a special edge $e_r\in G$ at
which $\calT$ is rooted, i.e., the root node of $\calT$ is the node
whose skeleton contains the real edge corresponding to $e_r$.

The reference edge of the skeleton of the root node is $e_r$, the
reference edge of the skeleton $S$ of a non-root node $v$ is the virtual
edge in $S$ that corresponds to the tree edge $(|father|(v),v)$.
}*/

public:

	friend class Skeleton;

	/*{\Mtypes}*/ 

	enum NodeType { SNode, PNode, RNode };

	/*{\Menum the type of a tree node in $\calT$.}*/


	/*{\Mcreation}*/

	SPQRTree(const leda::graph &G, bool is_map = false) : cgp(&G),
		_tree(2,2), _is_map(is_map), sk_of(G), copy_of(G)
	{
		init(G,G.first_edge(),is_map);
	}
	/*{\Mcreate creates an SPQR tree ${\calT}$ for graph $G$ rooted at
	the first edge of $G$. If |is_map| is set to |true|, $G$ must be a
	map, i.e., for each edge $e$ the reversal information |rev(e)|
	is set. In this case, $e$ and |rev(e)| are considered as a single
	edge. \precond $G$ is biconnected and contains at least 3 nodes,
	or |G| has exactly 2 nodes and at least 3 (resp.\ 6 if |is_map|
	is true) edges.}*/

	SPQRTree(const leda::graph &G, leda::edge e, bool is_map = false) : cgp(&G),
		_tree(2,2), _is_map(is_map), sk_of(G), copy_of(G)
	{
		init(G,e,is_map);
	}
	/*{\Mcreate creates an SPQR tree ${\calT}$ for graph $G$ rooted at edge $e$.
	If |is_map| is set to |true|, $G$ must be a
	map, i.e., for each edge $e$ the reversal information |rev(e)|
	is set. In this case, $e$ and |rev(e)| are considered as a single
	edge. \precond $e\in G$ and $G$ is biconnected and contains at least 3 nodes,
	or |G| has exactly 2 nodes and at least 3 (resp.\ 6 if |is_map|
	is true) edges.}*/

	~SPQRTree();


	/*{\Moperations}*/


	/*{\Mtext\headerline{a) Access operations}}*/

	const leda::graph &tree() const {
		return _tree;
	}
	/*{\Mop returns a reference to the tree ${\calT}$.}*/

	const leda::graph &original_graph() const {
		return *cgp;
	}
	/*{\Mop returns a reference to the original graph $G$.}*/

	leda::edge root_edge() const {
		return _root_edge;
	}
	/*{\Mop returns the edge $e_r$ in $G$ at which ${\calT}$ is rooted.}*/

	leda::node root_node() const {
		return _root_node;
	}
	/*{\Mop returns the root node of ${\calT}$.}*/

	int number_of_S_nodes() const {
		return num_S;
	}
	/*{\Mop returns the number of S-nodes in ${\calT}$.}*/

	int number_of_P_nodes() const {
		return num_P;
	}
	/*{\Mop returns the number of P-nodes in ${\calT}$.}*/

	int number_of_R_nodes() const {
		return num_R;
	}
	/*{\Mop returns the number of R-nodes in ${\calT}$.}*/

	leda::list<leda::node> nodes_of_type(NodeType t) const;
	/*{\Mop returns the list of all nodes with type $t$.}*/

	NodeType type(leda::node v) const {
		return _type[v];
	}
	/*{\Mop returns the type of node $v$. \precond $v\in{\calT}$}*/

	const Skeleton &skeleton(leda::node v) const {
		return *sk[v];
	}
	/*{\Mop returns the skeleton of node $v$. \precond $v\in{\calT}$}*/

	leda::edge skeleton_edge_src(leda::edge e) const {
		return sk_edge_src[e];
	}
	/*{\Mop returns the edge in skeleton of $source(e)$ that corresponds
	to tree edge $e$. \precond $e\in{\calT}$}*/

	leda::edge skeleton_edge_tgt(leda::edge e) const {
		return sk_edge_tgt[e];
	}
	/*{\Mop returns the edge in skeleton of $target(e)$ that corresponds
	to tree edge $e$. \precond $e\in{\calT}$}*/

	const Skeleton &skeleton_of_real(leda::edge e) const {
		return *sk_of[e];
	}
	/*{\Mop returns the skeleton that contains the real edge $e$.
	\precond $e\in G$}*/

	leda::edge copy_of_real(leda::edge e) const {
		return copy_of[e];
	}
	/*{\Mop returns the skeleton edge that corresponds to the real
	edge $e$. \precond $e\in G$}*/

	void pertinent_graph(leda::node vT, PertinentGraph &Gp);
	/*{\Mop returns the pertinent graph of tree node |vT| in |Gp|.
	\precond $|vT|\in{\calT}$}*/


	/*{\Mtext\medskip\headerline{b) Update operations}}*/

	leda::node root_tree_at(leda::edge e);
	/*{\Mop roots ${\calT}$ at edge $e$ and returns the new root
	node of ${\calT}$. \precond $e\in G$}*/


protected:
	void init(const leda::graph &G, leda::edge e, bool is_map);
	void root_rec(leda::node v, leda::edge ef);

	void expand_virtual(leda::edge eT, leda::node src, PertinentGraph &Gp);
	void create_inner_vertices(leda::node vT, PertinentGraph &Gp);
	leda::node cpV_add_node(leda::node v_orig, PertinentGraph &Gp)
	{
		leda::node &vP = (*_cpV)[v_orig];
		if (vP == nil) {
			cpV_added.append(v_orig);
			Gp.origV[vP = Gp.P.new_node()] = v_orig;
		}
		return vP;
	}


	const leda::graph *cgp;
	leda::graph        _tree;
	bool              _is_map;

	int  num_S, num_P, num_R;
	leda::edge _root_edge;
	leda::node _root_node;

	leda::node_array<NodeType>  _type;
	leda::node_array<Skeleton *> sk;

	leda::edge_array<leda::edge>  sk_edge_src, sk_edge_tgt;

	leda::edge_array<Skeleton *> sk_of;
	leda::edge_array<leda::edge>  copy_of;

	leda::slist<leda::node> cpV_added;
	leda::node_array<leda::node> *_cpV;
	leda::edge_array<leda::edge> *_cpE;
};


/*{\Mimplementation
Let $n$ be the number of vertices and $m$ be the number of edges in
$G$. Let $n_{\calT}$ be the number of nodes in $\calT$.
The construction of an SPQR-tree $\calT$ for graph $G$
is implemented using the algorithm \cite{HT73} by Hopcroft and
Tarjan for computing the triconnected components of a biconnected
graph and takes time $\BigO(n+m)$.

The operations root\_tree\_at() and nodes\_of\_type() take
time $\BigO(n_{\calT})$.
The first call of pertinent\_graph() takes time $\BigO(n+m)$,
subsequent calls take time $\BigO(\Labs{P})$, where $P$ is the
computed pertinent graph.
All other operations take constant time. The data structure uses
$\BigO(n+m)$ space.
}*/


#endif
