/*******************************************************************************
+
+  AGD 1.2
+
+
+  Skeleton.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                           Skeleton.h
//
//
//    author: Carsten Gutwenger
//    date:   August 1999
//-------------------------------------------------------------------


#ifndef _AGD_SKELETON_H
#define _AGD_SKELETON_H

#include <LEDA/graph/graph.h>
#include <AGD/basic.h>

class SPQRTree;
class PlanarSPQRTree;


/*{\Manpage {Skeleton} {} {Skeletons} {S} }*/

class __exportAGD Skeleton {

/*{\Mdefinition
The class |\Mname| maintains the skeleton |\Mvar| of a node
$v_{\calT}$ of an SPQR-tree $\calT$. We call $\calT$ the
owner tree of |\Mvar| and $v_{\calT}$ the corresponding
tree node. Let $G$ be the original graph of $\calT$.

|\Mvar| consists of an undirected multi-graph $M$, which is represented
as a map, i.e., for each edge $e=(u,v)\in M$ exists the reversal
edge $|rev|(e)=(v,u)\in M$. The two edges $e$ and $|rev|(e)$
represent one undirected edge. If the owner tree of |\Mvar| is
a |PlanarSPQRTree|, then $M$ is a planar map.
The vertices in $M$ correspond to vertices
in $G$. The edges in $M$ are of two types: Real edges correspond
to edges in $G$ and virtual edges correspond to tree-edges in
$\calT$ having $v_{\calT}$ as an endpoint.

Let $e\in M$ be a virtual edge and $e_{\calT}$ be the corresponding
tree-edge. Then there exists exactly one edge $e'$ in another
skeleton $S'$ of $\calT$ that corresponds to $e_{\calT}$ as well.
We call $e'$ the twin edge of $e$.
}*/

	friend class SPQRTree;
	friend class PlanarSPQRTree;

public:

	/*{\Mcreation}*/
	/*{\Mtext Skeletons are created by the constructor of |SPQRTree|
	and can be accessed with the skeleton() function of |SPQRTree|.}*/

	Skeleton(const SPQRTree *T, leda::node vT);
	/*creates a skeleton |\Mvar| with owner tree
	$T$ and corresponding node |vT|. \precond $|vT|\in T$*/


	~Skeleton() { }


	/*{\Moperations}*/

	const SPQRTree &owner() const {
		return *_owner;
	}
	/*{\Mop returns the owner tree ${\calT}$.}*/

	leda::node tree_node() const {
		return _tree_node;
	}
	/*{\Mop returns the corresponding node in the owner tree ${\calT}$
	to which |\Mvar| belongs.}*/

	leda::edge reference_edge() const {
		return ref_edge;
	}
	/*{\Mop returns the reference edge of |\Mvar| in $M$. The reference
	edge is a real edge if |\Mvar| is the skeleton of the root node
	of $\calT$, and a virtual edge otherwise.}*/

	const leda::graph &get_graph() const {
		return M;
	}
	/*{\Mop returns a reference to the skeleton graph $M$.}*/

	leda::node original (leda::node v) const {
		return _orig[v];
	}
	/*{\Mop returns the vertex in the original graph $G$
	that corresponds to $v$. \precond $v\in M$.}*/

	bool is_virt (leda::edge e) const {
		return (_real[e] == nil);
	}
	/*{\Mop returns |true| iff $e$ is a virtual edge. \precond $e\in M$}*/

	leda::edge real_edge (leda::edge e) const {
		return _real[e];
	}
	/*{\Mop returns the real edge that corresponds to
	skeleton edge $e$ if $e$ is not virtual, |nil| otherwiese.
	\precond $e\in M$}*/

	leda::edge tree_edge (leda::edge e) const {
		return _tree_edge[e];
	}
	/*{\Mop returns the tree edge which is associated with
	skeleton edge $e$ if $e$ is virtual, |nil| otherwise.
	\precond $e\in M$}*/

	leda::edge twin_edge (leda::edge e) const;
	/*{\Mop returns the twin edge of skeleton edge $e$
	if $e$ is virtual, |nil| otherwise.
	\precond $e\in M$.}*/

	leda::node twin_tree_node (leda::edge e) const;
	/*{\Mop returns the tree node in ${\calT}$ containing
	the twin edge of skeleton edge $e$ if $e$ is virtual,
	|nil| otherwise. \precond $e\in M$.}*/

protected:
	const SPQRTree *_owner;
	leda::node  _tree_node;
	leda::edge  ref_edge;
	leda::graph M;

	leda::node_array<leda::node> _orig;
	leda::edge_array<leda::edge> _real;
	leda::edge_array<leda::edge> _tree_edge;
};


/*{\Mimplementation
All operations take constant time. The data structure uses
$\BigO(\Labs{M})$ space.
}*/


#endif
