/*******************************************************************************
+
+  AGD 1.2
+
+
+  SpringLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       SpringLayout.h
//
//    author: Stefan Naeher, David Alberts
//    date:   April 1997
//-------------------------------------------------------------------


#ifndef _AGD_SPRING_LAYOUT_H
#define _AGD_SPRING_LAYOUT_H


#include <AGD/LayoutModuleImpl.h>
#include <AGD/ConstLedaGraphConcept.h>
#include <LEDA/core/array2.h>


/*{\Manpage {SpringLayout} {} {Spring Embedder} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5}{AGDModule}
\arhook{5}{3.5}{6}
\hdclass{1.5}{1.5}{5}{LayoutModule}
\arhook{6.5}{2}{7.5}
\hdclass{3}{0}{5}{SpringLayout}
\end{hierarchy}
\hfill\agdeps{spring.epsi}{3cm}
}*/

class __exportAGD SpringLayout : public LayoutModuleImpl<ConstLedaGraphConcept>
{

/*{\Mdefinition
The class |\Mname| represents the spring embedder layout algorithm
by Fruchterman and Reingold~\cite{fr-gdfdp-91}.

This algorithm draws a general graph |G| straight-line. The drawing
of a planar graph must contain crossings.

The idea of the algorithm is the one of simulating a system of mass particles.
The vertices simulate mass points repelling each other and
the edges simulate springs with attracting forces. The algorithm tries
to minimize the energy of this physical system.

\horizline
\commoninfo{Spring Embedder}{Spring Embedder (Fruchterman/Reingold)}{T. Fruchterman, E. Reingold}
{S. N\"aher, D. Alberts}{April 1997}{}

\begin{preposttab}
precondition & $\emptyset$ \\
postcondition(|PRE|) & $\{\, \emph{straight-line} \,\}$
\end{preposttab}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |int| {\bf iterations} = 400\\
the number of main steps that the algorithm performs.
\item |bool| {\bf noise} = |true|\\
determines if the algorithm performs random perturbations.
\end{options}\horizline
}*/

  public:

	/*{\Mcreation}*/

    SpringLayout();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

    ~SpringLayout() { }

  /*{\Moperations}*/

    leda::string name () const { return "Spring Embedder"; }

    leda::string long_name () const
    { return "Spring Embedder (Fruchterman/Reingold)"; }

    leda::string author () const { return "T. Fruchterman, E. Reingold"; }

    leda::string impl_author () const
    { return "S. N\204her, D. Alberts"; }

    leda::string impl_date () const
    { return "April 1997"; }

    void do_spring(const leda::graph& G, LayoutInterface& a)
    { leda::list<leda::node> L; do_spring(G,a,L); }

    void do_spring(const leda::graph& G, LayoutInterface& a,
                   const leda::list<leda::node>& fixed_nodes);


	AGDModule *clone () const;


	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|LayoutModule|).
	}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

    int iterations() const { return the_iterations; }
	/*{\Mop}*/

    void iterations(int i)
    { if(i>0) the_iterations = i; }
	/*{\Mop}*/

    bool noise() const { return the_noise; }
	/*{\Mop}*/

    void noise(bool on)
    { the_noise = on; }
	/*{\Mop}*/
      
	/*{\Mtext \stdend}*/


  private:
    
    bool initialize(const leda::graph& G, LayoutInterface& a,
                    const leda::list<leda::node>& fixed_nodes);
    void main_step(const leda::graph& G, LayoutInterface& a);
    void cleanup() { if(A) delete A; }

    void do_call(ConstLedaGraphConcept &C);
    
    bool _start_animation(const leda::graph& G, LayoutInterface& a)
    { leda::list<leda::node> L; initialize(G,a,L); return true; }
    
    bool _next_step(const leda::graph& G, LayoutInterface& a)
    { main_step(G,a); return true; }
    
    void _stop_animation(const leda::graph& /*G*/, LayoutInterface& /*a*/)
    { cleanup(); }

    
    leda::node_array<leda::list_item>  lit; 
    leda::node_array<bool>       fixed;

    int c_f;
    
    double width;
    double height; 

    double tx_null;
    double ty_null;
    double tx;
    double ty;

    double k;
    double k2;
    double kk;
    int ki;
    
    int xA;
    int yA;

    leda::array2<leda::list<leda::node> > *A;


    float log_2(int x)
    { float l = 0.0f; while(x) { l++; x >>= 1; } return l/2; }

    int    the_iterations;     // # of iterations

    double xleft;              // coordinates of bounding box
    double xright;
    double ysmall;
    double ybig;

    bool the_noise;            // perform random perturbations

};


#endif
