/*******************************************************************************
+
+  AGD 1.2
+
+
+  TutteLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       TutteLayout.h
//
//    author: David Alberts
//    date:   April 1997
//-------------------------------------------------------------------


#ifndef _AGD_TUTTE_LAYOUT_H
#define _AGD_TUTTE_LAYOUT_H


#include <AGD/LayoutModuleImpl.h>
#include <AGD/ConstLedaGraphConcept.h>
#include <AGD/DRect.h>

/*{\Manpage {TutteLayout} {} {Tutte's Algorithm} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{6}
\hdclass{0}{5}{5}{AGDModule}
\arhook{5}{5.5}{6}
\hdclass{1.5}{3.5}{5}{LayoutModule}
\arhook{6.5}{4}{7.5}
\hdclass{3}{2}{5}{TutteLayout}
\end{hierarchy}
\hfill\agdeps{tutte.epsi}{3.8cm}
}*/

class __exportAGD TutteLayout : public LayoutModuleImpl<ConstLedaGraphConcept>
{

/*{\Mdefinition
The class |\Mname| represents the layout algorithm by
Tutte~\cite{t-hdg-63}.

This algorithm draws a planar graph |G| straight-line
without crossings. It can also draw non-planar graphs.

The idea of the algorithm is to place every vertex into the
center of gravity by its neighbours. 

\horizline
\commoninfo{Tutte}{Tutte}{W. T. Tutte}
{D. Alberts}{April 1997}{}

\begin{preposttab}
precondition & $\{\, \emph{connected, no self loops} \,\}$ \\
postcondition(|PRE|) & $\{\, \emph{straight-line} \,\}$
\end{preposttab}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |DRect| {\bf bbox} = (0.0, 250.0, 0.0, 250.0)\\
a double rectangle determining the bounding box of the layout.
\end{options}\horizline
}*/

  public:

	/*{\Mcreation}*/

    TutteLayout();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

    ~TutteLayout() { }

  /*{\Moperations 1.5 4}*/

    leda::string name () const { return "Tutte"; }

    leda::string long_name () const { return "Tutte"; }

    leda::string author () const { return "W. T. Tutte"; }

    leda::string impl_author () const
    { return "D. Alberts"; }

    leda::string impl_date () const
    { return "April 1997"; }


	AGDModule *clone () const;

	/*{\Mtext       
	\headerline{Special Functionality} }*/

    bool do_tutte(const leda::graph& G, LayoutInterface& a,
		leda::list<leda::node>& fixed_nodes,
        leda::list<DPoint>& fixed_positions);
      /*{\Mop does the actual computation. |fixed_nodes| and |fixed_positions|
              contain the nodes with fixed positions.

              \precond All nodes in |fixed_nodes| are distinct and belong to
              |G|. The sizes of |fixed_nodes| and |fixed_positions| match.
      }*/

	/*{\Moptions warnings=no}*/

	/*{\Mtext\medskip\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|LayoutModule|).
	}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/
	
	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	DRect bbox () const {
		return the_bbox;
	}
	/*{\Mop}*/

	void bbox (const DRect& bb) {
		the_bbox = bb;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/


  protected:

    void do_call(ConstLedaGraphConcept &C);
	/*void do_call (GraphCopy& GC, LayoutInterface& A)
	{ LayoutModule::do_call (GC,A); }*/

  private:

    void set_fixed_nodes(const leda::graph& G, leda::list<leda::node>& nodes,
                         leda::list<DPoint>& pos, double radius = 1.0);
      // sets the positions of the nodes in a largest face of G in the
      // form of a regular k-gon with the prescribed radius. The
      // corresponding nodes and their positions are stored in nodes
      // and pos, respectively. G does not have to be planar!

    DRect the_bbox;
};

/*{\Mimplementation
The algorithm solves a system of linear equations and takes
$\BigO(n^3)$ time.
}*/

#endif
