/*******************************************************************************
+
+  AGD 1.2
+
+
+  TwoLayerCrossMin.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        TwoLayerCrossMin.h
//
//
//    author: Carsten Gutwenger
//    date:   february 1998
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_TWO_LAYER_CROSS_MIN_H
#define _AGD_TWO_LAYER_CROSS_MIN_H

#include <AGD/AGDModule.h>


class Layer;
class Hierarchy;

/*{\Manpage {TwoLayerCrossMin} {} {Two Layer Crossing Minimization} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5.5}{AGDModule}
\arhook{5.5}{2}{6.5}
\hdclass{0.9}{0}{6.1}{TwoLayerCrossMin}
\end{hierarchy}
}*/

class __exportAGD TwoLayerCrossMin : public AGDModule {

/*{\Mdefinition
Modules of type |\Mname| perform a two layer crossing minimization
in a node hierarchy |H|, that is, the nodes of a graph |G|
on a layer |L|, called the changeable layer, are reordered according
to a neighbored layer of |L|, called the fixed layer, in order
to reduce the number of edge crossings between these two layers.

These modules can be used in a process, which
iteratively performs two layer crossing minimization
between two neighbored layers in order
to reduce the total number of edge crossings in the hierarchy.
For efficiency reasons, the class |\Mname| contains two
notification methods init() and cleanup().
A user must first call the init() method with the graph |G|
and the node hierarchy |H|. Then, the call() method can be called
several times with layers belonging to |H|. Finally, the
cleanup() method should be called to free resources.

\headerline{Input and Output Parameters}

\begin{tabular}{lcc}
input parameters: & hierarchy & |in_hierarchy=1| \\
 & layer & |in_layer=2| \\
output parameter: & layer & |out_layer=1|
\end{tabular}

The integer constants |in_hierarchy|, |in_layer| and |out_layer| denote the
positions in the input and output parameter lists. The values can be used to
reference the desired parameter when setting the pre- and postcondition.  

\headerline{Implementing Two Layer Crossing Minimization Modules}

An implementation of a two layer crossing minimization module
must override the method call($G$,$L$), which gets as
input a graph $G$ and a layer $L$. The class |Layer| provides
methods to access the fixed and changeable layer, and to
manipulate the order of the nodes on the changeable layer.
The method init($G$,$H$) can be overridden to allocate or initialize
internal data structures, and the method cleanup() can be overridden
to free resources allocated by init().
}*/

public:

	// the input and output parameters of a two layer crossing
	// minimization module
	static const int in_hierarchy;
	static const int in_layer;
	static const int out_layer;

	/*{\Minitialization M 5}*/

	TwoLayerCrossMin () : AGDModule(2,1) { }

	/*{\Mcreate initializes a two layer crossing minimization module.}*/

	virtual ~TwoLayerCrossMin () { }


	/*{\Moperations}*/

	/*{\Mtext
	\headerline{Standard Interface} }*/

	bool check (const leda::graph& G, const Layer& L, AgdKey &p) const;
	/*{\Mop returns |true| if $G$ satisfies the precondition of
	|\Mvar|. Otherwise, |false| is returned and |p| contains a
	property that is not satisfied.}*/

	virtual void init (const leda::graph& G, const Hierarchy& H);
	/*{\Mop must be called before the crossing minimization. Can be used by derived
	classes in order to allocate and initialize data structures.}*/

	virtual void call (const leda::graph& G, Layer& L) = 0;
	/*{\Moppure implements the module call for graph $G$. Reorders the nodes
	on layer $L$.}*/

	virtual void cleanup () { }
	/*{\Mop must be called after the crossing minimization. Can be used by
	derived classes in order to free resources allocated by init().}*/
};

inline void TwoLayerCrossMin::init (const leda::graph& /*G*/, const Hierarchy& /*H*/)
{ }

inline bool TwoLayerCrossMin::check (const leda::graph& G, const Layer& /*L*/,
	AgdKey &p) const

{
	return check_graph (in_hierarchy, G, p);
}

/*{\Mexample
We perform one top-down traversal of the crossing minimization method
used in the Sugiyama algorithm. First, we build a hierarchy of our
input graph $G$. Then, we initialize the two layer crossing minimization
module |minimizer|. We must tell the hierarchy the direction of the crossing
minimization (in this case |downward|), and the current changeable layer.
This determines also the fixed layer (in this case $i-1$). We call
the two layer crossing minimization for all but the first layer.
Finally, we must call cleanup() of |minimizer| in order to free resources.
\small\tt
\begin{verbatim}
#include <AGD/TwoLayerCrossMin.h>
#include <AGD/LongestPathRanking.h>
#include <AGD/Hierarchy.h>
#include <AGD/GraphCopy.h>

void minimize_top_down(graph &G, TwoLayerCrossMin &minimizer)
{
  node_array<int> rank(G);

  // assign a non-negative level to each node in G
  LongestPathRanking().call(G,rank);

  // build the hierarchy
  GraphCopy GC(G);
  Hierarchy H(GC,rank);

  // initialize crossing minimization
  minimizer.init(GC,H);

  // perform one top-down traversal
  H.direction(Hierarchy::downward);
  for (int i = 1; i < H.number_of_layers(); i++) {
    H.set_changeable_layer(i);
    minimizer.call(G,H[i]);
  }

  // free resources
  minimizer.cleanup();
}
\end{verbatim}\normalsize\rm
}*/


#endif
