/*******************************************************************************
+
+  AGD 1.2
+
+
+  isugiyama.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       isugiyama.h
//
//    author: Thomas Ziegler
//    date:   January 2001
//-------------------------------------------------------------------



#ifndef _AGD_ISUGIYAMA_H
#define _AGD_ISUGIYAMA_H

#include <LEDA/graph/graph.h>
#include <LEDA/core/tuple.h>
#include <LEDA/core/array.h>
#include <LEDA/core/set.h>

#include <AGD/AGDModule.h>
#include <AGD/GraphCopy.h>
#include <AGD/Hierarchy.h>

// using namespace leda;


//-----------------------------------------------------------------------------

/*{\Manpage {InteractiveSugiyama} {} {Interactive Sugiyama Layout} {IS} }*/

class InteractiveSugiyama  {
/*{\Mdefinition
	The class |\Mname| is a modified version of the Sugiyama algorithm.  
	It allows to add nodes and edges to a given Sugiyama layout such 
	that the changes between the given layout and the new Sugiyama layout
	are small.

	The idea is that we are given a directed graph and  
	draw it using the Sugiyama algorithm. Afterwards nodes and edges
	are added to the graph. These new nodes
	and edges are inserted into the given Sugiyama layout in such a way 
	that the ranking of the old nodes and the relative ordering of the old 
	nodes and edges in each layer is preserved in the new Sugiyama Layout.

\horizline
\commoninfo{Interactive Sugiyama}{Interactive Sugiyama}{T. Ziegler}
{T. Ziegler}{January 2001}{0.1}
\horizline
}*/
public:
	/*{\Mcreation}*/
	InteractiveSugiyama(); 
	/*{\Mcreate creates an instance of |\Mname| and initializes the 
	options of the underlying Sugiyama algorithm to the default values.}*/

	/*{\Moperations}*/

	leda::string name() const 
	{ return leda::string("Interactive Sugiyama Layout"); }
  
	leda::string long_name() const 
	{ return leda::string("Interactive Sugiyama Layout"); }
  
	leda::string author() const { return leda::string("Thomas Ziegler"); }
  
	leda::string impl_author() const 
	{ return leda::string("Thomas Ziegler"); }
  
        leda::string impl_version() const { return leda::string("0.1"); }
  
	leda::string impl_date() const { return leda::string("January 2001");}

	/*{\Mtext\headerline{a) Layout operations}}*/

int redraw(const leda::graph& G); 
/*{\Mop  applies the standard Sugiyama algorithm to |G| and computes
	the corresponding internal data structures for |\Mvar|. 
	|\Mvar.redraw()| should be called for |G| before calling 
	|\Mvar.update()| for |G|.}*/

	int redraw(const leda::graph& G, const leda::node_map<int>& node_rank); 
	/*{\Mop applies the standard Sugiyama algorithm to |G| with the 
	ranking information given in |node_rank| and computes
	the corresponding internal data structures for |\Mvar|. 
	|\Mvar.redraw()| should be called for |G| before calling 
	|\Mvar.update()| for |G|.}*/

	int update(const leda::graph& G);  
	/*{\Mop inserts the nodes and edges that were added to |G| since the 
	last application of |\Mvar.redraw()|, respectively |\Mvar.update()|.
	If |\Mvar.redraw()| was not called for |G| before or edges or nodes 
	were deleted |\Mvar.redraw()| is called.}*/

	/*{\Mtext\headerline{b) Access operations}}*/
	int number_of_crossings() const {return _num_cross;}
	/*{\Mop returns the number of crossings in the layout.}*/

	int node_rank(leda::node v) const {return _node_rank[v];}
	/*{\Mop returns the rank of |node v|.}*/

	int node_pos(leda::node v) const {return _node_pos[v];}
	/*{\Mop returns the position of |node v| in its layer.}*/

	const leda::node_map<int>& node_rank() const {return _node_rank;}
	/*{\Mop returns a reference to the ranking information of the nodes
	of |G|.}*/

	const leda::node_map<int>& node_pos() const {return _node_pos;}
	/*{\Mop returns a reference to the position information of the nodes
	of |G|.}*/

	leda::list<leda::two_tuple<int,int> > edge_positions(leda::edge e) const 
	{return _edge_positions[e];}
	/*{\Mop returns the positions of the bends of |edge e| in the layers 
	between the layer of the source node and the target node of |e|. 
	The first elements of the |two_tuple|s are the layers and the second 
	elements the position in the layer.}*/
	
	/*{\Mtext
	\smallskip\headerline{c) Access operations to Sugiyama options}}*/

	//access to sugiyama options
	int ranking() const {return _ranking;}
	/*{\Mop}*/

	int cross_min() const {return _cross_min;}
	/*{\Mop}*/

	int cycle_removal() const {return _cycle_removal;}
	/*{\Mop}*/

	int width() const {return _width;}
	/*{\Mop}*/

	int runs() const {return _runs;}
	/*{\Mop}*/

	int  fails() const {return _fails;}
	/*{\Mop}*/

	bool init_order() const {return _init_order;}
	/*{\Mop}*/

	bool transpose() const {return _transpose;}
	/*{\Mop}*/

	double node_dist() const {return _node_dist;}
	/*{\Mop}*/

	double level_dist() const {return _level_dist;}
	/*{\Mop}*/

	bool fixed_level_dist() const {return _fixed_level_dist;}	
	/*{\Mop}*/

	/*{\Mtext
	\smallskip\headerline{d) Update operations for Sugiyama options}}*/

	void ranking(int ra) {_ranking=ra;}
	/*{\Mop}*/

	void cross_min(int cm) {_cross_min=cm;}
	/*{\Mop}*/

	void cycle_removal(int cr) {_cycle_removal=cr;}
	/*{\Mop}*/

	void width(int w) {_width=w;}
	/*{\Mop}*/

	void runs(int ru) {_runs=ru;}
	/*{\Mop}*/

	void  fails(int fa) {_fails=fa;}
	/*{\Mop}*/

	void init_order(bool i) {_init_order=i;}
	/*{\Mop}*/

	void transpose(bool t) {_transpose=t;}
	/*{\Mop}*/

	void node_dist(double n) {_node_dist=n;}
	/*{\Mop}*/

	void level_dist(double l) {_level_dist=l;}
	/*{\Mop}*/

	void fixed_level_dist(bool fi) {_fixed_level_dist=fi;}
	/*{\Mop}*/

	void set_options(int ranking,int cross_min,int cycle_removal,int width,
	int runs,int fails,bool init_order,bool transpose,double node_dist,
	double level_dist,bool fixed_level_dist);	
	/*{\Mop}*/


private:
	leda::node_map<int> _node_rank;  //layer of node
	leda::node_map<int> _node_pos;   //position of node in layer
	leda::edge_map<leda::list<leda::two_tuple<int,int> > > _edge_positions;
	                           //position of edge in all intermediate layers
	leda::array<int> _layersize;     
	leda::set<leda::node> _old_nodes;
	leda::set<leda::edge> _old_edges;
	bool _valid_hierarchy;  //complete redraw for first application
	int _num_cross;

	//sugiyama options
	int _ranking;
	int _cross_min;
	int _cycle_removal;
	int _width;
	int  _runs;
	int  _fails;
	bool _init_order;
	bool _transpose;
	double _node_dist;
	double _level_dist;
	bool _fixed_level_dist;

	//auxiliary functions used by update()
	void insert_single_node(const leda::graph& G, leda::node the_new_node);

	void insert_single_edge (const leda::graph& G, leda::edge the_new_edge,
                 leda::array<leda::list<leda::two_tuple<int,int> > >& sles);

	void set_positions_in_hierarchy(const leda::graph& G, const GraphCopy& GC, Hierarchy& H);

	double compute_layer(const leda::graph& G, int maxlayer, leda::node the_new_node);

	void insert_new_layer(const leda::graph& G,double nl);

	double compute_pos(const leda::graph& G, leda::node the_new_node);

	void insert_new_node_at_pos(const leda::graph& G, leda::node the_new_node,double p);

	void generate_sles
		(const leda::graph& G,const leda::list<leda::edge>& new_edges,
                 leda::array<leda::list<leda::two_tuple<int,int> > >& sles);

	void make_room_in_layer(const leda::graph& G,double l,int newpos);

	bool check_hierarchy(const leda::graph& G);

	void compute_auxiliary_graph
	(leda::graph& G_aux, leda::node_map<int>& aux_rank, 
                             leda::node_map<double>& aux_pos, 
	leda::node& s_aux, leda::node& t_aux);

	void compute_edge_weights (const leda::graph& G_aux, 
         const leda::node_map<int>& aux_rank,
         const leda::node_map<double>& aux_pos, 
	 leda::node s_aux, leda::node t_aux,
         leda::array<leda::list<leda::two_tuple<int,int> > >& sles,
         leda::edge_array<int>& edge_weight);	
};

/*{\Mimplementation
The update operation first inserts the new nodes and afterwards the new
edges. The layer of a new node depends on the layers of the source nodes of the
incoming edges and the layers of the target nodes of the outgoing edges. The
position of the new node in its layer depends on the average position of the 
neighbors. The new edges are inserted in such a way that the number of 
crossings is minimized. This is done by a shortest path computation in an
auxiliary graph.
}*/

#endif
