/*******************************************************************************
+
+  AGD 1.2
+
+
+  GridCoordinateMapper.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                  GridCoordinateMapper.cpp
//
//    author: Carsten Gutwenger
//    date:   september 1998
//-------------------------------------------------------------------


#include <AGD/GridCoordinateMapper.h>
#include <AGD/CGLayoutInterface.h>

using namespace leda;


void GridCoordinateMapper::call (GridLayoutInfo &gli, LayoutInterface& A)
{
	_cgp   = &A.get_graph();
	_ccgp  = nil;
	_glip  = &gli;
	_cgglip = nil;

	init(A);
	map_nodes(A);
	map_edges(A);

	A.update();
	cleanup(A);
}


void GridCoordinateMapper::call (CGGridLayoutInfo &gli, CGLayoutInterface& A)
{
	_cgp  = _ccgp = &A.get_graph();
	_glip = _cgglip = &gli;

	init(A);
	map_nodes(A);
	map_edges(A);
	map_clusters(A);

	A.update();
	cleanup(A);
}


void GridCoordinateMapper::get_node_grid_size(GridLayoutInfo &gli,
	const LayoutInterface& A)
{
	_cgp  = &A.get_graph();
	_glip = &gli;

	do_get_node_grid_size(A);
}

void GridCoordinateMapper::map_nodes(LayoutInterface &A)
{
	const graph& G = get_graph();
	node v;

	forall_nodes (v, G)
		A.set_position (v, map_node(v,A));
}

DPoint GridCoordinateMapper::map_node (node v, LayoutInterface &/*A*/)
{
	DPoint p1 (map_x(column(v)), map_y(row(v)));
	DPoint p2 (map_x(column(v)+grid_width(v)), map_y(row(v)+grid_height(v)));

	return 0.5 * (p1 + p2);
}

void GridCoordinateMapper::map_edges(LayoutInterface &A)
{
	const graph& G = get_graph();
	edge e;

	forall_edges (e, G) {
		IPolyline ipl = bends(e);
		node v = source(e), w = target(e);

		IPoint gsa(column(v),row(v));
		IPoint gta(column(w),row(w));

		if (!ipl.empty() && in_rect(ipl.head(),v))
			gsa = ipl.pop_front();
		if (!ipl.empty() && in_rect(ipl.tail(),w))
			gta = ipl.pop_back();

		DPolyline dpl;
		DPoint s_a, t_a;
		map_edge (e, ipl, gsa, gta, dpl, s_a, t_a, A);
		A.set_bend_points (e, dpl);
		A.set_source_anchor (e, s_a);
		A.set_target_anchor (e, t_a);
	}
}

void GridCoordinateMapper::get_anchor(DPoint &anchor, node v,
	const DPoint &abs_anchor, const LayoutInterface &A)
{
	double w = A.get_width(v), h = A.get_height(v);

	anchor.xcoord(w > 0 ? 2 * (abs_anchor.xcoord()-A.get_x(v)) / w : 0);
	anchor.ycoord(h > 0 ? 2 * (abs_anchor.ycoord()-A.get_y(v)) / h : 0);
}

void GridCoordinateMapper::map_edge (edge e, IPolyline &ipl, const IPoint &gsa,
	const IPoint &gta, DPolyline& dpl, DPoint& s_a, DPoint& t_a, LayoutInterface &A)
{
	get_anchor(s_a,source(e),map(gsa),A);
	get_anchor(t_a,target(e),map(gta),A);

	IPoint ip;
	forall (ip, ipl)
		dpl.append(map(ip));
}

bool GridCoordinateMapper::in_rect(const IPoint &ip, leda::node v) const
{
	return (column(v) <= ip.xcoord() && ip.xcoord() <= column(v)+grid_width(v) &&
		row(v) <= ip.ycoord() && ip.ycoord() <= row(v)+grid_height(v));
}

void GridCoordinateMapper::map_clusters(CGLayoutInterface &A)
{
	const list<Cluster*> cl_list = _ccgp->all_clusters();
	Cluster* cl;

	int x, y, width, height;

	forall(cl, cl_list) {  
		x = _cgglip->x(cl);
		y = _cgglip->y(cl);
		width  = _cgglip->width (cl);
		height = _cgglip->height(cl);

		A.set_cluster_region(cl, 
			map_x(x),
			map_y(y), 
			map_x(width), 
			map_y(height));
	}
}
