/*******************************************************************************
+
+  AGD 1.2
+
+
+  isugiyama.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


// Sugiyama version that allows to add and delete nodes and edges
// to a given sugiyama drawing such that the ranking and the 
// relative positions of old nodes remains unchanged

#include <AGD/GraphCopy.h>
#include <AGD/Hierarchy.h>
#include <AGD/SugiyamaLayout.h>

#include <AGD/DfsRanking.h>
#include <AGD/CoffmanGrahamRanking.h>
#include <AGD/LongestPathRanking.h>
#include <AGD/BarycenterHeuristic.h>
#include <AGD/MedianHeuristic.h>
#include <AGD/WeightedMedianHeuristic.h>
#include <AGD/GreedyInsertHeuristic.h>
#include <AGD/GreedySwitchHeuristic.h>
#include <AGD/SplitHeuristic.h>
#include <AGD/SiftingHeuristic.h>
#include <AGD/LEDAMakeAcyclic.h>
#include <AGD/GreedyCycleRemoval.h>

#include <LEDA/core/p_queue.h>
#include <LEDA/graph/graph_alg.h>

#include <AGD/isugiyama.h>

using namespace leda;

/*-----------------------------------------------------------------------------*/

InteractiveSugiyama::InteractiveSugiyama()
{
	_valid_hierarchy=false;  //complete redraw for first application

	//sugiyama options
	set_options(0,0,0,3,3,4,true,true,20.0,20.0,false);
}

/*-----------------------------------------------------------------------------*/

//redraw given graph completely with standard sugiyama algorithm 

int InteractiveSugiyama::redraw(const graph& G)
{	 
	if (G.empty()) {return 0;}

	_node_pos.init(G);
	_node_rank.init(G);
	_edge_positions.init(G);

	_valid_hierarchy=true;
	_old_nodes.clear();
	node v;forall_nodes(v,G) {_old_nodes.insert(v);}
	
	_old_edges.clear();
	edge e;forall_edges(e,G) {_edge_positions[e].clear();_old_edges.insert(e);}

    //compute rank of nodes
	switch (_ranking) {
		case 0:
			{LongestPathRanking LPR;

			if (_cycle_removal==0) {
				LEDAMakeAcyclic sub;
				LPR.set_subgraph(sub);
			}
			else {
				GreedyCycleRemoval sub;
				LPR.set_subgraph(sub);
			}

			LPR.call(G,_node_rank);}
			break;
		case 1:
			{CoffmanGrahamRanking CGR;	

			if (_cycle_removal==0) {
				LEDAMakeAcyclic sub;
				CGR.set_subgraph(sub);
			}
			else {
				GreedyCycleRemoval sub;
				CGR.set_subgraph(sub);
			}

			CGR.width(_width);
			CGR.call(G,_node_rank);}
			break;
		case 2:
			DfsRanking().call(G,_node_rank);
			break;
	}

	GraphCopy GC(G);Hierarchy H(GC,_node_rank);  //compute pos of nodes and bends

    SugiyamaLayout sugiyama;
	sugiyama.runs(_runs);
	sugiyama.fails(_fails);
	sugiyama.init_order(_init_order);
	sugiyama.transpose(_transpose);

	switch (_cross_min) {
		case 0:
			sugiyama.set_cross_min(BarycenterHeuristic());
			break;
		case 1:
			sugiyama.set_cross_min(MedianHeuristic());
			break;
		case 2:
			sugiyama.set_cross_min(WeightedMedianHeuristic());
			break;
		case 3:
			sugiyama.set_cross_min(SplitHeuristic());
			break;
		case 4:
			sugiyama.set_cross_min(SiftingHeuristic());
			break;
	}
	sugiyama.reduce_crossings(GC,H);
	_num_cross=sugiyama.number_of_crossings();

	array<int> locallayersize(H.number_of_layers());
	for (int i=0;i<H.number_of_layers();i++) locallayersize[i]=0;
	forall_nodes(v,GC) {
		if (GC.is_dummy(v)) {  //v correponds to bend
			two_tuple<int,int> p(H.rank(v),H.pos(v));
			edge elocal=GC.original(GC.first_adj_edge(v));
			_edge_positions[elocal].append(p);
			locallayersize[H.rank(v)]++;
		}

		else {
			_node_pos[GC.original(v)]=H.pos(v);
			locallayersize[_node_rank[GC.original(v)]]++;
		} //v original node
	}
	
	_layersize=locallayersize;  //store size of layer

	return H.number_of_layers()-1;
}

int InteractiveSugiyama::redraw(const graph& G, const node_map<int>& node_rank)
{	 
	if (G.empty()) {return 0;}

	_node_pos.init(G);
	_node_rank.init(G);
	_edge_positions.init(G);

	_valid_hierarchy=true;
	_old_nodes.clear();
	node v;forall_nodes(v,G) {_old_nodes.insert(v);}
	
	_old_edges.clear();
	edge e;forall_edges(e,G) {_edge_positions[e].clear();_old_edges.insert(e);}

	_node_rank=node_rank;

	GraphCopy GC(G);Hierarchy H(GC,_node_rank);  //compute pos of nodes and bends

    SugiyamaLayout sugiyama;
	sugiyama.runs(_runs);
	sugiyama.fails(_fails);
	sugiyama.init_order(_init_order);
	sugiyama.transpose(_transpose);

	switch (_cross_min) {
		case 0:
			sugiyama.set_cross_min(BarycenterHeuristic());
			break;
		case 1:
			sugiyama.set_cross_min(MedianHeuristic());
			break;
		case 2:
			sugiyama.set_cross_min(WeightedMedianHeuristic());
			break;
		case 3:
			sugiyama.set_cross_min(SplitHeuristic());
			break;
		case 4:
			sugiyama.set_cross_min(SiftingHeuristic());
			break;
	}
	sugiyama.reduce_crossings(GC,H);
	_num_cross=sugiyama.number_of_crossings();

	array<int> locallayersize(H.number_of_layers());
	for (int i=0;i<H.number_of_layers();i++) locallayersize[i]=0;
	forall_nodes(v,GC) {
		if (GC.is_dummy(v)) {  //v correponds to bend
			two_tuple<int,int> p(H.rank(v),H.pos(v));
			edge elocal=GC.original(GC.first_adj_edge(v));
			_edge_positions[elocal].append(p);
			locallayersize[H.rank(v)]++;
		}

		else {
			_node_pos[GC.original(v)]=H.pos(v);
			locallayersize[_node_rank[GC.original(v)]]++;
		} //v original node
	}
	
	_layersize=locallayersize;  //store size of layer

	return H.number_of_layers()-1;
}

/*-----------------------------------------------------------------------------*/

void InteractiveSugiyama::set_options(int ranking,int cross_min,int cycle_removal,int width,int runs,
		int  fails,bool init_order,bool transpose,double node_dist,double level_dist,
		bool fixed_level_dist) 
{
	_ranking=ranking;
	_cross_min=cross_min;
	_cycle_removal=cycle_removal;
	_width=width;
	_runs=runs;
	_fails=fails;
	_init_order=init_order;
	_transpose=transpose;
	_node_dist=node_dist;
	_level_dist=level_dist;
	_fixed_level_dist=fixed_level_dist;
}

/*-----------------------------------------------------------------------------*/

bool InteractiveSugiyama::check_hierarchy(const graph& G)
{	
	//bool wasvalid=true;

	array<int> locallayersize(_layersize.size());
	
	int i;for (i=0;i<_layersize.size();i++) locallayersize[i]=0;

	node v;
	forall_nodes(v,G) {if (_node_rank[v]!=-1) {locallayersize[_node_rank[v]]++;}}

	edge e;
	forall_edges(e,G) {
		two_tuple<int,int> p(0,0);
		forall(p,_edge_positions[e]) {locallayersize[p.first()]++;}
	}

	for (i=0;i<_layersize.size();i++) {
		if (locallayersize[i]!=_layersize[i]) {return false;}
	}

	return true;
}

/*-----------------------------------------------------------------------------*/

// inserting new nodes and edges

int InteractiveSugiyama::update(const graph& G)
{	
	if (!_valid_hierarchy) {return redraw(G);}

	list<node> new_nodes;
	node v;forall_nodes(v,G) {
		if (!_old_nodes.member(v)) {
			_node_pos[v]=-1;_node_rank[v]=-1;new_nodes.append(v);_old_nodes.insert(v);
		}
	}

	list<edge> new_edges;
	edge e;forall_edges(e,G) {
		if (!_old_edges.member(e)) {
			new_edges.append(e);_edge_positions[e].clear();_old_edges.insert(e);
		}
	}

	bool wasvalid=check_hierarchy(G);

	if (!wasvalid) {return redraw(G);}

	while (!new_nodes.empty()) {insert_single_node(G,new_nodes.pop());}	

	forall(e,new_edges) {
		if (_node_rank[G.source(e)]==_node_rank[G.target(e)]) {return redraw(G);}
	}

	while (!new_edges.empty()) 	{
		array<list<two_tuple<int,int> > > sles(_layersize.size()-1);
		generate_sles(G,new_edges,sles);	
		insert_single_edge(G,new_edges.pop(),sles);
	}

	return _layersize.size()-1;
}

/*-----------------------------------------------------------------------------*/

// generate sorted lists of edges between adjacent layers 

void InteractiveSugiyama::generate_sles
(const graph& G,const list<edge>& new_edges,array<list<two_tuple<int,int> > >& sles)
{
	int i;
	for (i=0;i<_layersize.size()-1;i++) {
		list<two_tuple<int,int> > dummylist;dummylist.clear();
		sles[i]=dummylist;
	}

	edge e;
	forall_edges(e,G) {
		if (e!=new_edges.head()) {
		    node s=G.source(e),t=G.target(e);
			if (_edge_positions[e].empty()&&
				_node_rank[s]-_node_rank[t]!=1&&_node_rank[s]-_node_rank[t]!=-1) {
				break;
			}
			
			list<two_tuple<int,int> > bends=_edge_positions[e];
			two_tuple<int,int> p1(_node_rank[s],_node_pos[s]);bends.append(p1);
			two_tuple<int,int> p2(_node_rank[t],_node_pos[t]);bends.append(p2);
			bends.sort();

			while (!bends.empty()) {			
				two_tuple<int,int> bendposition=bends.pop();
				if (bends.size()>0) {
					int i=bendposition.second();
					int j=bends.head().second();
					two_tuple<int,int> p(i,j);
					sles[bendposition.first()].append(p);
				}
			}
		}
	}

	for (i=0;i<_layersize.size()-1;i++) {sles[i].sort();}
}

/*-----------------------------------------------------------------------------*/

void InteractiveSugiyama::insert_single_node(const graph& G,node the_new_node)
{
	//compute layer of new node

	double layer_of_new_node=compute_layer(G,_layersize.size()-1,the_new_node);

	//insert new layer if necessary
	int lfloor=static_cast<int>(layer_of_new_node);
	if (layer_of_new_node!=lfloor) {  //new intermediate layer
		insert_new_layer(G,layer_of_new_node);
		_node_rank[the_new_node]=lfloor+1;	
	}
	else {_node_rank[the_new_node]=static_cast<int>(layer_of_new_node);}
	
	if (layer_of_new_node<0) {
		_node_rank[the_new_node]=0;_node_pos[the_new_node]=0;
		_layersize[0]=1;
		return;
	}
	
	if (layer_of_new_node==_layersize.size()) {
		int ls=_layersize.size();_layersize.resize(ls+1);
		_layersize[ls]=1;_node_rank[the_new_node]=ls;_node_pos[the_new_node]=0;return;
	}	

	//compute position of new node
	double pos_of_new_node=compute_pos(G,the_new_node);

	//insert new node at computed position in layering
	insert_new_node_at_pos(G,the_new_node,pos_of_new_node);
}

/*-----------------------------------------------------------------------------*/

double InteractiveSugiyama::compute_layer(const graph& G,int maxlayer, node the_new_node)
{
	//we use the following simple strategy: compute the maximum layer
	//of the source node of an in-edge and the minimum layer of the
	//target node of an out-edge. if maxin<minout

	double bin=0,bout=0;
	int indeg=0,outdeg=0;
	int maxin=-1,minout=-1;
	edge e;forall_in_edges(e,the_new_node) {
		int rd=_node_rank[G.source(e)];if (rd>=0) {
			bin+=rd;indeg++;
			if (maxin==-1||maxin<rd) maxin=rd;
		}
	}
	forall_out_edges(e,the_new_node) {
		int rd=_node_rank[G.target(e)];if (rd>=0) {
			bout+=rd;outdeg++;
			if (minout==-1||minout>rd) minout=rd;
		}
	}

	if (indeg==0&&outdeg==0) return maxlayer;   //no neighbors
    
	double b;
	if (indeg==0) {b=minout-1;}
	else if (outdeg==0) {b=maxin+1;}
	else if (maxin<minout) {b=maxin+(minout-maxin)/2.0;}
	else {
	  b=bin+bout;
	  b/=(indeg+outdeg);  //average layer of neigbors
	}

	if (b<0) return -0.5;

	//check if new node has neighbor in bfloor resp. bceil
	int bfloor=static_cast<int>(b); bool found=false;
	forall_in_edges(e,the_new_node) { if (_node_rank[G.source(e)]==bfloor) found=true;}
	forall_out_edges(e,the_new_node) { if (_node_rank[G.target(e)]==bfloor) found=true;}

	if (!found) {b=bfloor;}
	else {
		found=false;
		int bceil=bfloor+1;
		forall_in_edges(e,the_new_node) { if (_node_rank[G.source(e)]==bceil) found=true;}
		forall_out_edges(e,the_new_node) { if (_node_rank[G.target(e)]==bceil) found=true;}

		if (!found) {b=bceil;}
		else {b=bfloor+0.5;}
	}

	return b;
}

/*-----------------------------------------------------------------------------*/

void InteractiveSugiyama::insert_new_layer(const graph& G,double nl)
{
	//update ranks
	_layersize.resize(_layersize.size()+1);
	node v;forall_nodes(v,G) {if (_node_rank[v]>nl) _node_rank[v]++;}
	edge e;forall_edges(e,G) {
		if ((_node_rank[G.source(e)]>nl||_node_rank[G.target(e)]>nl)&&!_edge_positions[e].empty()) {
			list<two_tuple<int,int> > newlist;two_tuple<int,int> p;
			forall(p,_edge_positions[e]) {
				if (p.first()>nl) {
					two_tuple<int,int> pnew(p.first()+1,p.second());
					newlist.append(pnew);
				}
				else {newlist.append(p);}
			}
			_edge_positions[e]=newlist;
		}
	}

	if (nl<0) {	//update von layersize	and return
		int current_layer=_layersize.size()-1;
		while (current_layer>0) {
			_layersize[current_layer]=_layersize[current_layer-1];current_layer--;
		}

		return;
	}

	//sort edges crossing new layer
	int layer_i=static_cast<int>(nl);

	p_queue<two_tuple<int,int> ,edge> sorted_edges;
	forall_edges(e,G) {
		node s=G.source(e),t=G.target(e);
		if ((_node_rank[s]>=0&&_node_rank[t]>=0&&_node_rank[s]<nl&&_node_rank[t]>nl)||(_node_rank[s]>nl&&_node_rank[t]<nl)) {
			//if edge crosses new layer
			if (_edge_positions[e].empty()) {
				if (_node_rank[s]<nl&&_node_rank[t]-_node_rank[s]==2) {
					two_tuple<int,int> tempp(_node_pos[s],_node_pos[t]);
					sorted_edges.insert(tempp,e);
				}
				else if (_node_rank[t]-_node_rank[s]==-2) {
					two_tuple<int,int> tempp(_node_pos[t],_node_pos[s]);
					sorted_edges.insert(tempp,e);
				}
			}

			else {		
				int i=-1,j=-1;
				if (_node_rank[s]==layer_i) {i=_node_pos[s];}
				if (_node_rank[t]==layer_i) {i=_node_pos[t];}
				if (_node_rank[s]==layer_i+2) {j=_node_pos[s];}
				if (_node_rank[t]==layer_i+2) {j=_node_pos[t];}

				two_tuple<int,int> p(0,0);
				forall(p,_edge_positions[e]) {	
					if (p.first()==layer_i) {i=p.second();}
					else if (p.first()==layer_i+2) {j=p.second();}
				}

				two_tuple<int,int> tempp(i,j);
				sorted_edges.insert(tempp,e);
			}
		}
	}

	int i=0;
	while (!sorted_edges.empty()) {
		edge e=sorted_edges.inf(sorted_edges.find_min());
		sorted_edges.del_min();
		two_tuple<int,int> p(layer_i+1,i++);
		_edge_positions[e].append(p);
	}

	//update of layersize
	int current_layer=_layersize.size()-1;
	while (current_layer>layer_i+1) {
		_layersize[current_layer]=_layersize[current_layer-1];
		current_layer--;
	}
	_layersize[current_layer]=i;
}

/*-----------------------------------------------------------------------------*/

double InteractiveSugiyama::compute_pos(const graph& G,node the_new_node)
{	
	//we use the following simple strategy: compute the average pos of 
	//all neighbors of the new node
	//a more elaborate strategy would also take crossings into account 
	edge e;double b=0;int indeg=0,outdeg=0;
	forall_in_edges(e,the_new_node) {
		node d=G.source(e);
		if (_node_pos[d]>=0){
			b+=(1.0*_node_pos[d]/_layersize[_node_rank[d]]);indeg++;
		}
	}
	forall_out_edges(e,the_new_node) {
		node d=G.target(e);
		if (_node_pos[d]>=0) {	
			b+=(1.0*_node_pos[d]/_layersize[_node_rank[d]]);outdeg++;
		}
	}

	int lsrn=_layersize[_node_rank[the_new_node]];

	if (indeg==0&&outdeg==0) return lsrn; //no neighbors

	b/=(indeg+outdeg);  //average pos of neighbors
	b*=lsrn;

	int bfloor=static_cast<int>(b);
	if (b==0) b=-0.5;else if (b==bfloor&&b<lsrn) {b=b+0.5;}
	if (b>=lsrn) b=lsrn;

	return b;
}

/*-----------------------------------------------------------------------------*/

void InteractiveSugiyama::insert_new_node_at_pos
(const graph& G, node the_new_node,double p)
{
	int l=_node_rank[the_new_node];
	int newpos, pfloor=static_cast<int>(p);
	if (p==pfloor) newpos=pfloor; else newpos=pfloor+1;
	if (p<0) newpos=0;

	if (_layersize[l]!=0&&newpos<_layersize[l]) {
		make_room_in_layer(G,l,newpos);
	}

	_node_pos[the_new_node]=newpos;
	_layersize[l]++;	
}

void InteractiveSugiyama::make_room_in_layer(const graph& G, double l,int newpos)
{	
	node v;forall_nodes(v,G) {if (_node_rank[v]==l&&_node_pos[v]>=newpos) _node_pos[v]++;}
	edge e;forall_edges(e,G) {
		if ((_node_rank[G.source(e)]<l&&_node_rank[G.target(e)]>l)||
		    (_node_rank[G.source(e)]>l&&_node_rank[G.target(e)]<l)) {
			list<two_tuple<int,int> > newlist;
			two_tuple<int,int> p;
			forall(p,_edge_positions[e]) {
				if (p.first()==l&&p.second()>=newpos) {
					two_tuple<int,int> newp(p.first(),p.second()+1);
					newlist.append(newp);
				}
				else {newlist.append(p);}
			}

			_edge_positions[e]=newlist;
		}
	}
}

/*-----------------------------------------------------------------------------*/

void InteractiveSugiyama::compute_auxiliary_graph
(graph& G_aux, node_map<int>& aux_rank, node_map<double>& aux_pos, 
 node& s_aux, node& t_aux)
{
	node s_orig=s_aux,t_orig=t_aux;

    s_aux=G_aux.new_node();
	aux_rank[s_aux]=_node_rank[s_orig];
	aux_pos[s_aux]=_node_pos[s_orig];
	t_aux=G_aux.new_node();
	aux_rank[t_aux]=_node_rank[t_orig];
	aux_pos[t_aux]=_node_pos[t_orig];

    list<node> prev_layer;
	prev_layer.append(s_aux);

	int i;
	for (i=_node_rank[s_orig]+1;i<_node_rank[t_orig];i++) {
		list<node> current_layer;	
		int j;
		for (j=0;j<=_layersize[i];j++) {
			node n=G_aux.new_node();
			current_layer.append(n);
			node m;forall(m,prev_layer) {G_aux.new_edge(m,n);}
			aux_rank[n]=i;
			aux_pos[n]=j-0.5;
		}

		prev_layer=current_layer;
	}

	node n;forall(n,prev_layer) {G_aux.new_edge(n,t_aux);}
}

/*-----------------------------------------------------------------------------*/

void InteractiveSugiyama::compute_edge_weights
(const graph& G_aux, const node_map<int>& aux_rank, 
 const node_map<double>& aux_pos, node s_aux, node t_aux,
 array<list<two_tuple<int,int> > >& sles,
 edge_array<int>& edge_weight)
{
	edge e;
	forall_edges(e,G_aux) {
		node se=G_aux.source(e);
		node te=G_aux.target(e);
        int rse=aux_rank[se];
		edge_weight[e]=0;
		two_tuple<int,int> p;
		forall(p,sles[rse]) {
			if ((p.first()<aux_pos[se]&&p.second()>aux_pos[te])||
			    (p.first()>aux_pos[se]&&p.second()<aux_pos[te])) {
				edge_weight[e]++;
			}
		}
	}
}

/*-----------------------------------------------------------------------------*/

void InteractiveSugiyama::insert_single_edge
(const graph& G,edge the_new_edge, array<list<two_tuple<int,int> > >& sles)
{
	node s=G.source(the_new_edge),t=G.target(the_new_edge);
	if (_node_rank[s]>_node_rank[t]) {node help=s;s=t;t=help;}

	if (_node_rank[s]==_node_rank[t]-1) {
		two_tuple<int,int> p(0,0);
		forall(p,sles[_node_rank[s]]) {
			if ((p.first()<_node_pos[s]&&p.second()>_node_pos[t])||
			    (p.first()>_node_pos[s]&&p.second()<_node_pos[t])) {
				_num_cross++;
			}
		}
		return;
	}

	graph G_aux;
	node_map<int> aux_rank(G_aux);
	node_map<double> aux_pos(G_aux);
	node s_aux=s, t_aux=t;
	compute_auxiliary_graph(G_aux,aux_rank,aux_pos,s_aux,t_aux);

	edge_array<int> edge_weight(G_aux,0);
	compute_edge_weights(G_aux,aux_rank,aux_pos,s_aux,t_aux,sles,edge_weight);

	node_array<int> dist(G_aux);
	node_array<edge> pred(G_aux);
	SHORTEST_PATH(G_aux,s_aux,edge_weight,dist,pred);

	_num_cross+=dist[t_aux];

	list<node> dummies;
	node help_node=G_aux.source(pred[t_aux]);
	while (help_node!=s_aux) {
		dummies.push(help_node);
		help_node=G_aux.source(pred[help_node]);
	}

	while (!dummies.empty()) {
		help_node=dummies.pop();
		int newpos=static_cast<int>(aux_pos[help_node])+1;
		if (aux_pos[help_node]<0) newpos=0;

		if (newpos>=_layersize[aux_rank[help_node]]) {
			newpos=_layersize[aux_rank[help_node]];
		}

		else {make_room_in_layer(G,aux_rank[help_node],newpos);}

		two_tuple<int,int> p(aux_rank[help_node],newpos);
		_edge_positions[the_new_edge].append(p);
		_layersize[aux_rank[help_node]]++;
	}
}

