/*******************************************************************************
+
+  AGD 1.2
+
+
+  LongestPathCompaction.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



#include <AGD/LongestPathCompaction.h>
#include <LEDA/graph/node_partition.h>

#include <LEDA/geo/interval_set.h>

#include <AGD/GraphCopy.h>
#include <AGD/impl/CompactionPair.h>


using namespace leda;

AGDModule *LongestPathCompaction::clone() const { 
  LongestPathCompaction *C = new LongestPathCompaction(); 
  return C; 
}


double LongestPathCompaction::call(GridLayout& L) {

   double l_old, l_new; 
  node v;

  do {
    
    l_old = L.total_manhattan_edge_length();

    CompactionPair X(L);
    X.compute_coords(CompactionPair::vertical);
    forall_nodes(v, L.get_graph()) L.x()[v] = (int) X.x_coord(v);
       
    CompactionPair Y(L);
    Y.compute_coords(CompactionPair::horizontal);
    forall_nodes(v, L.get_graph()) L.y()[v] = (int) Y.y_coord(v);

    l_new = L.total_manhattan_edge_length();
    
  } while (l_old - l_new > 0.0);
  
  return l_new;  

}


void LongestPathCompaction::assign_gridlines(graph& G, 
                                             const node_array<double>& x, 
                                             node_map<int>& x_grid) {
  
  G.sort_nodes(x); double x_act = x[G.first_node()]; int row = 0;
  node v; forall_nodes(v, G) { 
    if (x[v] > x_act) ++row; 
    x_grid[v] = row; x_act = x[v]; 
  } 
}


int LongestPathCompaction::call(const graph& G, LayoutInterface& I) {
  
  GraphCopy GC(G); node_array<double> x(GC), y(GC); node v;
  forall_nodes(v, GC) { x[v] = I.get_x(v); y[v] = I.get_y(v); }
  // at this point I assume that the graph contains no bends
  
  GridLayout gl(GC);
  
  assign_gridlines(GC, x, gl.x()); assign_gridlines(GC, y, gl.y());
  
  int result = (int) call(gl);
  
  forall_nodes(v, GC) { I.set_x(v, gl.x(v)); I.set_y(v, gl.y(v)); }

  return result;
  
}

  
int LongestPathCompaction::call(const graph& G, node_map<int>& x,
                                node_map<int>& y) {
  
  
node_partition P(G);

edge e; forall_edges(e, G) if (x[source(e)] == x[target(e)])
  P.union_blocks(source(e), target(e));

graph C; node_array<node> link(G, nil);

node v; forall_nodes(v, G) if (!link[P(v)]) link[P(v)] =
C.new_node();


                                                            ;
  
interval_set<node> I;
typedef three_tuple<double, double, node> triple;   

node_array<int> fix(C);
node_array<double> hi(C, -MAXDOUBLE), lo(C, MAXDOUBLE);

forall_nodes(v, G) { 
  node c = link[P(v)]; 
  fix[c] = x[v]; 
  hi[c] = leda::max(hi[c], (double) y[v]); 
  lo[c] = leda::min(lo[c], (double) y[v]); 
}

C.bucket_sort_nodes(fix);

node c; forall_nodes(c, C) { 
  
  list<is_item> L = I.intersection(lo[c], hi[c]);
  list<triple> L_upd; // to update the data structure after having processed 
  // all segments on one vertical line
  
  //obacht: geht vielleicht eleganter mit list<is_item> statt list<triple>

  is_item it; forall(it, L) {
    
    double it_lo = I.left(it), it_hi = I.right(it); 
    node c_it = I.inf(it); I.del_item(it); 

    if (it_hi > hi[c]) L_upd.append(triple(hi[c], it_hi, c_it));
    if (it_lo < lo[c]) L_upd.append(triple(it_lo, lo[c], c_it)); 

    C.new_edge(c_it, c);
    
  }
  
  triple t; forall(t, L_upd) I.insert(t.first(), t.second(), t.third());
              
  I.insert(lo[c], hi[c], c);
  
}

                                             ;
      
  node_array<int> l(C, 0); longest_paths(C, l);
  
  int result = 0; forall_nodes(v, G) { 
    result += x[v]; 
    x[v] = l[link[P(v)]]; 
    result -= x[v]; 
  }
  
  return result;
  
}

// int LongestPathCompaction::call(const OrthogonalRepresentation& H, 
//                                 GridLayout& L, list<Label> &Lambda) {
 
//   return call(H, L);

// }


int LongestPathCompaction::call(const OrthogonalRepresentation& H, 
                                GridLayout& L) {

  string msg; if (!H.check(msg)) error_handler(1, msg + " (H)");

  OrthogonalRepresentation H_prime(H, true);
  
  if (!H_prime.check(msg)) error_handler(1, msg + " (H' after construction)");

  H_prime.dissect(); 

  if (!H_prime.check(msg)) error_handler(1, msg + " (H' after dissect)");

  CompactionPair C(H_prime);
  
  C.compute_coords(CompactionPair::vertical);
  C.compute_coords(CompactionPair::horizontal);

  if (debug_) {
   /*  W_debug.set_graph(H_prime); */
/*     node v; forall_nodes(v, H_prime) { */
/*       W_debug.set_position(v, point(100*C.x_coord(v),  */
/*                                100*C.y_coord(v))); */
/*     } */
/*     W_debug.open(); */
/*     W_debug.fill_window(); */

  }

  node v; forall_nodes(v, H) {
    L.x()[v] = (int) C.x_coord(H_prime.copy(v));
    L.y()[v] = (int) C.y_coord(H_prime.copy(v));
  }

  return L.total_manhattan_edge_length();
  
}

void LongestPathCompaction::call(const OrthogonalRepresentation& H, 
                                 node_map<int>& x,
                                 bool horizontally) {

  //const graph &G = H.get_graph();
  
  
node_partition P(H);

edge e; forall_edges(e, H) if ((horizontally && H.ver(e)) ||
                               (!horizontally && H.hor(e))) 
  P.union_blocks(source(e), target(e));

graph C; node_array<node> link(H, nil);

node v; forall_nodes(v, H) if (!link[P(v)]) link[P(v)] =
C.new_node();

                                                            ;
      
  forall_edges(e, H) if ((horizontally && H.hor(e)) ||
                         (!horizontally && H.ver(e)))
    C.new_edge(link[P.find(source(e))], link[P.find(target(e))]);
  
  node_array<int> l(C, 0); longest_paths(C, l);
  
  forall_nodes(v, H) x[v] = l[link[P(v)]];
 
}
  


#include <LEDA/core/stack.h>

void LongestPathCompaction::longest_paths(const graph& D, 
                                   node_array<int>& l) {

  node_array<int> delta_in(D, 0); stack<node> S;
  
  node v; forall_nodes(v, D) { delta_in[v] = indeg(v); if
    (delta_in[v] == 0) S.push(v); }
  
  while (!S.empty()) {
    v = S.pop(); 
    edge e; forall_out_edges(e, v) { 
      node w = target(e); 
      l[w] = leda::max(l[w], l[v] + 1); 
      if (--delta_in[w] == 0) S.push(w); 
    } 
  }
  
}

