/*******************************************************************************
+
+  AGD 1.2
+
+
+  heuristics.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        heuristics.c
//
//    author: Carsten Gutwenger
//            Michael Krger
//    date:   decembre 1998
//-------------------------------------------------------------------


#include <AGD/BarycenterHeuristic.h>
#include <AGD/MedianHeuristic.h>
#include <AGD/WeightedMedianHeuristic.h>
#include <AGD/GreedyInsertHeuristic.h>
#include <AGD/GreedySwitchHeuristic.h>
#include <AGD/SplitHeuristic.h>
#include <AGD/SiftingHeuristic.h>
#include <AGD/impl/CrossingsMatrix.h>
#include <AGD/Hierarchy.h>
#include <AGD/impl/misc.h>


using namespace leda;

//-------------------------------------------------------------------
//                        TwoLayerCrossMin
//-------------------------------------------------------------------

const int TwoLayerCrossMin::in_hierarchy = 1;
const int TwoLayerCrossMin::in_layer = 2;
const int TwoLayerCrossMin::out_layer = 1;


//-------------------------------------------------------------------
//                        BarycenterHeuristic
//-------------------------------------------------------------------

void BarycenterHeuristic::call (const graph& , Layer& L)
{
	const Hierarchy& H = L.hierarchy();
	node v, w;
	int i;

	for (i = 0; i < L.length(); i++) {
		v = L[i];
		long sumpos = 0L;
    
		forall (w, L.adj_nodes(v))
			sumpos += H.pos(w);

		weight[v] = L.adj_nodes(v).empty() ? 0.0 :
			(double) sumpos /
			(double) L.adj_nodes(v).size();
	}
	
	L.quicksort(weight);
}


//-------------------------------------------------------------------
//                        MedianHeuristic
//-------------------------------------------------------------------

void MedianHeuristic::call (const graph& , Layer& L)
{
	const Hierarchy& H = L.hierarchy();
	const Layer &LN = L.neighbored_layer();

	array<int> count(L.length()), odds(L.length());

	int i, vpos;
	node v;

	for (i = 0; i < L.length(); i++) {
		v = L[i];
		count [i] = (L.adj_nodes(v).size()+1) / 2;
		odds  [i] = 1;
		weight[v] = 0.0;
	}

	for (i = 0; i < LN.length(); i++) {
		const list<node>& adj = LN.adj_nodes(LN[i]);
		forall (v,adj) {
			vpos = H.pos(v);
			if (--count[vpos] == 0) {
				if (L.adj_nodes(v).length() & 1)
					weight[v] = i;
				else {
					weight[v] += (double)i / 2.;
					count[vpos] += odds[vpos]--;
				}
                        }
		}
	}

	L.quicksort(weight);
}


//-------------------------------------------------------------------
//                      WeightedMedianHeuristic
//-------------------------------------------------------------------

void WeightedMedianHeuristic::init(const leda::graph& G, const Hierarchy& )
{
	pos.init(G);

	node v;
	forall_nodes(v,G) {
		int m = leda::max(G.indeg(v),G.outdeg(v));
		if (m > 0)
			pos[v] = array<int>(m);
	}

	weight.init(G);
	size.init(G);
}


void WeightedMedianHeuristic::cleanup ()
{
	weight.init();
	size.init();
	pos.init();
}


void WeightedMedianHeuristic::call (const graph& , Layer& L)
{
	const Layer &LN = L.neighbored_layer();

	int i;
	node v;

	for (i = 0; i < L.length(); i++)
		size[L[i]] = 0;
	
	for (i = 0; i < LN.length(); i++) {
		const list<node>& adj = LN.adj_nodes(LN[i]);
		forall (v,adj)
			(pos[v])[size[v]++] = i;
	}

	for (i = 0; i < L.length(); i++) {
		v = L[i];
		const array<int> &P = pos[v];
		int len = size[v];
		int m = len / 2;

		if (len == 0) {
			weight[v] = 0.;
		} else if (len & 1) {
			weight[v] = P[m];
		} else if (len == 2) {
			weight[v] = (double)(P[0]+P[1]) / 2.;
		} else {
			int left  = P[m-1]   - P[0];
			int right = P[len-1] - P[m];
			weight[v] = (double)(P[m-1]*right+P[m]*left) / (left+right);
		}
	}

	L.quicksort(weight);
}


//-------------------------------------------------------------------
//                       GreedyInsertHeuristic
//-------------------------------------------------------------------

void GreedyInsertHeuristic::init (const leda::graph& G, const Hierarchy& H)
{
	weight.init(G);
	cp = new CrossingsMatrix(G,H);
}

void GreedyInsertHeuristic::cleanup()
{
	weight.init();
	delete cp;
}

void GreedyInsertHeuristic::call( const graph &, Layer &L )
{
	cp->init(L);
	int index, i;
	long prio;

	// Initialisierung und Prioritaetszuweisung 
	for (i = 0; i < L.length(); i++) {
		for (prio = index = 0; index < L.length(); index++)
			prio += (*cp)(i,index);

		// stable quicksort: no need for unique prio 
		weight[L[i]] = (double) prio;
	}

	L.quicksort(weight);
}


//-------------------------------------------------------------------
//                       GreedySwitchHeuristic
//-------------------------------------------------------------------

void GreedySwitchHeuristic::init (const leda::graph& G, const Hierarchy& H)
{
	cp = new CrossingsMatrix(G,H);
}

void GreedySwitchHeuristic::cleanup()
{
	delete cp;
}

void GreedySwitchHeuristic::call(const graph &, Layer &L)
{
	cp->init(L);
	int index;
	bool nolocalmin;

	do {
		nolocalmin = false;

		for (index = 0; index < L.length() - 1; index++)
			if ((*cp)(index,index+1) > (*cp)(index+1,index)) {

				nolocalmin = true;

				L.swap(index,index+1);
				cp->swap(index,index+1);
			}
	} while (nolocalmin);
}


//-------------------------------------------------------------------
//                       SiftingHeuristic
//-------------------------------------------------------------------

AGDModule *SiftingHeuristic::clone() const
{
	SiftingHeuristic *m = new SiftingHeuristic;
	m->strategy(strategy());
	return m;
}

void SiftingHeuristic::init (const leda::graph& G, const Hierarchy& H)
{
	cp = new CrossingsMatrix(G,H);
}

void SiftingHeuristic::cleanup()
{
	delete cp;
}

void SiftingHeuristic::call(const graph &, Layer &L)
{
	const Hierarchy &H = L.hierarchy();
	list<node> vertices;
	int i;

	const int n = L.length();

	cp->init(L); // initialize crossing matrix

	if (_strategy == left_to_right || _strategy == random) {
		for (i = 0; i < n; i++)
			vertices.push_back(L[i]);
		
		if (_strategy == random)
			permute(vertices);

	} else { // _strategy == desc_degree
		int max_deg = 0;
		
		for (i = 0; i < n; i++) {
			int deg = L.adj_nodes(L[i]).size();
			if (deg > max_deg) max_deg = deg;
		}

		array<list<node> > bucket(0,max_deg);
		for (i = 0; i < n; i++)
			bucket[L.adj_nodes(L[i]).size()].push_back(L[i]);

		for (i = max_deg; i >= 0; i--)
			while(!bucket[i].empty())
				vertices.push_back(bucket[i].pop_front());
	}

	node w;
	forall(w,vertices) {
		i = H.pos(w);
		int dev = 0;

		// sifting left
		for(; i > 0; --i) {
			dev = dev - (*cp)(i-1,i) + (*cp)(i,i-1);
			L.swap(i-1,i);
			cp->swap(i-1,i);
		}

		// sifting right and searching optimal position
		int opt = dev, opt_pos = 0;
		for (; i < n-1; ++i) {
			dev = dev - (*cp)(i,i+1) + (*cp)(i+1,i);
			L.swap(i,i+1);
			cp->swap(i,i+1);
			if (dev <= opt) {
				opt = dev; opt_pos = i+1;
			}
		}

		// set optimal position
		for (; i > opt_pos; --i) {
			L.swap(i-1,i);
			cp->swap(i-1,i);
		}
	}
}

  
//-------------------------------------------------------------------
//                          SplitHeuristic
//-------------------------------------------------------------------

void SplitHeuristic::init (const leda::graph& G, const Hierarchy& H)
{
	cp = new CrossingsMatrix(G,H);
}

void SplitHeuristic::cleanup()
{
	delete cp;
}

void SplitHeuristic::call( const graph &, Layer &L )
{
	cp->init(L); // initialize crossing matrix
	buffer = LEDA_NEW_VECTOR(node,L.length());

	_rec_call(L, 0, L.length() - 1);

	LEDA_DEL_VECTOR(buffer);
}


void SplitHeuristic::_rec_call(Layer &L, int low, int high )
{
	if (high <= low) return;

	const Hierarchy &H = L.hierarchy();
	CrossingsMatrix &crossings = *cp;

	int up = high, down = low;
	
	int i;
	for (i = low+1; i <= high; i++) {
		if (crossings(i,low) < crossings(low,i))
			buffer[down++] = L[i];
	}
	for (i = high; i >= low+1; i--) {
		if (crossings(i,low) >= crossings(low,i))
			buffer[up--] = L[i];
	}

	// set pivot in the middle (now, up == down holds!)
	buffer[down] = L[low];

	for (i = low; i < high; i++) { // i == high not necessary!
		int j = H.pos(buffer[i]);
		if (i != j) {
			L.swap(i,j);
			crossings.swap(i,j);
		}
	}

	_rec_call(L,low,down-1);
	_rec_call(L,up+1,high);

}


//-------------------------------------------------------------------
//                        CrossingMatrix
//-------------------------------------------------------------------

CrossingsMatrix::CrossingsMatrix(const leda::graph &, const Hierarchy &H) :
	matrix(1,1)
{
	int max_len = 0;
	for (int i = 0; i < H.number_of_layers(); i++) {
		int len = H[i].length();
		if (len > max_len) max_len = len;
	}

	_map = LEDA_NEW_VECTOR(int,max_len);
	matrix = array2<int>(max_len,max_len);
}

CrossingsMatrix::~CrossingsMatrix()
{
	matrix = array2<int>(0,0);
	LEDA_DEL_VECTOR(_map);
}

void CrossingsMatrix::init(const Layer &L)
{
	const Hierarchy &H = L.hierarchy();
	const int n = L.length();

	int i, j;
	for (i = 0; i < n; i++) {
		_map[i] = i;
		for (j = 0; j < n; j++)
			matrix(i,j) = 0;
	}

	for (i = 0; i < n; i++) {
		const list<node> &L_adj_i = L.adj_nodes(L[i]);

		node adj_v;
		forall (adj_v, L_adj_i) {
			int pos_adj_v = H.pos(adj_v);

			for (j = i + 1; j < n; j++) {
				const list<node> &L_adj_j = L.adj_nodes(L[j]);

				node adj_w;
				forall (adj_w, L_adj_j) {
					int pos_adj_w = H.pos(adj_w);
					matrix(i,j) += (pos_adj_v > pos_adj_w);
					matrix(j,i) += (pos_adj_w > pos_adj_v);
				}
			}
		}
	}
}
