// compression module includes
#include <LEDA/coding/RLE.h>

LEDA_BEGIN_NAMESPACE

/// RLECoder ////////////////////////////////////////////////////////////////////////////////

RLECoder::RLECoder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams, bool fast_mode)
: base(src_stream, tgt_stream, own_streams), FastMode(fast_mode)
{}

RLECoder::RLECoder(const char* src_file_name, const char* tgt_file_name, bool fast_mode)
: base(src_file_name, tgt_file_name), FastMode(fast_mode)
{}

void 
RLECoder::do_reset(reset_type t)
{
	base::do_reset(t);
	if (t == r_hard) {
		set_fast_compression_mode();
	}
}

void 
RLECoder::set_fast_compression_mode()
{
	if (! running()) FastMode = true;
	else error("coder_base::set_compression_mode: wrong state");
}

void 
RLECoder::set_high_compression_mode()
{
	if (! running()) FastMode = false;
	else error("coder_base::set_compression_mode: wrong state");
}

string 
RLECoder::get_description() const
{
	return base::get_description() + (FastMode ? string("[fast]") : string("[high]"));
}

void 
RLECoder::check_id(id_type id_to_check)
{
	if (id_to_check != get_id()) {
		// if the id does not match the current id, toggle the mode
		FastMode = !FastMode;
	}

	base::check_id(id_to_check);
}

void 
RLECoder::init_encoding()
{
	base::init_encoding();

	if (FastMode) init_encoding_fast_mode();
	else          init_encoding_slow_mode();
}

void 
RLECoder::encode_stream()
{
	if (FastMode) encode_stream_fast_mode();
	else          encode_stream_slow_mode();
}

void 
RLECoder::init_decoding()
{
	base::init_decoding();

	if (FastMode) init_decoding_fast_mode();
	else          init_decoding_slow_mode();
}

void 
RLECoder::decode_stream()
{
	while (Count > 0) {
		--Count;
		if (write_byte(PreviousSymbol) < 0) return; 
	}

	if (FastMode) decode_stream_fast_mode();
	else          decode_stream_slow_mode();
}


/// encode fast

void 
RLECoder::init_encoding_fast_mode()
{
	int c = read_byte();
	if (c >= 0 && c != EOF) {
		PreviousSymbol = (byte) c;
		write_byte(PreviousSymbol);
		Count = 0;
	}
	else { // c < 0
		if (c == EOF) encoding_finished();
		else          error("RLE2Coder::init_encoding: Unexpected STREAM_BLOCKED");
		return;
	}
}

void 
RLECoder::encode_stream_fast_mode()
{
	int c;
	while ((c = read_byte()) >= 0 && c != EOF) {
		byte symbol = (byte) c;

		if (symbol == PreviousSymbol) { // symbol repeated
			if (++Count == 257) {
				// overflow => write out symbol
				write_byte(symbol);
				write_byte((byte) Count-2);
				Count = 1;
			}
		}
		else { // new symbol, old one has to be written
			if (Count > 0) {
				write_byte(PreviousSymbol);
				write_byte((byte) Count-1);
			}
			write_byte(symbol);
			PreviousSymbol = symbol;
			Count = 0;
		}
	}

	if (c == EOF) {
		if (Count > 0) {
			write_byte(PreviousSymbol);
			write_byte((byte) Count-1);
		}
		encoding_finished();
	}
}


/// decode fast

void 
RLECoder::init_decoding_fast_mode()
{
	int c = read_byte();
	if (c >= 0 && c != EOF) {
		PreviousSymbol = (byte) c;
		Count = 1;
	}
	else { // c < 0
		if (c == EOF) decoding_finished();
		else          error("RLE2Coder::init_encoding: Unexpected STREAM_BLOCKED");
		return;
	}
}

void 
RLECoder::decode_stream_fast_mode()
{
	int c;
	while ((c = read_byte()) != EOF) {
		byte symbol = (byte) c;

		if (symbol == PreviousSymbol) {
			if ((c = read_byte()) == EOF) { error(); return; }
			Count = c + 1;
			while (Count > 0) {
				--Count;
				if (write_byte(symbol) < 0) return; 
			}
		}
		else {
			PreviousSymbol = symbol;
			if (write_byte(symbol) < 0) return;
		}
	}

	decoding_finished();
}


/// encode slow

void 
RLECoder::init_encoding_slow_mode()
{
	int c = read_byte();
	if (c >= 0 && c != EOF) {
		PreviousSymbol = (byte) c;
		Count = 1;
	}
	else { // c < 0
		if (c == EOF) encoding_finished();
		else          error("RLECoder::init_encoding: Unexpected STREAM_BLOCKED");
		return;
	}
}

void 
RLECoder::encode_stream_slow_mode()
{
	int c;
	for(;;) {
		if ((c = read_byte()) == StreamBlocked) return;

		if (c != PreviousSymbol || ++Count == 0) { // overflow or different symbol
			if (Count == 0) --Count;

			// encode previous symbol
			uint32 cnt = Count-1; // Count is at least one
			if (cnt < 2) {
				if (PreviousSymbol == sentinel) {
					write_byte(sentinel);
					write_byte((byte) cnt);
				}
				else do {
					write_byte(PreviousSymbol);
				} while (cnt--);
			}
			else {
				write_byte(sentinel);
				if (cnt < 255)
					write_byte((byte) cnt);
				else {
					write_byte(0xFF); cnt -= 255;
					do {
						byte b = byte(cnt & 0x7f);
						cnt >>= 7;
						if (cnt > 0) b |= 0x80;
						write_byte(b);
					} while (cnt > 0);
				}
				write_byte(PreviousSymbol);
			}

			if (c == EOF) break;

			PreviousSymbol = (byte) c;
			Count = 1;
		}
	}

	encoding_finished();
}


/// decode slow

void 
RLECoder::init_decoding_slow_mode()
{
	Count = 0;
}

void 
RLECoder::decode_stream_slow_mode()
{
	int c;
	while ((c = read_byte()) != EOF) {
		byte symbol = (byte) c;

		if (symbol != sentinel) {
			if (write_byte(symbol) < 0) return;
		}
		else {
			c = read_byte();
			if (c == EOF) { error("RLECoder::decode_stream: unexpected EOF"); return; }
			if (c < 2) { // -> symbol was sentinel itself
				if (c == 1 && write_byte(sentinel) < 0) { 
					Count = 1; PreviousSymbol = sentinel; 
					return; 
				}
				if (write_byte(sentinel) < 0) return;
			}
			else {
				uint32 cnt = (uint32) c + 1; // +1 -> undo "cnt = Count-1"
				if (c == 0xFF) {
					int shift = 0;
					do {
						c = read_byte();
						if (c == EOF) { error("RLECoder::decode_stream: unexpected EOF"); return; }
						uint32 bits = c & 0x7f;
						cnt += (bits << shift); shift += 7;
					} while (c & 0x80);
				}

				if ((c = read_byte()) == EOF) { error("RLECoder::decode_stream: unexpected EOF"); return; }
				symbol = (byte) c;

				while (cnt > 0) {
					--cnt;
					if (write_byte(symbol) < 0) {
						Count = cnt; PreviousSymbol = symbol; 
						return; 
					}
				}
			}
		}
	}

	decoding_finished();
}

LEDA_END_NAMESPACE
