// compression module includes
#include <LEDA/coding/auto_decoder.h>

LEDA_BEGIN_NAMESPACE

/// AutoDecoder /////////////////////////////////////////////////////////////////////////////

AutoDecoder::AutoDecoder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams),
  pActualCoder(0), pAdditionalStream(0), OwnCoder(false), OwnAdditionalStream(false)
{}

AutoDecoder::AutoDecoder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name),
  pActualCoder(0), pAdditionalStream(0), OwnCoder(false), OwnAdditionalStream(false)
{}

AutoDecoder::~AutoDecoder()
{
	set_state(s_finished); // clears errors
	//clear_parameters_lock();
	set_coder(0);
}

void
AutoDecoder::do_reset(reset_type t)
{
	bool was_decoding = decoding();
	base::do_reset(t);
	if (t == r_hard || (t == r_soft && was_decoding)) set_coder(0);
}

string 
AutoDecoder::get_description() const
{
	string result = base::get_description();

	if (get_coder()) result += string('[') + get_coder()->get_description() + string(']');

	return result;
}

bool 
AutoDecoder::is_optimizing_seeks() const
{
	return get_coder() ? get_coder()->is_optimizing_seeks() : false;
		// if in doubt, rather say no
}

bool 
AutoDecoder::needs_seeks_during_coding() const
{
	return get_coder() ? get_coder()->needs_seeks_during_coding() : false;
		// if in doubt, rather say no
}

bool 
AutoDecoder::is_secrecy_sensitive() const
{
	return get_coder() ? get_coder()->is_secrecy_sensitive() : true;
		// if in doubt, rather say yes
}

void
AutoDecoder::set_coder(coder_base* coder, bool own_coder)
{
	if (can_change_parameters()) {
		if (OwnCoder) delete pActualCoder;
		pActualCoder = coder; OwnCoder = own_coder;
	}
	else error("AutoDecoder::set_coder: wrong state");
}

void 
AutoDecoder::set_additional_stream(streambuf* additional_stream, bool own_stream)
{
	if (can_change_parameters()) {
		if (OwnAdditionalStream) delete pAdditionalStream;
		pAdditionalStream = additional_stream; OwnAdditionalStream = own_stream;
	}
	else error("AutoDecoder::set_additional_stream: wrong state");
}

void 
AutoDecoder::set_additional_file(const char* file_name, ios::openmode mode)
{
	if (file_name == 0) { set_src_stream(0, false); return; }

	filebuf* stream = new filebuf;
	if (! stream->open(file_name, mode)) {
		delete stream; stream = 0;
		error("AutoDecoder::set_additional_file: open failed");
	}

	set_additional_stream(stream, true); // own it and destroy it
}

void 
AutoDecoder::init_encoding()
{
	if (! get_coder()) {
		error("AutoDecoder::init_encoding: coder missing");
		return;
	}

	get_coder()->set_src_stream(get_src_stream(), false);
	get_coder()->set_tgt_stream(get_tgt_stream(), false);

	process_no_id_in_streams(); process_no_header_in_streams();
	base::init_encoding();
}

void 
AutoDecoder::encode_stream()
{
	get_coder()->do_encode();
	if (get_coder()->failed()) { error(); encoding_finished(); }
	else if (get_coder()->finished()) { encoding_finished(); }
}

void 
AutoDecoder::init_decoding()
{
	coder_base* coder = 0;
	
	if (get_src_stream()) coder = construct_from_stream(get_src_stream(), get_additional_stream());

	set_coder(coder, true);

	if (! coder) {
		error("AutoDecoder::init_decoding: failed to construct coder");
		return;
	}

	coder->set_src_stream(get_src_stream(), false);
	coder->set_tgt_stream(get_tgt_stream(), false);

	process_no_id_in_streams(); process_no_header_in_streams();
	base::init_decoding();
}

void 
AutoDecoder::decode_stream()
{
	get_coder()->do_decode();
	if (get_coder()->failed()) { error(); decoding_finished(); }
	else if (get_coder()->finished()) { decoding_finished(); }
}

streampos 
AutoDecoder::encode_seek(streampos current_pos, streampos new_pos)
{
	if (get_coder()) {
		new_pos = get_coder()->encode_seek(current_pos, new_pos);

		// synchronize *this and *pActualCoder ...
		if (get_coder()->failed()) { error(); encoding_finished(); }
		else if (get_coder()->finished()) { encoding_finished(); }
		else set_state(s_encoding);

		return new_pos;
	}
	return base::encode_seek(current_pos, new_pos);
}

streampos 
AutoDecoder::decode_seek(streampos current_pos, streampos new_pos)
{
	if (!decoding() || failed()) return base::decode_seek(current_pos, new_pos);
	if (get_coder()) {
		new_pos = get_coder()->decode_seek(current_pos, new_pos);

		// synchronize *this and *pActualCoder ...
		if (get_coder()->failed()) { error(); decoding_finished(); }
		else if (get_coder()->finished()) { decoding_finished(); }
		else set_state(s_decoding); // make it running ...

		return new_pos;
	}
	return base::decode_seek(current_pos, new_pos);
}

streambuf* 
AutoDecoder::change_src_stream_temporarily(streambuf* src_stream)
{
	if (running() && get_coder()) get_coder()->change_src_stream_temporarily(src_stream);
	return base::change_src_stream_temporarily(src_stream);
}

streambuf* 
AutoDecoder::change_tgt_stream_temporarily(streambuf* tgt_stream)
{
	if (running() && get_coder()) get_coder()->change_tgt_stream_temporarily(tgt_stream);
	return base::change_tgt_stream_temporarily(tgt_stream);
}

LEDA_END_NAMESPACE
