// compression module includes
#include <LEDA/coding/coder_factory.h>
#include <LEDA/coding/block_cipher_modes.h>
#include <LEDA/coding/coder_pipe.h>
#include <LEDA/coding/arithmetic_coding.h>
#include <LEDA/coding/block_coder.h>
#include <LEDA/coding/BWT.h>
#include <LEDA/coding/checksum.h>
#include <LEDA/coding/deflate.h>
#include <LEDA/coding/huffman.h>
#include <LEDA/coding/dict_coder.h>
#include <LEDA/coding/MTF.h>
#include <LEDA/coding/MTF2.h>
#include <LEDA/coding/PPMII.h>
#include <LEDA/coding/RLE.h>
#include <LEDA/coding/RLE0.h>

// core LEDA includes
#include <LEDA/system/error.h>

LEDA_BEGIN_NAMESPACE

/// coder_factory ///////////////////////////////////////////////////////////////////////////

#if defined(LEDA_MULTI_THREAD)
leda_lock coder_factory::mutex;
#endif

coder_factory CoderFactory; // The one and only instance

coder_factory::coder_factory()
{
	if (this != &CoderFactory)
		LEDA_EXCEPTION(1, "coder_factory is a singleton!");
}

#define REGISTER_CODER(CODER) register_id(CODER::ID, construct_##CODER)

// define a constructor function for a standalone coder (default)
#define DEF_CTOR_FUNC_SA(CODER) \
	static coder_base* construct_##CODER(coder_factory::id_type id, bool as_standalone) \
	{ return new CODER; }

// define a constructor function for a coder which cannot be used standalone
#define DEF_CTOR_FUNC_NS(CODER) \
	static coder_base* construct_##CODER(coder_factory::id_type id, bool as_standalone) \
	{ return as_standalone ? 0 : new CODER(); }

DEF_CTOR_FUNC_NS(block_coder);
DEF_CTOR_FUNC_NS(coder_pipe);
DEF_CTOR_FUNC_NS(cbc_mode);
DEF_CTOR_FUNC_NS(cfb_mode);
DEF_CTOR_FUNC_NS(ecb_mode);
DEF_CTOR_FUNC_NS(ofb_mode);
DEF_CTOR_FUNC_NS(omac_mode);
DEF_CTOR_FUNC_SA(A0Coder);
DEF_CTOR_FUNC_SA(A0sCoder);
DEF_CTOR_FUNC_SA(AdaptiveHuffmanCoder);
DEF_CTOR_FUNC_SA(Adler32Coder);
DEF_CTOR_FUNC_SA(BWTCoder);
DEF_CTOR_FUNC_SA(CCITTCoder);
DEF_CTOR_FUNC_SA(CRC32Coder);
DEF_CTOR_FUNC_SA(DeflateCoder);
DEF_CTOR_FUNC_SA(DictCoder);
DEF_CTOR_FUNC_SA(HuffmanCoder);
DEF_CTOR_FUNC_SA(MD5SumCoder);
DEF_CTOR_FUNC_SA(MTFCoder);
DEF_CTOR_FUNC_SA(MTF2Coder);
DEF_CTOR_FUNC_SA(PPMIICoder);
DEF_CTOR_FUNC_SA(RLE0Coder);
DEF_CTOR_FUNC_SA(SHACoder);

static coder_base* construct_RLECoder(coder_factory::id_type id, bool as_standalone)
{ return new RLECoder((streambuf*)0, (streambuf*)0, false, id == RLECoder::ID_Fast); }


void 
coder_factory::init()
{
	pCoderConstructorMap = new h_array<id_type, coder_construction_func>(0);

	// register the LEDA coders ...
	REGISTER_CODER(block_coder);
	REGISTER_CODER(coder_pipe);
	REGISTER_CODER(cbc_mode);
	REGISTER_CODER(cfb_mode);
	REGISTER_CODER(ecb_mode);
	REGISTER_CODER(ofb_mode);
	REGISTER_CODER(omac_mode);

	REGISTER_CODER(A0Coder);
	REGISTER_CODER(A0sCoder);
	REGISTER_CODER(AdaptiveHuffmanCoder);
	REGISTER_CODER(Adler32Coder);
	REGISTER_CODER(BWTCoder);
	REGISTER_CODER(CCITTCoder);
	REGISTER_CODER(CRC32Coder);
	REGISTER_CODER(DeflateCoder);
	REGISTER_CODER(DictCoder);
	REGISTER_CODER(HuffmanCoder);
	REGISTER_CODER(MD5SumCoder);
	REGISTER_CODER(MTFCoder);
	REGISTER_CODER(MTF2Coder);
	REGISTER_CODER(PPMIICoder);
	register_id(RLECoder::ID_Fast, construct_RLECoder);
	register_id(RLECoder::ID_High, construct_RLECoder);
	REGISTER_CODER(RLE0Coder);
	REGISTER_CODER(SHACoder);
}

void 
coder_factory::destroy()
{
	delete pCoderConstructorMap; pCoderConstructorMap = 0;
}

void 
coder_factory::register_id(id_type id, coder_construction_func ctor)
{
	lock();

	if (pCoderConstructorMap->defined(id)) { unlock(); return; }
	(*pCoderConstructorMap)[id] = ctor;
	unlock(); 
}

coder_base* 
coder_factory::construct_coder_by_id(id_type id)
{
	lock();

	coder_base* coder = 0;
	if (pCoderConstructorMap->defined(id)) {
		coder_construction_func ctor = (*pCoderConstructorMap)[id];
		coder = ctor(id, false);
	}

	unlock();

	return coder;
}

list<coder_base*> 
coder_factory::get_all_coders()
{
	list<coder_base*> all_coders;

	lock();

	id_type id;
	forall_defined(id, *pCoderConstructorMap) {
		coder_construction_func ctor = (*pCoderConstructorMap)[id];
		coder_base* p_coder = ctor(id, false);
		if (p_coder) all_coders.append(p_coder);
	}

	unlock();

	return all_coders;
}

list<coder_base*> 
coder_factory::get_standalone_coders()
{
	list<coder_base*> standalone_coders;

	lock();

	id_type id;
	forall_defined(id, *pCoderConstructorMap) {
		coder_construction_func ctor = (*pCoderConstructorMap)[id];
		coder_base* p_coder = ctor(id, true);
		if (p_coder) standalone_coders.append(p_coder);
	}

	unlock();

	return standalone_coders;
}

list<block_cipher_mode_base*> 
coder_factory::get_block_cipher_modes()
{
	list<block_cipher_mode_base*> bc_modes;
	bc_modes.append( new cbc_mode );
	bc_modes.append( new cfb_mode );
	bc_modes.append( new ecb_mode );
	bc_modes.append( new ofb_mode );
	bc_modes.append( new omac_mode );
	return bc_modes;
}

void 
coder_factory::print_registered_coders(ostream& out)
{
	list<coder_base*> coders = get_all_coders();

	coder_base* coder;
	forall(coder, coders) {
		out << string("%08x", coder->get_id()) << " " << coder->get_description();
		if (! coder->is_standalone_coder()) out << " (not standalone)";
		out << endl;
		delete coder;
	}
}

/// coder_factory_init //////////////////////////////////////////////////////////////////////

unsigned coder_factory_init::Count = 0;

typedef coder_base::byte    byte;
typedef coder_base::uint16  uint16;
typedef coder_base::uint32  uint32;
typedef coder_base::sint16  sint16;
typedef coder_base::sint32  sint32;
typedef coder_base::id_type id_type;

#define CF_Assert(COND) \
	leda_assert((COND), string("bad platform settings for compression module (see _coder_factory.cpp, line %d)", __LINE__));

coder_factory_init::coder_factory_init()
{
	if (Count++ == 0) {
		CoderFactory.init();

		// make some checks on this platform
		// int, streamsize: signed and bigger than a byte
		CF_Assert(int(-1) < 0 && sizeof(int) > sizeof(byte));
		CF_Assert(streamsize(-1) < 0 && sizeof(streamsize) > sizeof(byte));
		// byte: unsigned and exactly 8 bits
		byte b = 0; --b;
		CF_Assert(b > 0 && (unsigned long)(b) == 0xFFUL);
		// uint16: unsigned and at least 16 bits
		uint16 db = 0; --db;
#ifdef LEDA_CODER_EXACT_SIZE_INT16
		CF_Assert(db > 0 && sizeof(uint16) == 2);
#else
		CF_Assert(db > 0 && sizeof(uint16) >= 2);
#endif
		// uint32: unsigned and at least 32 bits
		uint32 qb = 0; --qb;
#ifdef LEDA_CODER_EXACT_SIZE_INT32
		CF_Assert(qb > 0 && sizeof(uint32) == 4);
#else
		CF_Assert(qb > 0 && sizeof(uint32) >= 4);
#endif
		// sint16: signed and at least 16 bits
		sint16 sdb = 0; --sdb;
#ifdef LEDA_CODER_EXACT_SIZE_INT16
		CF_Assert(sdb < 0 && sizeof(sint16) == 2);
#else
		CF_Assert(sdb < 0 && sizeof(sint16) >= 2);
#endif
		// sint32: signed and at least 32 bits
		sint32 sqb = 0; --sqb;
#ifdef LEDA_CODER_EXACT_SIZE_INT32
		CF_Assert(sqb < 0 && sizeof(sint32) == 4);
#else
		CF_Assert(sqb < 0 && sizeof(sint32) >= 4);
#endif
		// id_type: unsigned and at least 32 bits
		id_type id = 0; --id;
#ifdef LEDA_CODER_EXACT_SIZE_INT32
		CF_Assert(id > 0 && sizeof(id_type) == 4);
#else
		CF_Assert(id > 0 && sizeof(id_type) >= 4);
#endif
		// check byte order
		uint32 endian_check_uint = 1;
		byte* endian_check_char_ptr = (byte*) &endian_check_uint;
#ifdef LITTLE_ENDIAN_MACHINE
		CF_Assert(*endian_check_char_ptr == 1);
#else
		CF_Assert(*endian_check_char_ptr == 0);
#endif
	}
}

coder_factory_init::~coder_factory_init()
{
	if (--Count == 0) CoderFactory.destroy();
}

LEDA_END_NAMESPACE
