// compression module includes
#include <LEDA/coding/coder_pipe.h>

// core LEDA includes
#include <LEDA/system/error.h>

#if defined(LEDA_DEBUG)
#include <LEDA/system/assert.h>
#else
#undef assert
#define	assert(cond) ((void)0)
#endif

LEDA_BEGIN_NAMESPACE

typedef coder_types::uint32 uint32;

/// coder_pipe //////////////////////////////////////////////////////////////////////////////

coder_pipe::coder_pipe(unsigned num_coders)
: NumberOfCoders(num_coders), pCoder1(0), OwnCoder1(false), HasSeekAdapters(false)
{
	if (NumberOfCoders < 1) NumberOfCoders = 1;
	else if (NumberOfCoders > 254) NumberOfCoders = 254; // reserve 0xFF for the future
	init_adapter_array();
}

void 
coder_pipe::init_adapter_array()
{
	if (NumberOfCoders >= 2) {
		AdaptersForCoders = (new coder_strbuf_adapter*[NumberOfCoders-1]) - 2;
			// -> AdaptersForCoders[2] is the first allocated adapter!
		for (unsigned i = 2; i <= NumberOfCoders; ++i) 
			AdaptersForCoders[i] = new coder_strbuf_adapter;
	}
	else
		AdaptersForCoders = 0;

	HasSeekAdapters = false;
}

coder_pipe::~coder_pipe()
{
	set_state(s_finished); // clears errors
	set_coder1(0);
	if (AdaptersForCoders) {
		for (unsigned i = 2; i <= NumberOfCoders; ++i) delete AdaptersForCoders[i];
		delete[] (AdaptersForCoders+2); AdaptersForCoders = 0;
	}
}

void 
coder_pipe::do_reset(reset_type t)
{
	base::do_reset(t);

	if (t == r_seek) {
		assert(!HasSeekAdapters); // i.e. the pipe handles seeks globally
		// NOTE: the pipe must process the headers again after rewind (cf. init_encoding)
		ProcessHeaderInStreams = true;
		t = r_seek_without_rewind;
	}

	if (get_coder1()) get_coder1()->do_reset(t);
	for (unsigned i = 2; i <= NumberOfCoders; ++i)
		AdaptersForCoders[i]->do_reset(t);
}

string 
coder_pipe::get_description() const
{
	string result = base::get_description() + string('<');
	for (unsigned i = 1; i <= NumberOfCoders; ++i) {
		coder_base* coder_i = get_coder(i);
		result += coder_i ? coder_i->get_description() : string("unknown");
		if (i<NumberOfCoders) result += string(", ");
	}
	result += string('>');

	return result;
}

bool 
coder_pipe::is_single_pass_encoder() const
{
	if (! get_coder1()) {
		LEDA_EXCEPTION(1, "coder_pipe::is_single_pass_encoder: no coder1");
		return true;
	}
	return get_coder1()->is_single_pass_encoder();
}

bool 
coder_pipe::is_optimizing_seeks() const
{
	return get_coder1() ? get_coder1()->is_optimizing_seeks() : false;
}

bool 
coder_pipe::needs_seeks_during_coding() const
{
	for (unsigned i = 1; i <= NumberOfCoders; ++i) {
		coder_base* coder_i = get_coder(i);
		if (coder_i && coder_i->needs_seeks_during_coding()) return true;
	}

	return false;
}

bool 
coder_pipe::is_secrecy_sensitive() const
{
	for (unsigned i = 1; i <= NumberOfCoders; ++i) {
		coder_base* coder_i = get_coder(i);
		if (coder_i && coder_i->is_secrecy_sensitive()) return true;
	}

	return false;
}

void 
coder_pipe::ensure_suitable_adapters()
{
	// if we want to optimize seeks, our adapters must support it ...
	if (NumberOfCoders > 1 && !HasSeekAdapters) {
		// replace every adapter by one that supports seeks
		for (unsigned i = 2; i <= NumberOfCoders; ++i) {
			bool own_coder = AdaptersForCoders[i]->owns_coder();
			coder_base* coder = AdaptersForCoders[i]->release_ownership_of_coder();
			delete AdaptersForCoders[i];
			AdaptersForCoders[i] = new coder_strbuf_adapter_with_seek;
			AdaptersForCoders[i]->set_coder(coder, own_coder);
		}
		HasSeekAdapters = true;
	}
}

/*
	The pipe makes sure that in encode-mode all ids are written unencoded to the 
	target stream of the pipe; and in decode-mode they are read undecoded from the 
	source stream of the pipe.
*/
void 
coder_pipe::write_id_to_tgt_stream()
{
	base::write_id_to_tgt_stream();

	write_byte((byte) NumberOfCoders);

	for (unsigned i = 1; i <= NumberOfCoders; ++i) {
		coder_base* coder_i = get_coder(i);

		streambuf* save_tgt_stream = coder_i->change_tgt_stream_temporarily(get_tgt_stream());
		coder_i->write_id_to_tgt_stream();
		coder_i->change_tgt_stream_temporarily(save_tgt_stream);
	}
}

void 
coder_pipe::read_and_check_id_in_src_stream()
{
	base::read_and_check_id_in_src_stream();

	int c = read_byte();
	if (c == EOF || ((byte) c) != NumberOfCoders) {
		error("coder_pipe::read_and_check_id_in_src_stream: wrong number of coders");
		return;
	}

	for (unsigned i = 1; i <= NumberOfCoders; ++i) {
		coder_base* coder_i = get_coder(i);

		streambuf* save_src_stream = coder_i->change_src_stream_temporarily(get_src_stream());
		coder_i->read_and_check_id_in_src_stream();
		coder_i->change_src_stream_temporarily(save_src_stream);
	}
}

uint32 
coder_pipe::get_id_length(bool& is_exact) const
{
	uint32 len = base::get_id_length(is_exact) + 1 /* NumberOfCoders */;
	for (unsigned i = 1; i <= NumberOfCoders; ++i) {
		coder_base* coder_i = get_coder(i);
		if (! coder_i) { is_exact = false; len += coder_base::IdLength; continue; }
		bool coder_exact;
		len += coder_i->get_id_length(coder_exact);
		is_exact &= coder_exact;
	}
	return len;
}

void 
coder_pipe::finish_construction_from_stream(id_type id_in_stream, streambuf* stream, streambuf* additional_data)
{
	base::finish_construction_from_stream(id_in_stream, stream, additional_data);

	leda_assert(AdaptersForCoders == 0, "coder_pipe::finish_construction_from_stream");

	NumberOfCoders = (byte) stream->sbumpc();

	init_adapter_array();

	unsigned i;
	for (i = 1; i <= NumberOfCoders; ++i) {
		coder_base* coder = construct_from_stream(stream, additional_data);
		if (! coder) { 
			error("coder_pipe::finish_construction_from_stream: could not construct coder"); 
			return;
		}
		set_coder(i, coder, true);
	}
}


void 
coder_pipe::write_header_to_tgt_stream()
{
	if (!look_for_header_in_streams()) return;
		
	base::write_header_to_tgt_stream();

	for (unsigned i = 1; i <= NumberOfCoders; ++i) {
		coder_base* coder_i = get_coder(i);

		streambuf* save_tgt_stream = coder_i->change_tgt_stream_temporarily(get_tgt_stream());
		coder_i->ProcessHeaderInStreams = true;
		coder_i->write_header_to_tgt_stream();
		coder_i->change_tgt_stream_temporarily(save_tgt_stream);
	}
}

void 
coder_pipe::read_header_from_src_stream()
{
	if (!look_for_header_in_streams()) return;
		
	base::read_header_from_src_stream();

	for (unsigned i = 1; i <= NumberOfCoders; ++i) {
		coder_base* coder_i = get_coder(i);

		streambuf* save_src_stream = coder_i->change_src_stream_temporarily(get_src_stream());
		coder_i->ProcessHeaderInStreams = true;
		coder_i->read_header_from_src_stream();
		coder_i->change_src_stream_temporarily(save_src_stream);
	}
}

uint32 
coder_pipe::get_header_length(bool& is_exact) const
{
	uint32 len = base::get_header_length(is_exact);
	for (unsigned i = 1; i <= NumberOfCoders; ++i) {
		coder_base* coder_i = get_coder(i);
		if (! coder_i) { is_exact = false; continue; }
		bool coder_exact;
		len += coder_i->get_header_length(coder_exact);
		is_exact &= coder_exact;
	}
	return len;
}


void 
coder_pipe::init_encoding()
{
	bool only_single_pass_encoders = true;
	unsigned i;
	for (i = 1; i <= NumberOfCoders; ++i) {
		coder_base* coder_i = get_coder(i);
		if (! coder_i) {
			LEDA_EXCEPTION(1, "coder_pipe::init_encoding: coder(s) missing");
			error();
			return;
		}

		// the pipe takes over id and header processing
		coder_i->process_no_id_in_streams();
		coder_i->process_header_in_streams(look_for_header_in_streams());

		only_single_pass_encoders &= coder_i->is_single_pass_encoder();
	}

	if (!only_single_pass_encoders && is_secrecy_sensitive()) { 
		LEDA_EXCEPTION(1, "coder_pipe::init_encoding: do not use multi-pass coders with cryptography");
		// because data may be leaked in the temporary file (generated by coder_strbuf_adapter)
		return;
	}

	if (!HasSeekAdapters && needs_seeks_during_coding())
		ensure_suitable_adapters();

	get_coder1()->set_src_stream(get_src_stream(), false);
	if (NumberOfCoders == 1) {
		get_coder1()->set_tgt_stream(get_tgt_stream(), false);
	}
	else {
		get_coder1()->set_tgt_stream(AdaptersForCoders[2], false);
		for (i = 2; i < NumberOfCoders; ++i)
			AdaptersForCoders[i]->set_encoded_stream(AdaptersForCoders[i+1], false);
		AdaptersForCoders[NumberOfCoders]->set_encoded_stream(get_tgt_stream(), false);
	}

	/*
		We explicitly prevent calling coder_pipe::write_header/read_header 
		(for backward compatibility).
		Observe that coder_pipe itself no has header. (The number of coders and
		their ids are part of the id of this coder_pipe, because they are 
		needed during finish_construction_from_stream).
		Note that write_header is called (later!) for the different coders in 
		the pipe (when there respective init_encoding function is called)!
	*/
	process_no_header_in_streams(); // for backward compatibility
	base::init_encoding();
}

void 
coder_pipe::encode_stream()
{
	get_coder1()->do_encode();
	if (get_coder1()->failed()) { error(); encoding_finished(); }
	else if (get_coder1()->finished()) { encoding_finished(); }
}

void
coder_pipe::encoding_finished()
{
	for (unsigned i = 2; i <= NumberOfCoders; ++i) {
		AdaptersForCoders[i]->sputEOF();
		AdaptersForCoders[i]->pubsync();
		if (AdaptersForCoders[i]->failed()) error();
	}

	base::encoding_finished();
}

void 
coder_pipe::init_decoding()
{
	unsigned i;
	for (i = 1; i <= NumberOfCoders; ++i) {
		coder_base* coder_i = get_coder(i);
		if (! coder_i) {
			LEDA_EXCEPTION(1, "coder_pipe::init_decoding: coder(s) missing");
			error(); return;
		}

		// the pipe takes over id and header processing
		coder_i->process_no_id_in_streams();
		coder_i->process_header_in_streams(look_for_header_in_streams());
	}

	if (!HasSeekAdapters && (is_optimizing_seeks() || needs_seeks_during_coding()))
		ensure_suitable_adapters();

	if (NumberOfCoders == 1) {
		get_coder1()->set_src_stream(get_src_stream(), false);
	}
	else {
		AdaptersForCoders[NumberOfCoders]->set_encoded_stream(get_src_stream(), false);
		for (i = NumberOfCoders-1; i >= 2; --i)
			AdaptersForCoders[i]->set_encoded_stream(AdaptersForCoders[i+1], false);
		get_coder1()->set_src_stream(AdaptersForCoders[2], false);
	}
	get_coder1()->set_tgt_stream(get_tgt_stream(), false);

	process_no_header_in_streams(); // for backward compatibility (see init_encoding)
	base::init_decoding();
}

void 
coder_pipe::decode_stream()
{
	get_coder1()->do_decode();
	if (get_coder1()->failed()) { error(); decoding_finished(); }
	else if (get_coder1()->finished()) { decoding_finished(); }
}

void
coder_pipe::decoding_finished()
{
	for (unsigned i = 2; i <= NumberOfCoders; ++i) {
		if (AdaptersForCoders[i]->failed()) error();
	}

	base::decoding_finished();
}


streampos 
coder_pipe::decode_seek(streampos current_pos, streampos new_pos)
{
/*
There are two basic modes of operation:
1)	If we have no special seek support, then the pipe acts as big
	monolithic box, which handles the seeks globally on its own. 
	The coders in the pipe do not perform any seeks at all, they
	only decode sequentially!

2)	But if we have special seek support (HasSeekAdapters) and we
	have passed the initialization phase (i.e. decoding() == true),
	then the pipe is not involved at all in the seeks, but forwards
	them to coder1.
	Thus the individual coders in the pipe are responsible for the 
	seeks themselves.
*/
	if (!HasSeekAdapters || !decoding() || failed()) {
		return base::decode_seek(current_pos, new_pos);
	}

	new_pos = get_coder1()->decode_seek(current_pos, new_pos);

	// synchronize *this and coder1 ...
	if (get_coder1()->failed()) { error(); decoding_finished(); }
	else if (get_coder1()->finished()) { decoding_finished(); }
	else set_state(s_decoding); // make it running

	return new_pos;
}

streambuf* 
coder_pipe::change_src_stream_temporarily(streambuf* src_stream)
{
	if (encoding() && get_coder1()) get_coder1()->change_src_stream_temporarily(src_stream);
	if (decoding() && get_coder(NumberOfCoders)) get_coder(NumberOfCoders)->change_src_stream_temporarily(src_stream);
	return base::change_src_stream_temporarily(src_stream);
}

streambuf* 
coder_pipe::change_tgt_stream_temporarily(streambuf* tgt_stream)
{
	if (encoding() && get_coder(NumberOfCoders)) get_coder(NumberOfCoders)->change_tgt_stream_temporarily(tgt_stream);
	if (decoding() && get_coder1()) get_coder1()->change_tgt_stream_temporarily(tgt_stream);
	return base::change_tgt_stream_temporarily(tgt_stream);
}


void
coder_pipe::set_coder1(coder_base* coder1, bool own_coder)
{
	if (! running()) {
		if (OwnCoder1) delete pCoder1;
		pCoder1 = coder1; OwnCoder1 = own_coder;
	}
	else error("coder_pipe::set_coder1: wrong state");
}

coder_base*
coder_pipe::get_coder(unsigned coder_idx) const
{
	if (coder_idx < 1 || coder_idx > NumberOfCoders) {
		LEDA_EXCEPTION(1, "coder_pipe::get_coder: idx out of range");
		return 0;
	}

	return coder_idx == 1 ? get_coder1() : AdaptersForCoders[coder_idx]->get_coder();
}

void
coder_pipe::set_coder(unsigned coder_idx, coder_base* coder, bool own_coder)
{
	if (coder_idx < 1 || coder_idx > NumberOfCoders) {
		LEDA_EXCEPTION(1, "coder_pipe::set_coder: idx out of range");
		return;
	}

	if (coder_idx == 1)
		set_coder1(coder, own_coder);
	else {
		AdaptersForCoders[coder_idx]->set_coder(coder, own_coder);
	}
}

LEDA_END_NAMESPACE
