// compression module includes
#include <LEDA/coding/deflate.h>
#include <LEDA/coding/memory_streambuf.h>
#include "../zlib/zlib.h"
#include "../zlib/zutil.h" // for DEF_MEM_LEVEL

#define CHECK_ZLIB_ERR(err, msg) \
	if (err != Z_OK) { error(msg); return; }

/*
	The following files from zlib are needed:
	- adler32.c
	- compress.c
	- deflate.c
	- inffast.c
	- inflate.c
	- inftrees.c
	- trees.c
	- zutil.c

	From the zlib-FAQ: 
	21. Is zlib thread-safe? 
	Yes. However any library routines that zlib uses and any application- provided memory 
	allocation routines must also be thread-safe. zlib's gz* functions use stdio library 
	routines, and most of zlib's functions use the library memory allocation routines by default.
 */

LEDA_BEGIN_NAMESPACE

/// DeflateCoder ////////////////////////////////////////////////////////////////////////////

DeflateCoder::DeflateCoder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams,
						   uint32 enc_buffer_sz, uint32 dec_buffer_sz)
: base(src_stream, tgt_stream, own_streams), 
  CompressionLevel(Z_DEFAULT_COMPRESSION), Log2OfWindowSize(MAX_WBITS), 
  MemLevel(DEF_MEM_LEVEL), Strategy(Z_DEFAULT_STRATEGY), 
  EncBufferSize(0), DecBufferSize(0), pEncBuffer(0), pDecBuffer(0)
{
	p_zlib_stream = new z_stream;
	set_enc_buffer_size(enc_buffer_sz);
	set_dec_buffer_size(dec_buffer_sz);
}

DeflateCoder::DeflateCoder(const char* src_file_name, const char* tgt_file_name,
						   uint32 enc_buffer_sz, uint32 dec_buffer_sz)
: base(src_file_name, tgt_file_name), 
  CompressionLevel(Z_DEFAULT_COMPRESSION), Log2OfWindowSize(MAX_WBITS), 
  MemLevel(DEF_MEM_LEVEL), Strategy(Z_DEFAULT_STRATEGY), 
  EncBufferSize(0), DecBufferSize(0), pEncBuffer(0), pDecBuffer(0)
{
	p_zlib_stream = new z_stream;
	set_enc_buffer_size(enc_buffer_sz);
	set_dec_buffer_size(dec_buffer_sz);
}

DeflateCoder::~DeflateCoder()
{
	delete ((z_stream*) p_zlib_stream);
	delete[] pEncBuffer; pEncBuffer = 0;
	delete[] pDecBuffer; pDecBuffer = 0;
}

void 
DeflateCoder::do_reset(reset_type t)
{
	base::do_reset(t);
	if (t == r_hard) {
		set_compression_level(Z_DEFAULT_COMPRESSION);
		set_window_size(MAX_WBITS);
		set_memory_level(DEF_MEM_LEVEL);
		set_strategy(Z_DEFAULT_STRATEGY);

		set_enc_buffer_size(DefaultBufferSize);
		set_dec_buffer_size(DefaultBufferSize);
	}
}

string 
DeflateCoder::get_description() const
{
	string descr = base::get_description();
	if (CompressionLevel != Z_DEFAULT_COMPRESSION || Log2OfWindowSize != MAX_WBITS
	 || MemLevel != DEF_MEM_LEVEL || Strategy != Z_DEFAULT_STRATEGY)
	{
		descr += string("[%d,%d,%d,%d]", CompressionLevel, Log2OfWindowSize, MemLevel, Strategy);
	}
	return descr;
}

void
DeflateCoder::set_compression_level(int level)
{
	if (can_change_parameters()) {
		if (level < 0 || level > 9)
			CompressionLevel = Z_DEFAULT_COMPRESSION;
		else
			CompressionLevel = level;
	}
	else error("DeflateCoder::set_compression_level: wrong state");
}

void
DeflateCoder::set_window_size(int log2_of_sz)
{
	if (can_change_parameters()) {
		if (log2_of_sz < 8 || log2_of_sz > 15)
			Log2OfWindowSize = MAX_WBITS;
		else
			Log2OfWindowSize = log2_of_sz;
	}
	else error("DeflateCoder::set_window_size: wrong state");
}

void
DeflateCoder::set_memory_level(int level)
{
	if (can_change_parameters()) {
		if (level < 1 || level > 9)
			MemLevel = DEF_MEM_LEVEL;
		else
			MemLevel = level;
	}
	else error("DeflateCoder::set_memory_level: wrong state");
}

void
DeflateCoder::set_strategy(int strategy)
{
	if (can_change_parameters()) {
		switch (strategy) {
		case Z_FILTERED:
		case Z_HUFFMAN_ONLY:
		case Z_RLE:
			Strategy = strategy;
			break;
		default:
			Strategy = Z_DEFAULT_STRATEGY;
			break;
		}
	}
	else error("DeflateCoder::set_strategy: wrong state");
}

void 
DeflateCoder::get_memory_consumption(uint32& mem_for_enc, uint32& mem_for_dec)
{
	// see zconf.h for details ...
	mem_for_enc = (1 << (Log2OfWindowSize+2)) + (1 << (MemLevel+9))
	            + EncBufferSize + DecBufferSize;
	mem_for_dec = (1 << Log2OfWindowSize)
	            + EncBufferSize + DecBufferSize;
}

void
DeflateCoder::set_enc_buffer_size(uint32 buffer_size)
{
	if (can_change_parameters()) {
		int new_size = buffer_size > 0 ? buffer_size : DefaultBufferSize;
		if (new_size != EncBufferSize) {
			delete[] pEncBuffer;
			EncBufferSize = new_size;
			pEncBuffer = new byte[EncBufferSize];
		}
	}
	else error("DeflateCoder::set_enc_buffer_size: wrong state");
}

void
DeflateCoder::set_dec_buffer_size(uint32 buffer_size)
{
	if (can_change_parameters()) {
		int new_size = buffer_size > 0 ? buffer_size : DefaultBufferSize;
		if (new_size != DecBufferSize) {
			delete[] pDecBuffer;
			DecBufferSize = new_size;
			pDecBuffer = new byte[DecBufferSize];
		}
	}
	else error("DeflateCoder::set_dec_buffer_size: wrong state");
}

void 
DeflateCoder::init_encoding()
{
	base::init_encoding();

	z_stream* zlib_stream = (z_stream*) p_zlib_stream;

	zlib_stream->zalloc = (alloc_func) 0;
    zlib_stream->zfree = (free_func) 0;
    zlib_stream->opaque = (voidpf) 0;

    int err = deflateInit2(zlib_stream, CompressionLevel, Z_DEFLATED, 
						   Log2OfWindowSize, MemLevel, Strategy);
	CHECK_ZLIB_ERR(err, "deflateInit");

	zlib_stream->next_out = pEncBuffer; zlib_stream->avail_out = EncBufferSize;
	zlib_stream->avail_in = 0; pCurrent = pDecBuffer;
	State = Z_NO_FLUSH;
}

void 
DeflateCoder::encode_stream()
{
	z_stream* zlib_stream = (z_stream*) p_zlib_stream;

	for(;;) {
		if (zlib_stream->next_out != pEncBuffer) {
			// we have pending output
			int bytes_to_write = int(zlib_stream->next_out - pEncBuffer);
			if (write_bytes(pEncBuffer, bytes_to_write) != bytes_to_write) { error(); break; }
			zlib_stream->next_out = pEncBuffer; zlib_stream->avail_out = EncBufferSize;
		}

		if (! zlib_stream->avail_in) {
			// all input has been used
			if (State == Z_NO_FLUSH) {
				byte* stop = pDecBuffer + DecBufferSize;
				int c = EOF; // just to avoid a warning
				while ( pCurrent != stop && (c = read_byte()) >= 0 && c != EOF ) {
					*(pCurrent++) = (byte) c;
				}
				if (c == EOF) State = Z_FINISH;
				else if (c < 0) return; // StreamBlocked
				zlib_stream->avail_in = int(pCurrent - pDecBuffer);
				zlib_stream->next_in = pCurrent = pDecBuffer;
			}
			else if (State == Z_STREAM_END) {
				encoding_finished();
				return; 
			}
		}

        int res = deflate(zlib_stream, State);
		if (res != Z_OK) {
			deflateEnd(zlib_stream);
			if (res != Z_STREAM_END) { error("deflate"); return; }
			State = Z_STREAM_END;
		}
	}
}


void 
DeflateCoder::init_decoding()
{
    base::init_decoding();

    z_stream* zlib_stream = (z_stream*) p_zlib_stream;

    zlib_stream->zalloc = (alloc_func) 0;
    zlib_stream->zfree = (free_func) 0;
    zlib_stream->opaque = (voidpf) 0;

    int err = inflateInit(zlib_stream);
    CHECK_ZLIB_ERR(err, "inflateInit");

    zlib_stream->next_out = pCurrent = pDecBuffer;
    zlib_stream->avail_out = DecBufferSize;
    zlib_stream->avail_in = 0;
    State = Z_NO_FLUSH;
}

void 
DeflateCoder::decode_stream()
{
	z_stream* zlib_stream = (z_stream*) p_zlib_stream;

	for(;;) {
		if (zlib_stream->next_out != pDecBuffer) {
			// we have pending output
			int bytes_to_write = (int) (zlib_stream->next_out - pCurrent);
			int bytes_written = write_bytes(pCurrent, bytes_to_write);
			if (bytes_written != bytes_to_write) {
				// we have a problem: either an error or StreamBlocked
				if (bytes_written >= 0) error();
				pCurrent += StreamBlocked - bytes_written;
				return;
			}
			// old byte-wise output:
			//byte* stop = zlib_stream->next_out;
			//while (pCurrent != stop) if (write_byte(*(pCurrent++)) < 0) return;

			zlib_stream->next_out = pCurrent = pDecBuffer; zlib_stream->avail_out = DecBufferSize;
		}

		if (! zlib_stream->avail_in) {
			if (State == Z_NO_FLUSH) {
				int bytes_read = read_bytes(pEncBuffer, EncBufferSize);
				zlib_stream->next_in = pEncBuffer;
				zlib_stream->avail_in = bytes_read;
				if (bytes_read < EncBufferSize) 
					State = Z_FINISH;
			}
			else if (State == Z_STREAM_END) {
				decoding_finished();
				return;
			}
		}

        int res = inflate(zlib_stream, Z_NO_FLUSH);
		if (res != Z_OK) {
			inflateEnd(zlib_stream);
			if (res != Z_STREAM_END) { error("inflate"); return; }
			State = Z_STREAM_END;
		}
	}
}

void 
DeflateCoder::encode_memory(memory_streambuf& in, memory_streambuf& out)
{
	z_stream* zlib_stream = (z_stream*) p_zlib_stream;

	zlib_stream->zalloc = (alloc_func) 0;
    zlib_stream->zfree = (free_func) 0;
    zlib_stream->opaque = (voidpf) 0;

	zlib_stream->next_in = (Bytef*) in.get_remaining_in_buffer(); 
	zlib_stream->avail_in = in.in_avail();
	zlib_stream->next_out = (Bytef*) out.get_remaining_out_buffer(); 
	zlib_stream->avail_out = out.out_avail();

    int err = deflateInit2(zlib_stream, CompressionLevel, Z_DEFLATED, 
						   Log2OfWindowSize, MemLevel, Strategy);
	CHECK_ZLIB_ERR(err, "deflateInit");

	err = deflate(zlib_stream, Z_FINISH);
	if (err != Z_STREAM_END) error("deflate");

	// ensure that the put pointer of "out" points behind the last output byte
	out.pubseekoff(zlib_stream->total_out, ios::cur, ios::out);

    err = deflateEnd(zlib_stream);
	CHECK_ZLIB_ERR(err, "deflateEnd");
}

void 
DeflateCoder::decode_memory(memory_streambuf& in, memory_streambuf& out)
{
	z_stream* zlib_stream = (z_stream*) p_zlib_stream;

	zlib_stream->zalloc = (alloc_func) 0;
    zlib_stream->zfree = (free_func) 0;
    zlib_stream->opaque = (voidpf) 0;

	zlib_stream->next_in = (Bytef*) in.get_remaining_in_buffer(); 
	zlib_stream->avail_in = in.in_avail();
	zlib_stream->next_out = (Bytef*) out.get_remaining_out_buffer(); 
	zlib_stream->avail_out = out.out_avail();

	// workaround which is needed if the original stream was empty (and next_out is NULL)
	Bytef dummy;
	if (zlib_stream->avail_out == 0) zlib_stream->next_out = &dummy;

    int err = inflateInit(zlib_stream);
	CHECK_ZLIB_ERR(err, "inflateInit");

	err = inflate(zlib_stream, Z_FINISH);
	if (err != Z_STREAM_END) error("inflate");

	// ensure that the put pointer of "out" points behind the last output byte
	out.pubseekoff(zlib_stream->total_out, ios::cur, ios::out);

    err = inflateEnd(zlib_stream);
	CHECK_ZLIB_ERR(err, "inflateEnd");
}

const char* 
DeflateCoder::get_zlib_version()
{
	return zlibVersion();
}

LEDA_END_NAMESPACE
