// compression module includes
#include <LEDA/coding/huffman.h>

#if defined(LEDA_DEBUG)
#include <LEDA/system/assert.h>
#else
#undef assert
#define	assert(cond) ((void)0)
#endif

#define Left 0
#define Right 1

LEDA_BEGIN_NAMESPACE

/// HuffmanCoder ////////////////////////////////////////////////////////////////////////////

HuffmanCoder::HuffmanCoder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams), HuffTable(0)
{}

HuffmanCoder::HuffmanCoder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name), HuffTable(0)
{}

HuffmanCoder::~HuffmanCoder()
{
	delete[] HuffTable; HuffTable = 0;
}

void 
HuffmanCoder::init_encoding()
{
	base::init_encoding();

	// initialization
	delete[] HuffTable; HuffTable = 0;
	memset(HuffTree, 0, sizeof(HuffTree));
	for (uint32 i = 0; i < MaxSymbols; i++)
	{
		HuffTree[i].child[Left] = NIL;
		HuffTree[i].child[Right] = NIL;
	}
	ByteCount = 0;
}

void 
HuffmanCoder::encode_stream()
{
	int c;

	// huff_scan: count occurrences
	while ((c = read_byte()) >= 0 && c != EOF) {
		HuffTree[c].count++; // count # of occurences
		ByteCount++; // size of input file
	}
	if (c != EOF) return;

	rewind_src_stream();

	write_uint32(ByteCount);

	if (ByteCount == 0 || failed()) { encoding_finished(); return; }

	build_tree();

	// write tree
	write_uint16(NumActiveLeaves);

	for (uint16 i = 0; i <= Root; ++i) {
		if (HuffTree[i].count) {
			write_uint16(i);
			write_uint16(HuffTree[i].child[0]);
			write_uint16(HuffTree[i].child[1]);
		}
	}

	// huff_compress: compression
	int outchar = 0;
	int outbit = 8;

	while((c = read_byte()) != EOF) {
		byte* s = HuffTree[c].hcode;
		int count = HuffTree[c].bitcount;
		for (int k = 0; k < count; k++) {
			int _byte = k/8;
			int _bit = k - _byte * 8;
			int set = (s[_byte] & (1<<(7-_bit))) ? 1 : 0;

			outbit--;
			if (set) outchar |= 1 << outbit;

			if (outbit == 0) {
				write_byte((byte) outchar);
				outchar = 0;
				outbit = 8;
			}
		}
	}

	// last byte
	if (outbit != 8) {
		write_byte((byte) outchar);
	}

	delete[] HuffTable; HuffTable = 0;
	encoding_finished();
}

void 
HuffmanCoder::build_tree()
{
	LinkType i;

	NumActiveLeaves = 0;
	for (i = 0; i < MaxSymbols; ++i) {
		if (HuffTree[i].count) ++NumActiveLeaves;
	}

	// we need at least two active nodes
	if (NumActiveLeaves == 1) {
		if (! HuffTree[0].count) HuffTree[0].count++;
		else                     HuffTree[1].count++;
		++NumActiveLeaves;
	}

	Root = MaxSymbols + NumActiveLeaves - 2;

	// build internal nodes

	for (i = MaxSymbols; i <= Root; i++) {
	// TODO: das folgende kann man sicher mit priority queue beschleunigen
		LinkType lo1 = 0, lo2 = 0;

		LinkType k;
		for (k = 0; k < i; k++) {
			if (HuffTree[k].count && !HuffTree[k].parent) {
				lo1 = k; 
				k++;
				break;
			}
		}
		for (; k < i; k++) {
			if (HuffTree[k].count && !HuffTree[k].parent) {
				lo2 = k;
				k++;
				break;
			}
		}

		for (; k < i; k++) {
			if (HuffTree[k].count && !HuffTree[k].parent) {
				if (HuffTree[k].count < HuffTree[lo1].count)
					lo1 = k;
				else if (HuffTree[k].count < HuffTree[lo2].count)
					lo2 = k;
			}
		}

		// building the node
		HuffTree[i].count = HuffTree[lo1].count + HuffTree[lo2].count;
		HuffTree[i].child[Left] = lo1;
		HuffTree[i].child[Right] = lo2;
		HuffTree[lo1].parent = HuffTree[lo2].parent = i;
	}

	// building the bit codes
	delete[] HuffTable; HuffTable = 0;
	byte* ptr_in_table = 0;
	int bytes_needed_for_table = walk_tree(Root, 0, ptr_in_table);

	HuffTable = new byte[bytes_needed_for_table];
	memset(HuffTable, 0, bytes_needed_for_table);
	ptr_in_table = HuffTable;
	walk_tree(Root, 0, ptr_in_table);
}

int
HuffmanCoder::walk_tree(LinkType k, int depth, byte*& HuffTablePtr)
{
	int bytes_needed_for_table = 0;
	LinkType w, trace;

	for (int dir = Left; dir <= Right; ++dir) {
		if (HuffTree[k].child[dir] == NIL) {
			if (dir == Left) {
				bytes_needed_for_table += (depth+7)/8;
				if (HuffTablePtr) {
					HuffTree[k].bitcount = depth;
					HuffTree[k].hcode = HuffTablePtr;
					int bitno = depth-1;
					for (trace = k, w = HuffTree[k].parent; w; trace = w, w = HuffTree[w].parent)
					{
						int set = (HuffTree[w].child[Left] == trace) ? Left : Right;

						if (set) {
							int _byte = bitno/8;
							int _bit = bitno-_byte*8;
							HuffTablePtr[_byte] |= 1<<(7-_bit);
						}
						bitno--;
					}
					HuffTablePtr += (depth+7)/8;
				}
			}
		} // corresp. to if (HuffTree[k].child[dir] == NIL)
		else
			bytes_needed_for_table += walk_tree(HuffTree[k].child[dir], depth+1, HuffTablePtr);
	}

	return bytes_needed_for_table;
}


void 
HuffmanCoder::init_decoding()
{
	base::init_decoding();

	read_uint32(ByteCount);

	if (ByteCount == 0) return;

	// read tree
	read_uint16(NumActiveLeaves);

	Root = MaxSymbols + NumActiveLeaves - 2;

	memset(DecHuffTree, 0, sizeof(DecHuffTree));
	for (uint16 i = 0; i < NumActiveLeaves + (Root - MaxSymbols + 1); i++) {
		uint16 node_idx;
		read_uint16(node_idx);
		read_uint16(DecHuffTree[node_idx].child[0]);
		read_uint16(DecHuffTree[node_idx].child[1]);
	}

	BitNum = 0;
}

void 
HuffmanCoder::decode_stream()
{
	while (ByteCount > 0)
	{
		uint32 i = Root;
		for( ;; )
		{
			//need new input byte?
			if(--BitNum < 0)
			{
				if ((LastByte = read_byte()) == EOF) {
					error("HuffmanCoder::decode_stream: unexpected EOF");
					return;
				}
				BitNum = 7;
			}
			
			int bit = (LastByte & (1 << BitNum)) == 0 ? 0 : 1;

			i = DecHuffTree[i].child[bit];

			if(DecHuffTree[i].child[Left] == NIL) {
				--ByteCount;
				if (write_byte((byte) i) < 0) return;
				break;
			}
		}
	}

	decoding_finished();
}

/// floating_huff_tree (for AdaptiveHuffmanCoder) ///////////////////////////////////////////

floating_huff_tree::floating_huff_tree()
{
	initialize();
}

floating_huff_tree::~floating_huff_tree()
{
	destroy();
}

void
floating_huff_tree::destroy()
{
	block_t* cur_block = block(LastLeaf);
	if (cur_block == 0) return;

	cur_block->PrevBlk->NextBlk = 0; // break the cycle
	do {
		block_t* old_block = cur_block;
		cur_block = cur_block->NextBlk;
		delete old_block;
	} while (cur_block);

	ZeroLeaf = 0; block(ZeroLeaf) = 0;
}

void
floating_huff_tree::initialize()
{
	byte b = 0;
	do {
		SymbolToLeafIdx[b] = b; LeafIdxToSymbol[b] = b;
	} while (++b);

	ZeroLeaf = LastLeaf;
	block_t* blk_zero_leaf = new block_t;
	blk_zero_leaf->Weight         = 0;
	blk_zero_leaf->First          = ZeroLeaf;
	blk_zero_leaf->Last           = ZeroLeaf;
	blk_zero_leaf->ParentOfFirst  = NilIdx;
	blk_zero_leaf->RtChildOfFirst = NilIdx;
	blk_zero_leaf->ParityOfFirst  = p_Left;
	blk_zero_leaf->PrevBlk        = blk_zero_leaf;
	blk_zero_leaf->NextBlk        = blk_zero_leaf;
	block(ZeroLeaf) = blk_zero_leaf;
}

void 
floating_huff_tree::update(byte symbol)
{
	// FindNode
	node_idx q = corresp_leaf(symbol); node_idx leaf_to_increment = NilIdx;
	if (q > ZeroLeaf) { // tree already contains symbol
		// interchange q with the first node in its block
		block_t* blk = block(q);
		interchange_leaves(q, blk->First);
		q = blk->First; // now q is the leader of the block
		if (q == ZeroLeaf+1 && ZeroLeaf != NilIdx) {
			leaf_to_increment = q; q = blk->ParentOfFirst;
		}
		else slide_and_increment_leaf(q);
	}
	else {
		// a zero weight becomes positive
		interchange_leaves(q, ZeroLeaf);
		q = ZeroLeaf; // the new leaf is the old zero leaf
		if (--ZeroLeaf != NilIdx) {
			// split 0-leaf into an internal node with two children
			// new 0-leaf is stored in ZeroLeaf
			// old 0-leaf is stored in q (and equal to ZeroLeaf + 1)
			// common parent of old and new 0-leaf is ZeroLeaf + NumSymbols
			node_idx zero_inner = ZeroLeaf + NumSymbols;

			// update block containing the two zero weight leaves
			block_t* blk_zero_leaves = block(q); // block of old 0-leaf
			block(ZeroLeaf) = blk_zero_leaves;
			blk_zero_leaves->Last = ZeroLeaf;
			node_idx old_parent = blk_zero_leaves->ParentOfFirst;
			blk_zero_leaves->ParentOfFirst = zero_inner; blk_zero_leaves->ParityOfFirst = p_Right;

			// create a new internal block of zero weight for the parent
			block_t* blk_zero_inner = new block_t;
			blk_zero_inner->Weight = 0;
			blk_zero_inner->ParentOfFirst  = old_parent;
			blk_zero_inner->First          = zero_inner;
			blk_zero_inner->Last           = zero_inner;
			blk_zero_inner->ParentOfFirst  = old_parent;
			blk_zero_inner->RtChildOfFirst = q;
			blk_zero_inner->ParityOfFirst  = p_Left;
			blk_zero_inner->PrevBlk = blk_zero_leaves;
			blk_zero_inner->NextBlk = blk_zero_leaves->NextBlk;
			blk_zero_leaves->NextBlk->PrevBlk = blk_zero_inner;
			blk_zero_leaves->NextBlk = blk_zero_inner;
			block(zero_inner) = blk_zero_inner;

			// NOTE: The former 0-leaf was the LEFT child of old_parent
			//       so old_parent's RIGHT child remains the same!

			leaf_to_increment = q; q = zero_inner;
		}
		else slide_and_increment_leaf(q);
	}

	slide_and_increment_inner_nodes(q);

	if (leaf_to_increment != NilIdx)
		slide_and_increment_leaf(leaf_to_increment);
}

floating_huff_tree::node_idx 
floating_huff_tree::child(node_idx idx, parity_t which) const
// The remark in the paper that the else-part is never executed
// for a well-formed tree is wrong!
{
	block_t* blk = block(idx);
	node_idx delta = 2*(blk->First - idx) + 1 - which;
	node_idx right = blk->RtChildOfFirst;
	node_idx gap = right - block(right)->Last;
	if (delta <= gap) return right - delta;
	else {
		delta -= gap + 1;
		right = block(right)->PrevBlk->First;
		gap = right - block(right)->Last;
		if (delta <= gap) return right - delta;
		else return block(right)->PrevBlk->First - delta + gap + 1;
	}
}

void
floating_huff_tree::slide_and_increment_inner_nodes(node_idx q)
{
	assert(is_inner_node(q));

	while (q != NilIdx) {
		// q is currently the first node in its block
		block_t* blk_q = block(q); block_t* next_blk_q = blk_q->NextBlk;
		node_idx par_q = blk_q->ParentOfFirst; 
		node_idx old_parent = par_q; parity_t old_parity = blk_q->ParityOfFirst;

		bool slide = false;
		if ( is_leaf(next_blk_q->First) && blk_q->Weight+1 ==next_blk_q->Weight ) {
			// slide q over the next block
			slide = true;

			old_parent = next_blk_q->ParentOfFirst; old_parity = next_blk_q->ParityOfFirst;

			// adjust child pointers for next higher level in the tree
			if (par_q != NilIdx) {
				block_t* blk_par_q = block(par_q);
				if (blk_par_q->RtChildOfFirst == q) blk_par_q->RtChildOfFirst = next_blk_q->Last;
				else if (blk_par_q->RtChildOfFirst == next_blk_q->First) blk_par_q->RtChildOfFirst = q;
				else ++(blk_par_q->RtChildOfFirst);

				if (par_q != LastNode && block(par_q + 1) != blk_par_q) {
					block_t* blk_par_q_succ = block(par_q + 1);
					if (blk_par_q_succ->RtChildOfFirst == next_blk_q->First)
						blk_par_q_succ->RtChildOfFirst = q;
					else if (block(blk_par_q_succ->RtChildOfFirst) == next_blk_q)
						++(blk_par_q_succ->RtChildOfFirst);
				}
			}

			// adjust parent pointers for next_blk_q
			next_blk_q->ParentOfFirst -= 1 - next_blk_q->ParityOfFirst;
			next_blk_q->ParityOfFirst = parity_t(1 - next_blk_q->ParityOfFirst);
			next_blk_q = next_blk_q->NextBlk;
		}

		// increment phase
		if ( is_inner_node(next_blk_q->First) && blk_q->Weight+1 == next_blk_q->Weight )
			// q and the nodes in the next block are of the same kind 
			// and their weight is one higher than that of q
		{
			// merge q into next_blk_q
			block(q) = next_blk_q; next_blk_q->Last = q;

			// remove q from its old block
			if (blk_q->Last == q) {
				// q was the only node in this block, so it disappears ...
				blk_q->PrevBlk->NextBlk = blk_q->NextBlk;
				blk_q->NextBlk->PrevBlk = blk_q->PrevBlk;
				delete blk_q;
			}
			else {
				blk_q->RtChildOfFirst = child(q-1, p_Right);
				if (blk_q->ParityOfFirst == p_Left) --(blk_q->ParentOfFirst);
				blk_q->First = q-1;
				blk_q->ParityOfFirst = parity_t(1 - blk_q->ParityOfFirst);
			}
		}
		else if (blk_q->Last == q) {
			// q is the only node in its block
			if (slide) { // slide q's block forward in the block list
				// remove blk_q from block list
				blk_q->PrevBlk->NextBlk = blk_q->NextBlk;
				blk_q->NextBlk->PrevBlk = blk_q->PrevBlk;
				// reinsert it before next_blk_q
				blk_q->PrevBlk = next_blk_q->PrevBlk; blk_q->NextBlk = next_blk_q;
				next_blk_q->PrevBlk = blk_q; blk_q->PrevBlk->NextBlk = blk_q;
				// adjust parent info
				blk_q->ParentOfFirst = old_parent;
				blk_q->ParityOfFirst = old_parity;
			}
			++(blk_q->Weight);
		}
		else {
			// a new block is created for q
			block_t* new_blk_q = new block_t;
			new_blk_q->First = new_blk_q->Last = q; block(q) = new_blk_q;
			new_blk_q->RtChildOfFirst = blk_q->RtChildOfFirst;
			blk_q->RtChildOfFirst = child(q-1, p_Right);
			if (new_blk_q->RtChildOfFirst == q-1) blk_q->ParentOfFirst = q;
			else if (blk_q->ParityOfFirst == p_Left) --(blk_q->ParentOfFirst);
			blk_q->First = q-1;
			blk_q->ParityOfFirst = parity_t(1 - blk_q->ParityOfFirst);

			// insert new_blk_q into block list before next_blk_q
			new_blk_q->PrevBlk = next_blk_q->PrevBlk; new_blk_q->NextBlk = next_blk_q; 
			next_blk_q->PrevBlk = new_blk_q; new_blk_q->PrevBlk->NextBlk = new_blk_q;

			new_blk_q->Weight = blk_q->Weight + 1;
			new_blk_q->ParentOfFirst = old_parent; new_blk_q->ParityOfFirst = old_parity;
		}

		// move q one level higher in the tree
		q = par_q;
	}
}

void
floating_huff_tree::slide_and_increment_leaf(node_idx& q)
{
	assert( is_leaf(q) );

	// q is currently the first node in its block
	block_t* blk_q = block(q); block_t* next_blk_q = blk_q->NextBlk;
	node_idx par_q = blk_q->ParentOfFirst; 
	node_idx old_parent = par_q; parity_t old_parity = blk_q->ParityOfFirst;

	bool slide = false;
	if ( is_inner_node(next_blk_q->First) && blk_q->Weight == next_blk_q->Weight ) {
		// slide q over the next block
		slide = true;

		old_parent = next_blk_q->ParentOfFirst; old_parity = next_blk_q->ParityOfFirst;

		// adjust child pointers for next higher level in the tree
		if (par_q != NilIdx) {
			block_t* blk_par_q = block(par_q);
			if (blk_par_q->RtChildOfFirst == q) blk_par_q->RtChildOfFirst = next_blk_q->Last;
			else if (blk_par_q->RtChildOfFirst == next_blk_q->First) blk_par_q->RtChildOfFirst = q;
			else ++(blk_par_q->RtChildOfFirst);

			if (par_q != LastNode && block(par_q + 1) != blk_par_q) {
				block_t* blk_par_q_succ = block(par_q + 1);
				if (blk_par_q_succ->RtChildOfFirst == next_blk_q->First)
					blk_par_q_succ->RtChildOfFirst = q;
				else if (block(blk_par_q_succ->RtChildOfFirst) == next_blk_q)
					++(blk_par_q_succ->RtChildOfFirst);
			}
		}

		// adjust parent pointers for next_blk_q
		next_blk_q->ParentOfFirst -= 1 - next_blk_q->ParityOfFirst;
		next_blk_q->ParityOfFirst = parity_t(1 - next_blk_q->ParityOfFirst);
		next_blk_q = next_blk_q->NextBlk;
	}

	// increment phase
	if ( is_leaf(next_blk_q->First)	&& blk_q->Weight+1 == next_blk_q->Weight )
		// q and the nodes in the next block are of the same kind 
		// and their weight is one higher than that of q
	{
		// merge q into next_blk_q
		block(q) = next_blk_q; next_blk_q->Last = q;

		// remove q from its old block
		if (blk_q->Last == q) {
			// q was the only node in this block, so it disappears ...
			blk_q->PrevBlk->NextBlk = blk_q->NextBlk;
			blk_q->NextBlk->PrevBlk = blk_q->PrevBlk;
			delete blk_q;
		}
		else {
			if (blk_q->ParityOfFirst == p_Left) --(blk_q->ParentOfFirst);
			blk_q->First = q-1;
			blk_q->ParityOfFirst = parity_t(1 - blk_q->ParityOfFirst);
		}
	}
	else if (blk_q->Last == q) {
		// q is the only node in its block
		if (slide) { // slide q's block forward in the block list
			// remove blk_q from block list
			blk_q->PrevBlk->NextBlk = blk_q->NextBlk;
			blk_q->NextBlk->PrevBlk = blk_q->PrevBlk;
			// reinsert it before next_blk_q
			blk_q->PrevBlk = next_blk_q->PrevBlk; blk_q->NextBlk = next_blk_q;
			next_blk_q->PrevBlk = blk_q; blk_q->PrevBlk->NextBlk = blk_q;
			// adjust parent info
			blk_q->ParentOfFirst = old_parent;
			blk_q->ParityOfFirst = old_parity;
		}
		++(blk_q->Weight);
	}
	else {
		// a new block is created for q
		block_t* new_blk_q = new block_t;
		new_blk_q->First = new_blk_q->Last = q; block(q) = new_blk_q;
		if (blk_q->ParityOfFirst == p_Left) --(blk_q->ParentOfFirst);
		blk_q->First = q-1;
		blk_q->ParityOfFirst = parity_t(1 - blk_q->ParityOfFirst);

		// insert new_blk_q into block list before next_blk_q
		new_blk_q->PrevBlk = next_blk_q->PrevBlk; new_blk_q->NextBlk = next_blk_q; 
		next_blk_q->PrevBlk = new_blk_q; new_blk_q->PrevBlk->NextBlk = new_blk_q;

		new_blk_q->Weight = blk_q->Weight + 1;
		new_blk_q->ParentOfFirst = old_parent; new_blk_q->ParityOfFirst = old_parity;
	}

	// move q one level higher in the tree
	q = old_parent;
}

void
floating_huff_tree::print()
{
	print_subtree(root(), 0);
	print_block_list();

	cout << endl;
}

void
floating_huff_tree::print_subtree(node_idx idx, int level)
{
	for (int l = 0; l < level; ++l) cout << "  ";
	cout << "idx: " << idx << " w: " << weight(idx);
	if (is_leaf(idx)) {
		if (is_zero_leaf(idx)) cout << " 0-leaf";
		else                   cout << " sym: " << symbol_at(idx);
	}
	cout << endl;

	if (is_inner_node(idx)) {
		print_subtree(child(idx, p_Left), level+1);
		print_subtree(child(idx, p_Right), level+1);
	}
}

void
floating_huff_tree::print_block_list()
{
	cout << "block list:" << endl;
	node_idx smallest_leaf = ZeroLeaf; 
	if (smallest_leaf == NilIdx) smallest_leaf = 0;
	block_t* head = block(smallest_leaf);
	block_t* cur_blk = head;
	do {
		cout << "-> [" << cur_blk->Last;
		if (cur_blk->Last != cur_blk->First) cout << "-" << cur_blk->First;
		cout << "]:" << cur_blk->Weight;
		cout << " / DadRt: " << cur_blk->ParentOfFirst;
		cout << " / ParRt: " << cur_blk->ParityOfFirst;
		if (is_inner_node(cur_blk->First)) cout << " / RtCldOfRt: " << cur_blk->RtChildOfFirst;
		cout << endl;
		cur_blk = cur_blk->NextBlk;
	} while (cur_blk != head);
}

/// AdaptiveHuffmanCoder ////////////////////////////////////////////////////////////////////

AdaptiveHuffmanCoder::AdaptiveHuffmanCoder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams)
{}

AdaptiveHuffmanCoder::AdaptiveHuffmanCoder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name)
{}

void 
AdaptiveHuffmanCoder::init_encoding()
{
	base::init_encoding();
	
	Tree.clear();
	BitsInBuffer = 0;
	Buffer       = 0;
}

void 
AdaptiveHuffmanCoder::encode_stream()
{
	int c;
	while ((c = read_byte()) >= 0 && c != EOF) {
		byte symbol = (byte) c;

		Stack.clear();
		floating_huff_tree::node_idx q = Tree.corresp_leaf(symbol);

		// encode the symbol on the stack if it is new
		if (! Tree.contains(symbol)) {
			int M = Tree.zero_leaf() + 1; // M = # zero weight letters
			assert(M > 0);
			int E = 0, R = 1;
			while (M > R) { ++E; R <<= 1; }
			R = M - R; // now M = 2^E + R with 0<=E<=8 and 0<=R<2^E

			int t = E + 1;
			if (q >= 2*R) { q -= R; --t; }
			while (--t >= 0) {
				Stack.push(huff_bit_stack::bit(q & 1));
				q >>= 1;
			}

			// encode path to 0-leaf
			q = Tree.zero_leaf();
		}

		// encode the path in the tree
		floating_huff_tree::node_idx root = Tree.root();
		while (q != root) {
			Stack.push(Tree.parity(q));
			q = Tree.parent(q);
		}

		// transmit bits on stack in reverse order
		while (!Stack.empty()) {
			if (Stack.pop()) Buffer |= 1 << BitsInBuffer;
			++BitsInBuffer;
			if (BitsInBuffer == 8) {
				if (write_byte(Buffer) < 0) { error(); return; }
				BitsInBuffer = 0; Buffer = 0;
			}
		}

		// update the tree
		Tree.update(symbol);
	}

	if (c == EOF) {
		// flush buffer if necessary
		// in order to make decoding possible 
		// we record in the very last byte the number the number 
		// of used bits in the last data byte (i.e. the one before the very last)
		if (BitsInBuffer > 0) {
			if (write_byte(Buffer) < 0) { error(); return; }	
			if (write_byte((byte) BitsInBuffer) < 0) { error(); return; }
		}
		else {
			if (write_byte(8) < 0) { error(); return; }
		}

		encoding_finished();
	}
}

void 
AdaptiveHuffmanCoder::init_decoding()
{
	base::init_decoding();
	
	Tree.clear();
	BitsInBuffer = 0;
	Buffer       = 0;
	ReadAhead1   = read_byte();
	ReadAhead2   = read_byte();

	if (ReadAhead2 == EOF) { decoding_finished(); return; } // empty file

	if (ReadAhead1 < 0 || ReadAhead2 < 0)
		error("AdaptiveHuffmanCoder::init_decoding: unexpected byte");
}

/*
#define GET_BIT(B) \
	if (BitsInBuffer == 0) { \
	if (ReadAhead2 == EOF) { error(); return; } \
		Buffer = (byte) ReadAhead1; \
		ReadAhead1 = ReadAhead2; \
		ReadAhead2 = read_byte(); \
		BitsInBuffer = ReadAhead2 != EOF ? 8 : (byte) ReadAhead1; \
	} \
	B = Buffer & 1; Buffer >>= 1; --BitsInBuffer;
*/

void 
AdaptiveHuffmanCoder::decode_stream()
{
	while (!failed()) {
		// decode next character
		if (! has_more_bits()) { decoding_finished(); return; }

		// trace path
		floating_huff_tree::node_idx q = Tree.root();
		while (Tree.is_inner_node(q)) {
			q = Tree.child(q, floating_huff_tree::parity_t(get_bit()));
		}

		// decode 0-leaf if necessary
		if (Tree.is_zero_leaf(q)) {
			int M = Tree.zero_leaf() + 1; // M = # zero weight letters
			assert(M > 0);
			int E = 0, R = 1;
			while (M > R) { ++E; R <<= 1; }
			R = M - R; // now M = 2^E + R with 0<=E<=8 and 0<=R<2^E

			q = 0;
			while (--E >= 0) {
				q = (q << 1) + get_bit();
			}

			if (q < R) {
				q = (q << 1) + get_bit();
			}
			else {
				q += R;
			}
		}

		// update the tree
		byte symbol = Tree.symbol_at(q);
		Tree.update(symbol);

		// write symbol
		if (write_byte(symbol) < 0) return;
	}
}

AdaptiveHuffmanCoder::byte
AdaptiveHuffmanCoder::get_bit()
{
	if (BitsInBuffer == 0) {
		if (ReadAhead2 == EOF) { error(); return 1; }
		Buffer = (byte) ReadAhead1;
		ReadAhead1 = ReadAhead2;
		ReadAhead2 = read_byte();
		BitsInBuffer = ReadAhead2 != EOF ? 8 : (byte) ReadAhead1;
	}
	byte result = Buffer & 1; Buffer >>= 1; --BitsInBuffer;
	return result;
}

LEDA_END_NAMESPACE
