// compression module includes
#include <LEDA/coding/memory_streambuf.h>

// core library includes
#include <string.h>

#if defined(LEDA_DEBUG)
#include <LEDA/system/assert.h>
#else
#undef assert
#define	assert(cond) ((void)0)
#endif

LEDA_BEGIN_NAMESPACE

/// memory_streambuf ////////////////////////////////////////////////////////////////////////

void 
memory_streambuf::free_buffer()
{
	//if (WipeBuffer) memset(Buffer, 0, BufferCapacity);
	if (WipeBuffer) memset(Buffer, 0, (size_t)BufferCapacity);
	if (OwnBuffer) delete[] Buffer;
	Buffer = 0; BufferCapacity = Size = 0; OwnBuffer = false;
	setg(0,0,0); setp(0,0); Max_gptr = Max_pptr = 0;
}

void 
memory_streambuf::set_size(streamsize n)
{
	if (n > BufferCapacity) {
		//streamsize new_cap = 256;
		streamsize new_cap = 256;
		while (new_cap < n) new_cap *= 2;
		set_buffer(new char[(size_t)new_cap], 
                                    new_cap, true, WipeBuffer);
	}
	Size = n>=0 ? n : BufferCapacity; // full capacity if n < 0
	reset();
}

void 
memory_streambuf::truncate(streamsize n)
{
	if (n < 0 || n >= BufferCapacity)
		Size = BufferCapacity;
	else
		Size = n;

	reset();
}

void 
memory_streambuf::set_buffer(char* buf, streamsize buf_sz, bool own_buf, bool wipe_buf)
{ 
	free_buffer();
	if (buf != 0 && buf_sz > 0) {
		Buffer = buf; Size = BufferCapacity = buf_sz; 
		OwnBuffer = own_buf; WipeBuffer = wipe_buf;
	}
	reset();
}

void 
memory_streambuf::reset()
{
	setg(Buffer, Buffer, Buffer + Size);
	setp(Buffer, Buffer + Size);
	Max_gptr = Max_pptr = Buffer;
	OverflowOccurred = UnderflowOccurred = false;
}

void 
memory_streambuf::wipe_buffer()
{
	//memset(Buffer, 0, BufferCapacity);
	memset(Buffer, 0, (size_t)BufferCapacity);
	reset();
}

streamsize 
memory_streambuf::in_count() const
{
	if (gptr() > Max_gptr) Max_gptr = gptr();
	return streamsize(Max_gptr - eback());
}

streamsize 
memory_streambuf::out_count() const
{
	if (pptr() > Max_pptr) Max_pptr = pptr();
	return streamsize(Max_pptr - pbase());
}

streampos 
memory_streambuf::seekoff(streamoff off, ios::seekdir way, ios::openmode which)
{
	if (gptr() > Max_gptr) Max_gptr = gptr();
	if (pptr() > Max_pptr) Max_pptr = pptr();

	switch (way) {
	case ios::beg:	return seekpos(off, which);
	case ios::cur:	if (which == ios::in || which == ios::out) 
						return seekpos(get_current_pos(which) + off, which);
					break;
	case ios::end:	if (which == ios::in) return seekpos(streampos(Size) + off, which);
					if (which == ios::out) return seekpos(streamoff(Max_pptr - pbase()) + off, which);
	default:	return EOF; // just to avoid a warning
	}
	return streampos(-1);
}

streampos 
memory_streambuf::seekpos(streampos sp, ios::openmode which)
{
	// NOTE: sp == SIZE is okay! (e.g. seek(ios::end))
	if (sp < 0 || sp > Size) return streampos(-1); 
	if (which & ios::in) {
		if (gptr() > Max_gptr) Max_gptr = gptr();
		setg(Buffer, Buffer + streamoff(sp), Buffer + Size); 
	}
	if (which & ios::out) {
		if (pptr() > Max_pptr) Max_pptr = pptr();
		setp(Buffer, Buffer + Size); pbump(int(sp));
	}
	return sp;
}

streampos 
memory_streambuf::get_current_pos(ios::openmode which) const
{
	if (which & ios::in) return streamoff(gptr() - eback());
	else                 return streamoff(pptr() - pbase());
}

/// mem_strbuf_with_callbacks ///////////////////////////////////////////////////////////////

int 
mem_strbuf_with_callbacks::overflow(int c) 
{ 
	if (OverflowCallbackFct != 0 && out_count() > 0) {
		bool ok = OverflowCallbackFct(get_buffer(), out_count(), OverflowCallbackInfo);
		if (ok) {
			reset();
#ifndef LEDA_COMPRESSION_WA6
			return sputc( (char) c);
#else
			return sputc(c);
#endif
		}
	}

	return base::overflow(c);
}

int mem_strbuf_with_callbacks::underflow() 
{ 
	if (UnderflowCallbackFct != 0) {
		streamsize n = UnderflowCallbackFct(get_buffer(), get_capacity(), UnderflowCallbackInfo);
		if (n > 0) {
			truncate(n);
			return sgetc();
		}
	}

	return base::underflow();
}

int 
mem_strbuf_with_callbacks::sync()
{
	if (OverflowCallbackFct != 0 && out_count() > 0) {
		bool ok = OverflowCallbackFct(get_buffer(), out_count(), OverflowCallbackInfo);
		if (!ok) return -1;
		reset();
	}

	return base::sync();
}

LEDA_END_NAMESPACE
