#include <LEDA/coding/secure_socket_streambuf.h>

#if defined(LEDA_DEBUG)
#include <LEDA/system/assert.h>
#else
#undef assert
#define	assert(cond) ((void)0)
#endif

LEDA_BEGIN_NAMESPACE

/// secure_socket_streambuf /////////////////////////////////////////////////////////////////

secure_socket_streambuf::secure_socket_streambuf(leda_socket& s, 
												 const CryptByteString& auth_passphrase, 
												 const CryptByteString& cipher_passphrase, 
												 uint32 out_buf_sz, uint32 in_buf_sz, 
												 bool enable_compression, bool send_ack)
 : Socket(s), 
   PutBuffer(0), GetBuffer(0), 
   OutgoingPacketSize(out_buf_sz), IncomingPacketSize(in_buf_sz), 
   OutgoingMessageCnt(0), IncomingMessageCnt(0), 
   Failed(false), HasPutEOF(false), HasGotEOF(false), 
   CompressionEnabled(false), 
   WaitForAcknowlegdeOfOutgoingPackets(false), SendAcknowlegdeForIncomingPackets(send_ack)
{ init(auth_passphrase, cipher_passphrase, enable_compression); }

secure_socket_streambuf::secure_socket_streambuf(leda_socket& s, 
												 const CryptByteString& passphrase, 
												 uint32 out_buf_sz, uint32 in_buf_sz, 
												 bool enable_compression, bool send_ack)
 : Socket(s), 
   PutBuffer(0), GetBuffer(0), 
   OutgoingPacketSize(out_buf_sz), IncomingPacketSize(in_buf_sz), 
   OutgoingMessageCnt(0), IncomingMessageCnt(0), 
   Failed(false), HasPutEOF(false), HasGotEOF(false), 
   CompressionEnabled(false), 
   WaitForAcknowlegdeOfOutgoingPackets(false), SendAcknowlegdeForIncomingPackets(send_ack)
{ init(passphrase, passphrase, enable_compression); }

void 
secure_socket_streambuf::init(const CryptByteString& auth_passphrase, 
							  const CryptByteString& cipher_passphrase, 
							  bool enable_compression)
{
	negotiate(auth_passphrase, cipher_passphrase, enable_compression);
	allocate_buffers();

	setg(0,0,0); setp(0,0);
}

secure_socket_streambuf::~secure_socket_streambuf()
{
	if (!failed() && !HasPutEOF) sputEOF();
	free_buffers();
}


void 
secure_socket_streambuf::allocate_buffers()
{
	PutBuffer = OutgoingPacketSize > 0 ? new char[OutgoingPacketSize] : 0;
	GetBuffer = IncomingPacketSize > 0 ? new char[IncomingPacketSize] : 0;
}

void 
secure_socket_streambuf::free_buffers()
{
	delete[] PutBuffer; PutBuffer = 0;
	delete[] GetBuffer; GetBuffer = 0;
}

void 
secure_socket_streambuf::error(const char* msg) 
{
	if (!Failed) {
		Failed = true; ErrorMsg = msg;
	}
	setp(0,0); setg(0,0,0);

#if defined(LEDA_DEBUG) || defined(LEDA_SOCK_DEBUG)
	cout << msg << endl;
#endif
}

void   
secure_socket_streambuf::store_uint32(uint32 val, byte* mem)
{
	mem[0] = (byte) (val);
	mem[1] = (byte) (val >> 8);
	mem[2] = (byte) (val >> 16);
	mem[3] = (byte) (val >> 24);
}

secure_socket_streambuf::uint32 
secure_socket_streambuf::load_uint32(const byte* mem)
{
	return (uint32(mem[0])) | (uint32(mem[1])<<8) | (uint32(mem[2])<<16) | (uint32(mem[3])<<24);
}

void 
secure_socket_streambuf::negotiate(const CryptByteString& auth_passphrase, 
								   const CryptByteString& cipher_passphrase, 
								   bool enable_compression)
{
	// no compression during negotiations
	CryptStreambuf.set_coder(new crypt_coder(), true);
	CompressionEnabled = false;

	AuthKey   = CryptKey::generate_key(256/8, auth_passphrase, CryptByteString("a"));
	CipherKey = CryptKey::generate_key(256/8, cipher_passphrase, CryptByteString("c"));

	CryptByteString session_seed;
	if (Socket.get_host() == "") negotiate_as_server(session_seed, enable_compression);
	else                         negotiate_as_client(session_seed, enable_compression);

	if (failed()) {
		AuthKey.wipe(); CipherKey.wipe(); return;
	}

	// compute session keys
	AuthKey   = CryptKey::generate_key(256/8, auth_passphrase, session_seed+CryptByteString("a"));
	CipherKey = CryptKey::generate_key(256/8, cipher_passphrase, session_seed+CryptByteString("c"));

	// enable compression if appropriate
	if (enable_compression) {
		CryptStreambuf.set_coder(new crypt_and_compress_coder(), true);
		CompressionEnabled = true;
	}
}

/*
Protocol: A = client, B = server

-	Initial negotiation:
	A: 1) E_K(A_K(Ok|0||V|N_A))
	B: 2) E_K(A_K(Ok|0||N_A|N_B))
	A: 3) E_K(A_K(Ok|1||N_B|PBS_A|GBS_A|F_A|R_A))
	B: 4) E_K(A_K(Ok|1||N_A|PBS_B|GBS_B|F_B|R_B|S))

	where
	E_K(...)     : enciphering with master-cipher-key (generated from passphrase)
	A_K(...)     : authentication with master-auth-key (generated from passphrase)
	Ok           : Ok-flag that starts transmission (as opposed to EOF or Ack)
	0, 1, ...    : message numbers: A and B maintain counters of outgoing and incoming
	               messages to avoid replay attacks
	V            : protocol version
	N_A, N_B     : nonces generated by A and B respectively (to avoid replay attacks)
	PBS_A, PBS_B : requested put-buffer size (for outgoing packets)
	GBS_A, GBS_B : requested get-buffer size (for incoming packets)
	F_A, F_B     : session flags (send acknowledge, use compression, ...)
	R_A, R_B     : reserved for future extensions
	S            : seesion-seed (generated by B, i.e. server)

	-> The size of the Ok-Flag is 8 bits, the nonces N_A and N_B are 8 bytes long, 
	   the session-seed S is 32 bytes long, all remaining parameters have size 32 bits.
	-> After receiving message 2), A checks verifies N_A. So does B after message 3).
	-> A chooses as put-buffer size min(put-buffer size requ. by A, get-buffer size requ. by B)
	   (similar for get-buffer and for B's buffers)
	-> A and B decide independently whether they want to send acknowledgements or not, the
	   other side respects the decision of its mate
	-> compression is enabled iff both parties want to

-	Regular message exchange (later):
	A: E_K*(A_K*(Ok|MsgNr_A||Msg)) or E_K*(A_K*(EOF|MsgNr_A||Msg))
	B: E_K*(A_K*(Ok|MsgNr_B||Msg)) or E_K*(A_K*(EOF|MsgNr_B||Msg))

	where
	K* : session key generated from passphrase and session-id

	-> After receiving an EOF-packet no further packets are accepted!

-	Acknowlegement (of received message):
	A: E_K*(A_K*(Ack|MsgNr_A||AckNonce))
	B: E_K*(A_K*(Ack|MsgNr_B||AckNonce))

	where
	AckNonce : nonce generated to avoid that an attacker might guess the 
	           plaintext of an acknowledge-packet
*/

void 
secure_socket_streambuf::negotiate_as_client(CryptByteString& session_seed, bool& enable_compression)
{
	// A = client, B = server
	char msg[MaxMsgSize];
	int num_rcv;	

	session_seed = CryptByteString(SeedSize);
	CryptByteString N_A = CryptKey::generate_nonce(NonceSize);

	// send E_K(A_K(Ok|0||V|N_A))
	store_uint32(ProtocolVersion, msg);
	memcpy(msg+4, N_A.get_bytes(), NonceSize);
	encrypt_and_send(msg, 4+NonceSize);

	// receive E_K(A_K(Ok|0||N_A|N_B))
	num_rcv = receive_and_decrypt(msg, MaxMsgSize);
	if (num_rcv != 2*NonceSize) {
		error("ssb:negotiate_as_client: wrong length msg 0"); return;
	}
	if (CryptByteString((const byte*) msg, NonceSize) != N_A) {
		error("ssb:negotiate_as_client: wrong nonce msg 0"); return;
	}
	CryptByteString N_B((const byte*) msg+NonceSize, NonceSize);

	// send E_K(A_K(Ok|1||N_B|PBS_A|GBS_A|F_A|R_A))
	// where PBS_A=out_buf_sz, GBS_A=in_buf_sz, F_A = flags, R_A = reserved
	uint32 F_A = 0, R_A = 0;
	if (enable_compression) F_A |= f_Compress;
	if (SendAcknowlegdeForIncomingPackets) F_A |= f_Ack;
	memcpy(msg, N_B.get_bytes(), NonceSize);
	store_uint32(OutgoingPacketSize, msg+NonceSize);
	store_uint32(IncomingPacketSize, msg+NonceSize+4);
	store_uint32(F_A, msg+NonceSize+8);
	store_uint32(R_A, msg+NonceSize+12);
	encrypt_and_send(msg, NonceSize+4*4);

	// receive E_K(A_K(Ok|1||N_A|PBS_B|GBS_B|F_B|R_B|S))
	// where PBS_B=out_buf_sz, GBS_B=in_buf_sz, F_B = flags, R_B = reserved
	//       S=session_seed
	num_rcv = receive_and_decrypt(msg, MaxMsgSize);
	if (num_rcv != NonceSize + 4*4 + SeedSize) {
		error("ssb:negotiate_as_client: wrong length msg 1"); return;
	}
	if (CryptByteString((const byte*) msg, NonceSize) != N_A) {
		error("ssb:negotiate_as_client: wrong nonce msg 1"); return;
	}
	uint32 mate_out_buf_sz = load_uint32(msg+NonceSize);
	uint32 mate_in_buf_sz  = load_uint32(msg+NonceSize+4);
	uint32 mate_F          = load_uint32(msg+NonceSize+8);
	// ignored: uint32 mate_R_A        = load_uint32(msg+NonceSize+12);
	bool   mate_compress   = (mate_F & f_Compress) != 0;
	bool   mate_sends_ack  = (mate_F & f_Ack) != 0;
	memcpy(session_seed.get_bytes(), msg+NonceSize+4*4, SeedSize);

	// calculate session parameters
	OutgoingPacketSize = leda_min(OutgoingPacketSize, mate_in_buf_sz);
	IncomingPacketSize = leda_min(IncomingPacketSize, mate_out_buf_sz);
	enable_compression = enable_compression && mate_compress; // only if both agree
	WaitForAcknowlegdeOfOutgoingPackets = mate_sends_ack;
}

void 
secure_socket_streambuf::negotiate_as_server(CryptByteString& session_seed, bool& enable_compression)
{
	// A = client, B = server
	char msg[MaxMsgSize];
	int num_rcv;	

	session_seed = CryptKey::generate_salt(SeedSize);
	CryptByteString N_B = CryptKey::generate_nonce(NonceSize);

	// receive E_K(A_K(Ok|0||V|N_A))
	num_rcv = receive_and_decrypt(msg, MaxMsgSize);
	if (num_rcv != 4+NonceSize) {
		error("ssb:negotiate_as_server: wrong length msg 0"); return;
	}
	if (load_uint32(msg) != ProtocolVersion) {
		error("ssb:negotiate_as_server: wrong protocol version"); return;
	}
	CryptByteString N_A((const byte*) msg+4, NonceSize);

	// send E_K(A_K(Ok|0||N_A|N_B))
	memcpy(msg,           N_A.get_bytes(), NonceSize);
	memcpy(msg+NonceSize, N_B.get_bytes(), NonceSize);
	encrypt_and_send(msg, 2*NonceSize);

	// receive E_K(A_K(Ok|1||N_B|PBS_A|GBS_A|F_A|R_A))
	// where PBS_A=out_buf_sz, GBS_A=in_buf_sz, F_A = flags, R_A = reserved
	num_rcv = receive_and_decrypt(msg, MaxMsgSize);
	if (num_rcv != NonceSize+4*4) {
		error("ssb:negotiate_as_server: wrong length msg 1"); return;
	}
	if (CryptByteString((const byte*) msg, NonceSize) != N_B) {
		error("ssb:negotiate_as_server: wrong nonce msg 1"); return;
	}
	uint32 mate_out_buf_sz = load_uint32(msg+NonceSize);
	uint32 mate_in_buf_sz  = load_uint32(msg+NonceSize+4);
	uint32 mate_F          = load_uint32(msg+NonceSize+8);
	// ignored: uint32 mate_R_A        = load_uint32(msg+NonceSize+12);
	bool   mate_compress   = (mate_F & f_Compress) != 0;
	bool   mate_sends_ack  = (mate_F & f_Ack) != 0;

	// send E_K(A_K(Ok|1||N_A|PBS_B|GBS_B|F_B|R_B|S))
	// where PBS_B=out_buf_sz, GBS_B=in_buf_sz, F_B = flags, R_B = reserved
	//       S=session_seed
	uint32 F_B = 0, R_B = 0;
	if (enable_compression) F_B |= f_Compress;
	if (SendAcknowlegdeForIncomingPackets) F_B |= f_Ack;
	memcpy(msg, N_A.get_bytes(), NonceSize);
	store_uint32(OutgoingPacketSize, msg+NonceSize);
	store_uint32(IncomingPacketSize, msg+NonceSize+4);
	store_uint32(F_B, msg+NonceSize+8);
	store_uint32(R_B, msg+NonceSize+12);
	memcpy(msg+NonceSize+16, session_seed.get_bytes(), SeedSize);
	encrypt_and_send(msg, NonceSize + 4*4 + SeedSize);

	// calculate session parameters
	OutgoingPacketSize = leda_min(OutgoingPacketSize, mate_in_buf_sz);
	IncomingPacketSize = leda_min(IncomingPacketSize, mate_out_buf_sz);
	enable_compression = enable_compression && mate_compress; // only if both agree
	WaitForAcknowlegdeOfOutgoingPackets = mate_sends_ack;
}


void 
secure_socket_streambuf::reset_crypt_streambuf()
{
	CryptStreambuf.reset();
	CryptStreambuf.set_encoded_stream(&InternalBuffer);

	coder_base* coder = CryptStreambuf.get_coder();
	if (!coder) {
		error("ssb:reset_crypt_streambuf: no coder"); return;
	}

	if (CompressionEnabled) {
		// crypt & compress
		crypt_and_compress_coder* c = (crypt_and_compress_coder*) coder;
		c->get_coder1()->set_key(AuthKey);
		c->get_coder3()->set_key(CipherKey);
	}
	else {
		// only crypt
		crypt_coder* c = (crypt_coder*) coder;
		c->get_coder1()->set_key(AuthKey);
		c->get_coder2()->set_key(CipherKey);
	}
}

void 
secure_socket_streambuf::encrypt_and_send(char* buf, int n, Header h)
{
	InternalBuffer.set_size(streamsize(HeaderSize + n + Overshoot));

	// prepare CryptStreambuf
	reset_crypt_streambuf();

	// put header
	char header[HeaderSize];
	header[0] = h;
	store_uint32(OutgoingMessageCnt++, header + 1);
	CryptStreambuf.sputn(header, HeaderSize);

	// put data
	CryptStreambuf.sputn(buf, n);

	// finish message
	CryptStreambuf.sputEOF();

	// check
	if (CryptStreambuf.failed() || CryptStreambuf.get_coder()->failed() || InternalBuffer.had_overflow()) { 
		error("ssb::encrypt_and_send: encode failed!"); return; 
	}

	// send encrypted data
	Socket.send_bytes(InternalBuffer.get_buffer(), (int) InternalBuffer.out_count());
//	cout << "ssb::encrypt_and_send: num_sent = " << (int) InternalBuffer.out_count() << endl;
}

int 
secure_socket_streambuf::receive_and_decrypt(char* buf, int n, Header expected_header)
{
	InternalBuffer.set_size(streamsize(HeaderSize + n + Overshoot));

	// receive encrypted data
	int num_rcv = Socket.receive_bytes(InternalBuffer.get_buffer(), (int) InternalBuffer.get_capacity());
//	cout << "ssb::receive_and_decrypt: num_rcv = " << num_rcv << endl;

	// verify length of received message
	if (num_rcv <= 0 || num_rcv > (int) InternalBuffer.get_capacity()) {
		string msg = "ssb::receive_and_decrypt: receive error: ";
		if (num_rcv <= 0) msg += Socket.get_error();
		else              msg += string("num_rcv = %d larger than expected size", num_rcv);
		error(msg); return -1;
	}

	InternalBuffer.truncate((streamsize) num_rcv);

	// prepare CryptStreambuf
	reset_crypt_streambuf();

	// verify header
	char header[HeaderSize];
	if (CryptStreambuf.sgetn(header, HeaderSize) != HeaderSize) {
		error("ssb::receive_and_decrypt: could not read header"); return -1;
	}
	char h = header[0];
	bool ok = (h==expected_header) || (expected_header==h_OkOrEOF && (h==h_Ok || h==h_EOF));
	ok &= load_uint32(header+1) == IncomingMessageCnt++;
	if (! ok) {
		error("ssb::receive_and_decrypt: bad header!"); return -1; 
	}
	if (h == h_EOF) HasGotEOF = true;

	// get data
	int data_sz = (int)CryptStreambuf.sgetn(buf, n);
	if (CryptStreambuf.failed() || CryptStreambuf.get_coder()->failed() || !InternalBuffer.had_underflow()) { 
		error("ssb::receive_and_decrypt: decode failed!"); return -1; 
	}
	
	return data_sz;
}


void 
secure_socket_streambuf::send_put_buffer(Header header)
{
	if (failed()) {	setp(0,0); return; }

	if (OutgoingPacketSize == 0 || HasPutEOF) { 
		error("send_put_buffer"); return;
	}

	if (pbase() && (pptr() != PutBuffer || header != h_Ok)) { 
		// (do not send empty data buffer)
//		if (header == h_Ok) cout << "+"; else cout << "/"; cout.flush();

		// send put buffer
		encrypt_and_send(PutBuffer, int(pptr()-PutBuffer), header);
		if (failed()) return;

		// and wait for acknowledge
		if (WaitForAcknowlegdeOfOutgoingPackets) {
			char answer[AckNonceSize];
			if (receive_and_decrypt(answer, sizeof(answer), h_Ack) != AckNonceSize) {
				error("send_put_buffer: did not get acknowledge"); return;
			}
		}
	}

	setp(PutBuffer, PutBuffer+OutgoingPacketSize);
}

int 
secure_socket_streambuf::fill_get_buffer()
{
	if (failed() || HasGotEOF) { setg(0,0,0); return EOF; }

	if (IncomingPacketSize == 0) { 
		error("fill_get_buffer: no get buffer"); return EOF; 
	}

	int num_rcv = receive_and_decrypt(GetBuffer, IncomingPacketSize, h_OkOrEOF);
	if (failed()) return EOF;

	// received correctly -> send acknowledge (if appropriate)
	if (SendAcknowlegdeForIncomingPackets) {
		CryptByteString ack_nonce = CryptKey::generate_nonce(AckNonceSize, CryptByteString(), 10);
		encrypt_and_send((char*) ack_nonce.get_bytes(), AckNonceSize, h_Ack);
	}

//	if (header == h_Ok) cout << "."; else cout << "*"; cout.flush();

	if (HasGotEOF) {
		// explicit EOF received
		if (num_rcv == 0) { setg(0,0,0); return EOF; }
	}
	else {
		// standard message: verify non-emptyness
		if (num_rcv <= 0) {
			error("fill_get_buffer: no data received"); return EOF;
		}
	}

	// non-empty message:
	setg(GetBuffer, GetBuffer, GetBuffer+num_rcv);
	return byte(GetBuffer[0]);
}

int 
secure_socket_streambuf::overflow(int c)
{
	send_put_buffer();

	return (!failed() && c != EOF) ? sputc(c) : EOF;
}

int 
secure_socket_streambuf::underflow()
{
	return fill_get_buffer();
}

int 
secure_socket_streambuf::sync()
{
	if (OutgoingPacketSize > 0) send_put_buffer();
	return failed() ? -1 : 0;
}

void 
secure_socket_streambuf::sputEOF()
{
	if (OutgoingPacketSize == 0) { error("sputEOF: no put buffer"); return; }

	// send (unsent contents of put buffer and) EOF
	send_put_buffer(h_EOF);

	if (!failed()) HasPutEOF = true; 
        setp(0,0);
}

LEDA_END_NAMESPACE
