#include <LEDA/coding/socket_streambuf.h>

#if defined(LEDA_DEBUG)
#include <LEDA/system/assert.h>
#else
#undef assert
#define	assert(cond) ((void)0)
#endif

LEDA_BEGIN_NAMESPACE

/// socket_streambuf ////////////////////////////////////////////////////////////////////////

socket_streambuf::socket_streambuf(leda_socket& s, uint32 out_buf_sz, uint32 in_buf_sz, 
								   bool send_ack)
 : Socket(s), PutBuffer(0), GetBuffer(0), 
   OutgoingPacketSize(0), IncomingPacketSize(0), 
   Failed(false), HasPutEOF(false), HasGotEOF(false), HasHadPuts(false),
   WaitForAcknowlegdeOfOutgoingPackets(false), SendAcknowlegdeForIncomingPackets(send_ack)
{
	negotiate_packet_sizes(out_buf_sz, in_buf_sz);
	allocate_buffers();

	setg(0,0,0); setp(0,0);
}

socket_streambuf::~socket_streambuf()
{
	if (!failed() && HasHadPuts && !HasPutEOF) sputEOF();
	free_buffers();
}

void 
socket_streambuf::error(const char* msg) 
{
	if (!Failed) {
		Failed = true; ErrorMsg = msg;
	}
	setp(0,0); setg(0,0,0);

#if defined(LEDA_DEBUG) || defined(LEDA_SOCK_DEBUG)
	cout << msg << endl;
#endif
}

void   
socket_streambuf::store_uint32(uint32 val, byte* mem)
{
	mem[0] = (byte) (val);
	mem[1] = (byte) (val >> 8);
	mem[2] = (byte) (val >> 16);
	mem[3] = (byte) (val >> 24);
}

socket_streambuf::uint32 
socket_streambuf::load_uint32(const byte* mem)
{
	return (uint32(mem[0])) | (uint32(mem[1])<<8) | (uint32(mem[2])<<16) | (uint32(mem[3])<<24);
}

void 
socket_streambuf::negotiate_packet_sizes(uint32 out_buf_sz, uint32 in_buf_sz)
{
	byte my_params[11], mate_params[11];

	// send my parameters
	my_params[0] = 0x73; // magic byte
	store_uint32(out_buf_sz, my_params+1);
	store_uint32(in_buf_sz,  my_params+5);
	my_params[9] = SendAcknowlegdeForIncomingPackets ? 1 : 0;
	my_params[10] = 0x12; // magic byte

	Socket.send_bytes((char*) my_params, sizeof(my_params));

	// receive mate parameters
	bool ok = Socket.receive_bytes((char*) mate_params, sizeof(mate_params)) == sizeof(mate_params);

	if (!ok || mate_params[0] != my_params[0] || mate_params[sizeof(my_params)-1] != my_params[sizeof(mate_params)-1]) {
		OutgoingPacketSize = IncomingPacketSize = 0;
		error("negotiate_packet_sizes: mate params could not be determined!");
	}

	uint32 mate_out_buf_sz = load_uint32(mate_params+1);
	uint32 mate_in_buf_sz  = load_uint32(mate_params+5);
	bool   mate_will_acknowledge_incoming_packets = mate_params[9] != 0;

	// calculate session parameters
	OutgoingPacketSize = leda_min(out_buf_sz, mate_in_buf_sz);
	IncomingPacketSize = leda_min(in_buf_sz,  mate_out_buf_sz);
	WaitForAcknowlegdeOfOutgoingPackets = mate_will_acknowledge_incoming_packets;

//	cout << OutgoingPacketSize << " " << IncomingPacketSize << endl;
}

void 
socket_streambuf::allocate_buffers()
{
	PutBuffer = OutgoingPacketSize > 0 ? new char[HeaderSize+OutgoingPacketSize] : 0;
	GetBuffer = IncomingPacketSize > 0 ? new char[HeaderSize+IncomingPacketSize] : 0;
}

void 
socket_streambuf::free_buffers()
{
	delete[] PutBuffer; PutBuffer = 0;
	delete[] GetBuffer; GetBuffer = 0;
}


void 
socket_streambuf::send_put_buffer(Header header)
{
	if (failed()) {	setp(0,0); return; }

	if (OutgoingPacketSize == 0 || HasPutEOF) { 
		error("send_put_buffer"); return;
	}

	if (pbase() && (pptr() != PutBuffer+HeaderSize || header != h_Ok)) { 
		// (do not send empty data buffer)
//		if (header == h_Ok) cout << "+"; else cout << "/"; cout.flush();

		// send put buffer
		PutBuffer[0] = header;
		Socket.send_bytes(PutBuffer, int(pptr()-PutBuffer));

		// and wait for acknowledge
		if (WaitForAcknowlegdeOfOutgoingPackets) {
			char answer;
			bool ok = Socket.receive_bytes(&answer, sizeof(answer)) == sizeof(answer);
			if (answer != h_Ack || !ok) {
				error("send_put_buffer: did not get acknowledge"); return;
			}
		}
	}

	setp(PutBuffer+HeaderSize, PutBuffer+HeaderSize+OutgoingPacketSize);
}

int 
socket_streambuf::fill_get_buffer()
{
	if (failed() || HasGotEOF) { setg(0,0,0); return EOF; }

	if (IncomingPacketSize == 0) { 
		error("fill_get_buffer: no get buffer"); return EOF; 
	}

	int num_rcv = Socket.receive_bytes(GetBuffer, HeaderSize+IncomingPacketSize);

	// verify received message
	if (num_rcv <= 0 || uint32(num_rcv) > HeaderSize+IncomingPacketSize) {
		string msg = "fill_get_buffer: receive error: ";
		if (num_rcv <= 0) msg += Socket.get_error();
		else              msg += string("num_rcv = %d larger than packet size", num_rcv);
		error(msg); return EOF;
	}
	char header = GetBuffer[0];
	if (header != h_Ok && header != h_EOF) {
		error(string("fill_get_buffer: bad header %d", header)); return EOF;
	}

	// received correctly -> send acknowledge (if appropriate)
	if (SendAcknowlegdeForIncomingPackets) {
		char acknowledge = h_Ack;
		Socket.send_bytes(&acknowledge, sizeof(acknowledge));
	}

//	if (header == h_Ok) cout << "."; else cout << "*"; cout.flush();

	if (header == h_EOF) {
		// explicit EOF received
		HasGotEOF = true;
		if (num_rcv == HeaderSize) { setg(0,0,0); return EOF; }
	}
	else {
		// standard message: verify non-emptyness
		if (num_rcv <= HeaderSize) {
			error("fill_get_buffer: no data received"); return EOF;
		}
	}

	// non-empty message:
	setg(GetBuffer+HeaderSize, GetBuffer+HeaderSize, GetBuffer+num_rcv);
	return byte(GetBuffer[HeaderSize]);
}

int 
socket_streambuf::overflow(int c)
{
	HasHadPuts = true;
	send_put_buffer();

	return (!failed() && c != EOF) ? sputc(c) : EOF;
}

int 
socket_streambuf::underflow()
{
	return fill_get_buffer();
}

int 
socket_streambuf::sync()
{
	if (OutgoingPacketSize > 0) send_put_buffer();
	return failed() ? -1 : 0;
}

void 
socket_streambuf::sputEOF()
{
	if (OutgoingPacketSize == 0) { error("sputEOF: no put buffer"); return; }

	// send (unsent contents of put buffer and) EOF
	send_put_buffer(h_EOF);

	if (!failed()) HasPutEOF = true; 
        setp(0,0);
}

LEDA_END_NAMESPACE
