/*******************************************************************************
+
+  LEDA 7.0  
+
+
+  minizip.cpp
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


#include <LEDA/core/list.h>
#include <LEDA/core/string.h>
#include <LEDA/core/array.h>
#include <LEDA/system/file.h>

#include <LEDA/coding/crypt.h>

using namespace leda;

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::cin;
using std::cerr;
using std::endl;
using std::ostream;
using std::istream;
using std::ofstream;
using std::ifstream;
using std::streambuf;
using std::filebuf;
using std::ios;
#endif

#ifdef LEDA_COMPRESSION_WA1
#define openmode open_mode
#endif

#define open_in  coder_base::binary_in()
#define open_out coder_base::binary_out()

//typedef CoderPipe3<BWTCoder, MTFCoder, A0Coder> Coder;
typedef PPMIICoder Coder;
typedef CFBCoder<Rijndael> CipherCoder;

// display short help message
static void usage()
{
	cout << "minizip [options] [files_to_process]" << endl;
	cout << "options:" << endl;
	cout << "  -f archive_file: specifies the name of the archive" << endl;
	cout << "  -p path: specifies the path which is appended to the names of decoded files" << endl;
	cout << "  -d: switch to decode mode, default is encode" << endl;
	cout << "  -t: show table of contents of archive file" << endl;
	cout << "  -c: use cryptography" << endl;
}

// copy src to tgt (till src signals EOF)
static void copy_streambuffer(streambuf* src, streambuf* tgt)
{
	char buffer[1024];

	int bytes_read;
	do {
		bytes_read = src->sgetn(buffer, 1024);
		tgt->sputn(buffer, bytes_read);
	} while (bytes_read == 1024);
}

// copy n bytes from src to tgt
static void copy_n_bytes_between_streambuffers(streambuf* src, streambuf* tgt, int n)
{
	char buffer[1024];

	while (n > 0) {
		int bytes_to_read = n > 1024 ? 1024 : n;

		int bytes_read = src->sgetn(buffer, bytes_to_read);
		tgt->sputn(buffer, bytes_read);

		if (bytes_read != bytes_to_read) return; // error

		n -= bytes_to_read;
	}
}

static ostream* setup_ostream(string archive_file, bool crypt)
{
	typedef encoding_ostream<Coder>                           CompressOnlyOStream;
	typedef encoding_ostream< CoderPipe2<Coder,CipherCoder> > CompressAndCryptOStream;

	filebuf* p_raw_filebuf = new filebuf;
	if (! p_raw_filebuf->open(archive_file, open_out)) {
		cerr << "could not open archive " << archive_file << endl; return 0;
	}
	ostream arc_raw(p_raw_filebuf); // data written to this stream is not encoded

	// every archive starts with "MZ." where "." encodes the format
	// so far we have only two formats: 0x00: compressed, 0x01: compressed & encrypted
	arc_raw << "MZ";

	if (!crypt) { // no encryption
		arc_raw.put((char) 0x00); arc_raw.flush();
		CompressOnlyOStream* p_arc = new CompressOnlyOStream(p_raw_filebuf, true);
		cout << "  using coder: " << p_arc->get_coder()->get_description() << endl;
		return p_arc;
	}
	else { // encryption
		arc_raw.put((char) 0x01); 

		// generate key (and salt)
		CryptByteString passphrase = CryptKey::read_passphrase("  enter passphrase (>= 6 letters): ", 6);
		CryptByteString salt;
		CryptKey key = CryptKey::generate_key_and_salt(256/8, 128/8, passphrase, salt, true);
		if (key.is_empty()) return 0;
		arc_raw << salt; arc_raw.flush(); // write salt (unencrypted!)

		// now the encoding_ostream takes over the filebuf of arc_raw ...
		CompressAndCryptOStream* p_arc = new CompressAndCryptOStream(p_raw_filebuf, true);
		cout << "  using coder: " << p_arc->get_coder()->get_description() << endl;
		p_arc->get_coder()->get_coder2()->set_key(key);
		return p_arc;
	}
}

static istream* setup_istream(string archive_file)
{
	typedef decoding_istream<CryptAutoDecoder> AutoDecIStream;

	filebuf* p_raw_filebuf = new filebuf;
	if (! p_raw_filebuf->open(archive_file, open_in)) {
		cerr << "could not open archive " << archive_file << endl; return 0;
	}
	istream arc_raw(p_raw_filebuf); // data read from this stream is not decoded

	// check for id "MZ."
	char id_buffer[3];
	arc_raw.read(id_buffer, 3);
	if (arc_raw.fail() || id_buffer[0] != 'M' || id_buffer[1] != 'Z') { 
		cerr << "archive corrupted" << endl; return 0;
	}
	char format = id_buffer[2];
	if (format > 0x01) {
		cerr << "warning: unknown archive format" << endl;
	}
	bool crypt = (format & 0x01) != 0;

	// prepare decoding (with CryptAutoDecoder this is easy)
	AutoDecIStream* p_arc = new AutoDecIStream(p_raw_filebuf, true);
	if (crypt) {
		CryptByteString passphrase = CryptKey::read_passphrase("  enter passphrase (>= 6 letters): ", 6);
		CryptByteString salt;
		arc_raw >> salt; // read salt (unencrypted)
		CryptKey key = CryptKey::generate_key(256/8, passphrase, salt, true);
		p_arc->get_coder()->add_key(key);
	}

	// display some coder info
	char ch; p_arc->get(ch); p_arc->putback(ch); // we have to read s.th. first!
	cout << "  using coder: " << p_arc->get_coder()->get_description() << endl;

	return p_arc;
}


static int encode_files(const list<string>& files_to_process, string archive_file, bool crypt)
{
	cout << "Packing " << archive_file << " ... " << endl;

	// generate the archive-ostream "arc"
	ostream* p_arc = setup_ostream(archive_file, crypt);
	if (! p_arc) return 1;
	ostream& arc = *p_arc;

	// write table_of_contents
	arc << "#" << files_to_process.size() << '\n';
	string filename;
	forall(filename, files_to_process) {
		if (!is_file(filename)) { cerr << filename << " not found" << endl; return 1; }
		arc << filename << endl << size_of_file(filename) << endl;
	}

	// put TOC end marker
	arc.put('#');

	// encode the files
	forall(filename, files_to_process) {
		cout << "  encoding: " << filename << " ... "; cout.flush();

		// open input file
		filebuf in;
		in.open(filename, open_in);
		if (! in.is_open()) { cerr << filename << "not opened" << endl; return 1; }

		// encode file into archive
		copy_streambuffer(&in, arc.rdbuf());

		// write EOF-marker (not really necessary)
		arc.put('#'); arc.put('L'); arc.put('#');

		if (arc.fail()) { cerr << "failure" << endl; return 1; }
		cout << "ok" << endl;
	}

	delete p_arc;
	return 0;
}


static int read_toc(istream& arc, array<string>& name_of_file, array<int>& size_of_file)
{
	// check TOC start marker
	char ch = 0;
	arc >> ch; if (arc.fail() || ch != '#') { cerr << "archive table of contents corrupted" << endl; return 1; }

	// read table_of_contents
	int num_of_files; arc >> num_of_files; 
	arc.get(ch); if (arc.fail() || ch != '\n') { cerr << "archive table of contents corrupted" << endl; return 1; }
	if (num_of_files == 0) return 2;
	name_of_file.resize(1, num_of_files);
	size_of_file.resize(1, num_of_files);
	int i;
	for (i=1; i<=num_of_files; ++i) {
		arc >> name_of_file[i] >> size_of_file[i];
	}

	// check TOC end marker
	arc >> ch; if (arc.fail() || ch != '#') { cerr << "archive table of contents corrupted" << endl; return 1; }

	return 0;
}

static int decode_files(string path, string archive_file)
{
	cout << "Unpacking " << archive_file << " ... " << endl;

	// generate the archive-istream "arc"
	istream* p_arc = setup_istream(archive_file);
	if (! p_arc) return 1;
	istream& arc = *p_arc;

	// read table of contents
	array<string> name_of_file;
	array<int>    size_of_file;
	if (read_toc(arc, name_of_file, size_of_file) != 0) return 1;

	// decode the files
	for (int i=name_of_file.low(); i<=name_of_file.high(); ++i) {
		string filename = path + name_of_file[i];
		cout << "  decoding: " << filename << " ... "; cout.flush();

		// open output file
		filebuf out;
		out.open(filename, open_out);
		if (! out.is_open()) { cerr << filename << "not opened" << endl; return 1; }

		// decode file from archive
		copy_n_bytes_between_streambuffers(arc.rdbuf(), &out, size_of_file[i]);

		// check EOF-marker
		char ch;
		arc.get(ch); if (arc.fail() || ch != '#') { cerr << "archive corrupted" << endl; return 1; }
		arc.get(ch); if (arc.fail() || ch != 'L') { cerr << "archive corrupted" << endl; return 1; }
		arc.get(ch); if (arc.fail() || ch != '#') { cerr << "archive corrupted" << endl; return 1; }

		cout << "ok" << endl;
	}

	delete p_arc;
	return 0;
}


static int show_toc(string archive_file)
{
	cout << "Displaying contents of " << archive_file << " ... " << endl;

	// generate the archive-istream "arc"
	istream* p_arc = setup_istream(archive_file);
	if (! p_arc) return 1;
	istream& arc = *p_arc;

	// read table of contents
	array<string> name_of_file;
	array<int>    size_of_file;
	if (read_toc(arc, name_of_file, size_of_file) != 0) return 1;

	// display toc
	for (int i=name_of_file.low(); i<=name_of_file.high(); ++i) {
		cout << "  file: " << name_of_file[i] << " (" << size_of_file[i] << " Bytes)" << endl;
	}

	return 0;
}


enum { m_encode, m_decode, m_show_toc };

int main(int argc, char* argv[])
{
	if (argc < 2) { usage(); return 0; }

	int mode = m_encode;

	string archive_file = "leda.enc";
	string path;
	bool crypt = false;

	list<string> files_to_process;
	for (int i=1; i<argc; ++i) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case 'u':
				case 'd':	mode = m_decode; break;
				case 't':	mode = m_show_toc; break;
				case 'f':	if (++i < argc) { archive_file = string(argv[i]); break; }
							else            { cerr << "no filename after -f" << endl; return 1; }
				case 'p':	if (++i < argc) { path = string(argv[i]); break; }
							else            { cerr << "no path after -p" << endl; return 1; }
				case 'c':	crypt = true; break;
				default:	cerr << "unknown option: " << argv[i] << endl; return 1;
			}
		}
		else files_to_process.append(string(argv[i]));
	}
	if (files_to_process.empty() && mode == m_encode) mode = m_decode;

	switch (mode) {
	case m_encode:		return encode_files(files_to_process, archive_file, crypt);
	case m_decode:		return decode_files(path, archive_file);
	case m_show_toc:	return show_toc(archive_file);
	}

	return 0;
}
