#ifndef _DIRECTORY_TREE_H_
#define _DIRECTORY_TREE_H_

/// DirectoryTreeNode //////////////////////////////////////////////////////////////////////

// we assume that a node owns its children, i.e. it deletes them when it is deleted itself.

class DirectoryTreeNode {
public:
	typedef DirectoryTreeNode Node;

public:
	DirectoryTreeNode(const string& dir = "");
	virtual ~DirectoryTreeNode() { clear(); }

	string get_name() const { return Name; }
	string get_path() const;
	Node*  get_father() const { return Father; }
	const list<Node*>& get_children() const { return Children; }

	bool is_leaf() const { return Children.empty(); }
	int  get_depth() const { return Father ? 1+Father->get_depth() : 0; }

	Node* get_child(string name) const;

	Node* get_first_child() const { return is_leaf() ? 0 : Children.head(); }
	Node* get_last_child() const { return is_leaf() ? 0 : Children.tail(); }

	Node* get_left_sibling() const;
	Node* get_right_sibling() const;

	Node* get_predecessor() const;
	Node* get_successor() const;

	Node* follow_path(string path) const;

	void set_mark_of_ancestors(bool mark); // the node itself counts as ancestor too
	void set_mark_of_descendants(bool mark); // the node itself counts as descendant too

	void add_child(Node* child);
	void  remove_child(Node* child); // but does not delete the child
	Node* remove_child(string name); // does not delete the child, but returns it

	void clear_children(); // removes and deletes all children

private:
	void clear();

private:
	string                   Name;
	DirectoryTreeNode*       Father;
	list_item                ItemInFather;
	list<DirectoryTreeNode*> Children;
	bool                     Mark;
};

/// DirectoryTreeIter //////////////////////////////////////////////////////////////////////

class DirectoryTreeIter {
public:
	typedef DirectoryTreeNode Node;
	typedef DirectoryTreeIter Iter;

public:
	DirectoryTreeIter(Node* n = 0) : CurrentNode(n) {}

	Iter& operator++() { CurrentNode = CurrentNode->get_successor(); return *this; }
	Iter operator++(int) { Iter tmp = *this; ++(*this); return tmp; }

	bool valid() const { return CurrentNode != 0; }

	operator Node*() const { return CurrentNode; }
	Node& operator*() { return *CurrentNode; }
	Node* operator->() const { return CurrentNode; }

	friend bool operator==(const Iter& i1, const Iter& i2) { return i1.CurrentNode == i2.CurrentNode; }
	friend bool operator!=(const Iter& i1, const Iter& i2) { return !(i1==i2); }

private:
	Node* CurrentNode;
};

/// DirectoryTree //////////////////////////////////////////////////////////////////////////

class DirectoryTree {
public:
	typedef DirectoryTreeNode Node;
	typedef DirectoryTreeIter Iter;

public:
	DirectoryTree(const string& path_to_root);
	~DirectoryTree() { clear(); }

	Node* get_root() const { return Root; }
	string get_path_to_root() const { return PathToRoot; }
	string get_absolute_path(Node* n) const;

	Iter begin() const { return Iter(Root); }
	Iter end() const { return Iter(); }

	void rebuild();

	void print(ostream& out = cout) const;
	void remove_nodes_with_name(string name);
	list<string> find_files(string pattern) const;

	void clear_subtree(Node* subtree_root);
	void clear_subtree(string relative_path);

private:
	void rebuild_children(Node* n);
	void clear();

private:
	string PathToRoot;
	Node*  Root;
};

#endif
