#ifdef _MSC_IDE

#include <LEDA/core/d_array.h>
#include <LEDA/core/set.h>
#include <LEDA/system/file.h>

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::cerr;
using std::endl;
using std::flush;
using std::ostream;
using std::istream;
using std::ofstream;
using std::ifstream;
using std::streambuf;
using std::filebuf;
using std::ios;
#endif

using namespace leda;

#endif // ifdef _MSC_IDE

#include "LedaTester.h"
#include "HtmlReport.h"

typedef DirectoryTree::Iter Iter;
typedef DirectoryTree::Node Node;

/// static data (partially platform dependant) /////////////////////////////////////////////

#if defined(__win32__) || defined(__win64__)
#define USE_NMAKE
string TestTarget::ObjExtension = ".obj";
string TestTarget::ExeExtension = ".exe";
string LedaTester::RunPrefix;
#else
string TestTarget::ObjExtension = ".o";
string TestTarget::ExeExtension = "";
string LedaTester::RunPrefix = string(".") + get_directory_delimiter();
#endif

string TestTarget::MakeOutExtension = ".mk";
string TestTarget::MakeErrExtension = ".mke";
string TestTarget::RunInExtension = ".in";
string TestTarget::RunOutExtension = ".out";
string TestTarget::RunErrExtension = ".err";
string TestTarget::RunParamExtension = ".par";

string LedaTester::ErrRedirection = " 2> ";

#ifdef USE_NMAKE
string LedaTester::MakeCmd = "nmake -nologo";
#else
string LedaTester::MakeCmd = "make";
#endif

/// TestTarget /////////////////////////////////////////////////////////////////////////////

TestTarget::TestTarget(string name, DirectoryTree& tree, DirectoryTreeNode* dir, bool uses_GUI)
 : Status(S_None), MyTree(tree), MyDirectory(dir), UsesGUI(uses_GUI)
{
	int dot_pos = name.pos(".");
	if (dot_pos < 0) dot_pos = name.length();
	BaseName = name(0, dot_pos-1); SourceExtension = name(dot_pos, name.length()-1);
}

bool 
TestTarget::delete_file_with_check(string path)
{
	if (is_file(path)) {
		delete_file(path);
		return !is_file(path);
	}
	return true;
}

string 
TestTarget::find_run_in_file() const
{
	string run_in_file = get_run_in_path();
	if (is_file(run_in_file)) return run_in_file;
	Node* current = MyDirectory;
	string default_in_file = get_directory_delimiter() + get_default_run_in_name();
	while (current) {
		run_in_file = MyTree.get_absolute_path(current) + default_in_file;
		if (is_file(run_in_file)) return run_in_file;
		current = current->get_father();
	}
	return string();
}

string 
TestTarget::find_run_param_file() const
{
	string run_param_file = get_run_param_path();
	if (is_file(run_param_file)) return run_param_file;
	Node* current = MyDirectory;
	string default_param_file = get_directory_delimiter() + get_default_run_param_name();
	while (current) {
		run_param_file = MyTree.get_absolute_path(current) + default_param_file;
		if (is_file(run_param_file)) return run_param_file;
		current = current->get_father();
	}
	return string();
}

string 
TestTarget::read_run_parameters() const
{
	string param_file = find_run_param_file();
	if (!is_file(param_file)) return string();
	ifstream in(param_file);
	string line; line.read_line(in);
	return in ? line : string();
}

/// LedaTester /////////////////////////////////////////////////////////////////////////////

LedaTester::LedaTester(string leda_directory)
 : Verbose(true),
   ReportLowStream(0), ReportHighStream(0),
   LedaDirectory(leda_directory), 
   Tree(leda_directory + get_directory_delimiter() + "test"),
   ShowOnlyErrorsAndWarnings(true)
{
	OriginalWorkingDirectory = get_directory();
	ReportPath = Tree.get_path_to_root() + get_directory_delimiter() + "report.txt";
}

LedaTester::~LedaTester()
{
	if (ReportLowStream != ReportHighStream && ReportLowStream != &cout && ReportLowStream != &cerr) {
		delete ReportLowStream;
	}
	ReportLowStream = 0;
	if (ReportHighStream != &cout && ReportHighStream != &cerr) {
		delete ReportHighStream;
	}
	ReportHighStream = 0;
	if (DummyReportPath != "") delete_file(DummyReportPath);
	set_directory(OriginalWorkingDirectory);
}

void
LedaTester::init()
{
	init_streams();

	// remove the paths that should be ignored
	string path_to_ignore;
	forall(path_to_ignore, PathsToIgnore)
		Tree.clear_subtree(path_to_ignore);

	// gather the targets
	for (Iter it = Tree.begin(); it != Tree.end(); ++it) {
		string makefile_path = Tree.get_absolute_path(it) + get_directory_delimiter() + "Makefile";
		string make_lst_path = Tree.get_absolute_path(it) + get_directory_delimiter() + "Make.lst";
		if (!is_file(makefile_path) || !is_file(make_lst_path)) continue;

		// parse Makefile to determine whether the targets use GUI
		bool uses_GUI = false;
		ifstream in1(makefile_path);
		if (! in1.good()) { error("Cannot open " + makefile_path); continue; }
		while (in1) {
			string line; line.read_line(in1); if (!in1) break;
			if (line.pos("$(XLIB)") >= 0) uses_GUI = true;
		}

		// parse Make.lst in current directory
		string relative_path_to_dir = it->get_path() + get_directory_delimiter();
		ifstream in2(make_lst_path);
		if (! in2.good()) { error("Cannot open " + make_lst_path); continue; }
		while (in2) {
			string line; line.read_line(in2); if (!in2) break;
			int pos = line.pos("$(e)");
			if (pos < 0) continue;
			string target_base = line(0, pos-1);
			while ( (pos=target_base.pos(" ")) >= 0 ) target_base = target_base(pos+1, target_base.length()-1);
			string target_name = target_base + ".cpp";
			string target_path = Tree.get_absolute_path(it) + get_directory_delimiter() + target_name;
			if (! is_file(target_path)) { error("Could not find source file " + target_path); continue; }
			list_item lit = Targets.append( TestTarget(target_name, Tree, it, uses_GUI) );
			TestTarget& tgt = Targets[lit];

			// check for exceptions
			string relative_path = relative_path_to_dir + target_base;
			if (PathsToIgnore.member(relative_path)) {
				Targets.del_item(lit);
			}
			else if (FilesToSkip.defined(relative_path)) {
				tgt.set_run_skip();
				tgt.set_skip_comment(FilesToSkip[relative_path]);
			}
			else if (NoGUIExceptions.member(relative_path)) {
				tgt.set_GUI_flag(false);
			}
		}
	}
	if (Verbose) cout << "Found " << Targets.size() << " test targets ..." << endl;
}

void 
LedaTester::init_streams()
{
	if (!ReportHighStream) {
		ReportHighStream = new ofstream(ReportPath);
	}
	if (!ReportLowStream) {
		if (ShowOnlyErrorsAndWarnings) {
			DummyReportPath = tmp_file_name();
			ReportLowStream = new ofstream(DummyReportPath);
		}
		else {
			ReportLowStream = ReportHighStream;
		}
	}
}

ostream& 
LedaTester::report(report_level l)
{
	init_streams();

	switch (l) {
		case L_LOW:
			return *ReportLowStream;
		default:
			return *ReportHighStream;
	}
}

void
LedaTester::error(string msg)
{
	report() << "!ERROR: " << msg << endl;
}

void 
LedaTester::copy_file_to_report_stream(string file_name, report_level l)
{
	ifstream in(file_name);
	while (in) {
		string line; line.read_line(in); if (!in) break;
		report(l) << "  " << line << endl;
	}
}

void 
LedaTester::make_clean()
{
	if (Verbose) cout << "Deleting all obj and exe files ..." << endl;
	report() << endl << ">>> Make clear >>>>>>>>>>>>>>>>>>>>>>>>>>>" << endl << endl;

	list_item it;
	forall_items(it, Targets) {
		TestTarget& tgt = Targets[it];

		if (! tgt.delete_obj_file()) error("Could not delete " + tgt.get_obj_path());
		if (! tgt.delete_exe_file()) error("Could not delete " + tgt.get_exe_path());
		if (! tgt.delete_make_out_file()) error("Could not delete " + tgt.get_make_out_path());
		if (! tgt.delete_make_err_file()) error("Could not delete " + tgt.get_make_out_path());
		if (! tgt.delete_run_out_file()) error("Could not delete " + tgt.get_run_out_path());
		if (! tgt.delete_run_err_file()) error("Could not delete " + tgt.get_make_out_path());
	}

	report() << endl << "<<< Make clear <<<<<<<<<<<<<<<<<<<<<<<<<<<" << endl;
}

void 
LedaTester::make_all()
{
	// make the exe files
	if (Verbose) cout << "Making all exe files ..." << endl;
	report() << endl << ">>> Make step >>>>>>>>>>>>>>>>>>>>>>>>>>>>" << endl << endl;

	list_item it;
	forall_items(it, Targets) {
		TestTarget& tgt = Targets[it];
		make(tgt);

		if (tgt.make_error() || !tgt.make_ok()) {
			report() << "!Error(s) making " << tgt.get_exe_path() 
				            << " (see " << tgt.get_make_out_name() << ")" << endl;
		}
		else if (tgt.make_warning()) {
			report(L_MED) << "?Warning(s) making " << tgt.get_exe_path() 
				          << " (see " << tgt.get_make_out_name() << ")" << endl;
		}
		else {
			report(L_LOW) << "+Successfully made " << tgt.get_exe_path() << endl;
		}
	}

	report() << endl << "<<< Make step <<<<<<<<<<<<<<<<<<<<<<<<<<<<" << endl;
}

void 
LedaTester::parse_make_output(string file_name, TestTarget& tgt)
{
	string base_name = tgt.get_base_name();
	ifstream in(file_name);
	while (in) {
		string line; line.read_line(in); if (!in) break;
		line = line.replace_all(base_name, string()); // important! (cf. win_error.exe)
		if (line.pos("Error") >= 0) tgt.set_make_error();
		if (line.pos("error") >= 0) tgt.set_make_error();
		if (line.pos("Fehler") >= 0) tgt.set_make_error();
		if (line.pos("Warning") >= 0) tgt.set_make_warning();
		if (line.pos("warning") >= 0) tgt.set_make_warning();
		if (line.pos("Warnung") >= 0) tgt.set_make_warning();
	}
}

void
LedaTester::make(TestTarget& tgt)
{
	string target_file = tgt.get_exe_name();
	string out_file = tgt.get_make_out_name();
	string err_file = tgt.get_make_err_name();

	set_directory(tgt.get_directory());
        if (!is_file(tgt.get_source_name())) { error("Could not find " + tgt.get_source_path()); return; }

	// try to make the target_file
	string make_cmd = MakeCmd + " " + target_file + " > " + out_file;
	if (ErrRedirection != string())
		make_cmd += " 2> " + err_file;
	else
		err_file = out_file;

	if (Verbose) cout << "Making " << tgt.get_exe_path() << endl;
	int ret_val = system(make_cmd);
	if (ret_val == 0) tgt.set_make_ok();
	else              tgt.set_make_error();
	if (!is_file(target_file)) tgt.set_make_error();

	if (is_file(out_file) && size_of_file(out_file) == 0) delete_file(out_file);
	if (is_file(err_file) && size_of_file(err_file) == 0) delete_file(err_file);

	// parse out and error file
	if (out_file != err_file && is_file(out_file)) {
		parse_make_output(out_file, tgt);
	}
	if (is_file(err_file)) {
		parse_make_output(err_file, tgt);
	}
}

void 
LedaTester::run_all()
{
	// make the exe files
	if (Verbose) cout << "Running all exe files ..." << endl;
	report() << endl << ">>> Run step >>>>>>>>>>>>>>>>>>>>>>>>>>>>" << endl << endl;

	list_item it;
	forall_items(it, Targets) {
		TestTarget& tgt = Targets[it];
		if (tgt.make_error()) {
			report(L_LOW) << "?Skipped " << tgt.get_exe_path() << " (make error)" << endl;
			continue;
		}
		if (tgt.run_skip()) {
			report(L_LOW) << "?Skipped " << tgt.get_exe_path();
			if (tgt.get_skip_comment() != "") report(L_LOW) << " (" << tgt.get_skip_comment() << ")";
			report(L_LOW) << endl;
			continue;
		}
		if (!is_file(tgt.get_exe_path())) {
			report(L_LOW) << "?Skipped " << tgt.get_exe_path() << " (not found)" << endl;
			tgt.set_run_skip();
			continue;
		}
		if (tgt.uses_GUI()) {
			report(L_LOW) << "?Skipped " << tgt.get_exe_path() << " (uses GUI)" << endl;
			tgt.set_run_skip();
			continue;
		}

		run(tgt);

		if (tgt.run_error() || !tgt.run_ok()) {
			report() << "!Error(s) running " << tgt.get_exe_path();
			if (is_file(tgt.get_run_out_path()) || is_file(tgt.get_run_err_path())) 
				report() << " (see below)" << endl;
			else
				report() << " (no output)" << endl;
		}
		else {
			report(L_LOW) << "+Successfully run " << tgt.get_exe_path() << endl;
		}
	}

	report() << endl << "<<< Run step <<<<<<<<<<<<<<<<<<<<<<<<<<<<" << endl;

	// report the results of the run step
	report() << endl << ">>> Results of run >>>>>>>>>>>>>>>>>>>>>>" << endl << endl;
	forall_items(it, Targets) {
		TestTarget& tgt = Targets[it];
		string out_file = tgt.get_run_out_path();
		string err_file = tgt.get_run_err_path();
		if (tgt.run_skip() || (!is_file(out_file) && !is_file(err_file))) continue;

		// copy the out_file into the report
		if (is_file(out_file)) {
			report() << out_file;
			if (! tgt.run_ok()) report() << " (Error!)";
			report() << endl;
			copy_file_to_report_stream(out_file);
			report() << endl;
		}

		// copy the err_file into the report
		if (is_file(err_file)) {
			report() << err_file;
			if (! tgt.run_ok()) report() << " (Error!)";
			report() << endl;
			copy_file_to_report_stream(err_file);
			report() << endl;
		}

		report() << endl;
	}
	report() << endl << "<<< Results of run <<<<<<<<<<<<<<<<<<<<<<" << endl;
}

void
LedaTester::run(TestTarget& tgt)
{
	string target_file  = tgt.get_exe_name();
	string run_in_file  = tgt.find_run_in_file();
	string run_out_file = tgt.get_run_out_name();
	string run_err_file = tgt.get_run_err_name();
	set_directory(tgt.get_directory());
	if (!is_file(target_file)) { error("Could not find " + tgt.get_exe_path()); return; }

	// run the target_file
	string run_cmd = RunPrefix + target_file;
	run_cmd += string(" ") + tgt.read_run_parameters();
	if (is_file(run_in_file)) run_cmd += string(" < ") + run_in_file;
	run_cmd += string(" > ") + run_out_file;
	if (ErrRedirection != string())	run_cmd += " 2> " + run_err_file;
	
	if (Verbose) cout << "Running " << tgt.get_exe_path() << endl;
	int ret_val = system(run_cmd);
	if (ret_val == 0) tgt.set_run_ok();
	else              tgt.set_run_error();

	if (is_file(run_out_file) && size_of_file(run_out_file) == 0) delete_file(run_out_file);
	if (is_file(run_err_file) && size_of_file(run_err_file) == 0) delete_file(run_err_file);
}

void 
LedaTester::add_path_to_ignore(string relative_path)
{
	relative_path = relative_path.replace_all("\\", get_directory_delimiter());
	relative_path = relative_path.replace_all("/", get_directory_delimiter());
	PathsToIgnore.insert(relative_path);
	report(L_LOW) << "IGNORE Path: " << relative_path << endl;
}

void 
LedaTester::add_file_to_skip(string relative_path, string comment)
{
	relative_path = relative_path.replace_all("\\", get_directory_delimiter());
	relative_path = relative_path.replace_all("/", get_directory_delimiter());
	FilesToSkip[relative_path] = comment;
	report(L_LOW) << "SKIP File: " << relative_path;
	if (comment != "") report(L_LOW) << " (" << comment << ")";
	report(L_LOW) << endl;
}

void 
LedaTester::add_noGUI_exception(string relative_path)
{
	relative_path = relative_path.replace_all("\\", get_directory_delimiter());
	relative_path = relative_path.replace_all("/", get_directory_delimiter());
	NoGUIExceptions.insert(relative_path);
}

void 
LedaTester::generate_html_report()
{
	string html_path = get_test_directory() + get_directory_delimiter() + "report.html";
	generate_html_report(html_path);
}

void 
LedaTester::generate_html_report(string html_path)
{
	HtmlReport report(html_path);

	report.begin_document();

	list_item it;
	forall_items(it, Targets) {
		TestTarget& tgt = Targets[it];
		report.add_target(tgt);
	}

	report.end_document();
}
