/*******************************************************************************
+
+  LEDA 7.0  
+
+
+  gw_tutte.c
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


#include<LEDA/graphics/graphwin.h>
#include<LEDA/graph/graph_alg.h>

using namespace leda;


// ------------------------------------------------------------- //
//                    GraphWin interface                         //
// ------------------------------------------------------------- //



void fix_node(GraphWin& gw, const point&)
{ // position no used
  node v = gw.get_found_node();
  if (gw.get_color(v) != red) 
     gw.set_color(v,red);
  else
     gw.set_color(v,gw.get_node_color());
 }


void update_drawing(GraphWin& GW, const graph& G)
// redraw the graph according to the positions of the red nodes
{
  // get fixed nodes
  list<node> fixed_nodes;

  node v;
  forall_nodes(v,G) 
     if(GW.get_color(v) == red) fixed_nodes.append(v);

  node_array<double> xpos(G);
  node_array<double> ypos(G);

  forall(v,fixed_nodes) 
  { point p = GW.get_position(v);
    xpos[v] = p.xcoord();
    ypos[v] = p.ycoord();
   }

  if (TUTTE_EMBEDDING(G,fixed_nodes,xpos,ypos)) GW.set_layout(xpos,ypos);
}


// --------------------- handler ------------------------------ //


void move_node_handler(GraphWin& GW, node v)
{
  if(GW.get_color(v) != red) return;
  const graph& G = GW.get_graph();
  update_drawing(GW,G);
}

void del_node_handler(GraphWin& GW)
{
  const graph& G = GW.get_graph();
  update_drawing(GW,G);
}

void new_edge_handler(GraphWin& GW, edge e)
{
  const graph& G = GW.get_graph();
  update_drawing(GW,G);
}

void del_edge_handler(GraphWin& GW)
{
  const graph& G = GW.get_graph();
  update_drawing(GW,G);
}


void about(GraphWin& GW)
{ panel P("About Tutte");

  P.text_item("\
  \\bf\\blue Tutte Drawing Demo \\black\\tf \\c4\
  This program demonstrates the \\bf Tutte Drawing \\tf\
  procedure. Some designated nodes of the graph have fixed positions,\
  and all other nodes are drawn in the center of gravity\
  of their neighbors.\
  Edit a graph and select some nodes by clicking with the left\
  mouse button while simultaneously holding down the CTRL key.\
  The selected nodes will become red and be the fixed nodes as\
  described above. Now, move the red nodes or change the graph.\\4 \
  You get the best results for planar triangulated graph\
  ");
  P.button("ok");
  GW.open_panel(P);
}


int main()
{
  GraphWin GW("Tutte Demo");

  GW.set_action(A_LEFT|A_NODE|A_CTRL, fix_node);
  GW.add_node_menu("fix node",fix_node);

  GW.set_animation_steps(0);
  GW.set_move_node_handler(move_node_handler);
  //GW.set_end_move_node_handler(move_node_handler);
  GW.set_del_node_handler(del_node_handler);
  GW.set_new_edge_handler(new_edge_handler);
  GW.set_del_edge_handler(del_edge_handler);

  GW.display();

  int h_menu = GW.get_menu("Help");
  gw_add_simple_call(GW,about,"About Tutte Demo",h_menu);

  about(GW);
  GW.edit();

  return 0;
}
