#if !defined(LEDA_ROOT_INCL_ID)
#define LEDA_ROOT_INCL_ID 430970
#include <LEDA/PREAMBLE.h>
#endif

//----------------------------------------------------------------------------
//
// MAX_FLOW
//
// preflow-push + highest level + gap heuristic 
//
//-----------------------------------------------------------------------------

#include <LEDA/graph.h>
#include <LEDA/b_queue.h>
#include <LEDA/std/assert.h>

LEDA_BEGIN_NAMESPACE


template <class graph_t>
class node_level_queue {

typedef graph_t graph;
typedef typename graph_t::node node;
typedef typename graph_t::edge edge;

node*  T;
node*  head;
node*  max;

#if defined(LEDA_GRAPH_ARRAY1)
  node_slot<node,0,graph_t> NEXT;
#else
  node_array<node,graph_t> NEXT;
#endif

public:

node_level_queue(const graph& G)
{ 
  assert(NEXT.use_node_data(G,nil));
  int sz = G.number_of_nodes();
  T = new node[sz+1];
  for(node* p = T+sz; p >= T;  p--) *p = nil; 
  head = T+1;
  max = T;
 }

~node_level_queue() { delete[] T; }

void insert(node v, int i)
{ // insert on level i
  node* h = head + i;
  if (h > max) max = h;
  NEXT[v] = *h;
  *h = v;
 }


void insert_non_max(node v, int i)
{ // do not update max pointer
  node* h = head + i;
  NEXT[v] = *h;
  *h = v;
 }


void insert_max(node v, int i)
{ // new max at level i
  max = head + i;
  *max = v;
  NEXT[v] = 0;
 }

node del_max() 
{ while (*max == nil) max--;
  node v = *max;
  *max = NEXT[v];
  return v; 
}


void clear() { while (max >= T) *max-- =  nil; }


};




template<class NT, class graph_t>
class max_flow  
{
  typedef graph_t graph;
  typedef typename graph_t::node node;
  typedef typename graph_t::edge edge;


#if defined(LEDA_GRAPH_ARRAY1)
  typedef node_slot.<int,1,graph>  excess_array;
  typedef node_slot.<int,2,graph>  dist_array;

  typedef edge_slot.<int,0,graph>  cap_array;
  typedef edge_slot.<int,1,graph>  flow_array;
#else
  typedef node_array<int,graph>  excess_array;
  typedef node_array<int,graph>  dist_array;

  typedef edge_array<int,graph>  cap_array;
  typedef edge_array<int,graph>  flow_array;
#endif


  float h;

  NT fval;

  // statistics (phase 1, 2, and sum)

  int num_pushes[3];
  int num_relabels[3];
  int num_updates[3];
  int num_gap_nodes[3]; 
  int num_inspections[3];

  float cputime[3];


  const graph* g_p; 
  node s_p; 
  node t_p;
  const cap_array* cap_p; 
  const flow_array* flow_p;

public:

int pushes(int i=0)      { return num_pushes[i]; }
int relabels(int i=0)    { return num_relabels[i]; }
int updates(int i=0)     { return num_updates[i]; }
int gap_nodes(int i=0)   { return num_gap_nodes[i]; }
int inspections(int i=0) { return num_inspections[i]; }

float cpu_time(int i=0) { return cputime[i]; }

void set_heuristic(float hh) { h = hh; }


NT flow() { return fval; }


void reset_counters()
{ for(int i=0; i<3; i++) 
  { num_pushes[i] = 0;
    num_relabels[i] = 0;
    num_updates[i] = 0;
    num_gap_nodes[i] = 0;
    num_inspections[i] = 0;
    cputime[i] = 0;
   }
}


max_flow(float heur = 5.0)
{ reset_counters();
  fval = 0;
  h  = heur;
}


void compute_dist0(const graph& G, node s,
                   const excess_array& excess,
                   dist_array& dist, 
                   node_level_queue<graph_t>& U, int* count)
{ 
  int n = G.number_of_nodes();

  for(int i=0; i<n; i++) count[i] = 0;

  node v;
  forall_nodes(v,G)
  { dist[v] = 0;
    if (excess[v] > 0) U.insert(v,0);
   }
  count[0] = n-1;
  dist[s] = n;
}



void compute_dist1(const graph& G, node s, node t,
                   const cap_array& cap,
                   const flow_array& flow,
                   const excess_array& excess,
                   dist_array& dist,
                   node_level_queue<graph_t>& U, int* count)
{ 
  // compute exact distance values by a "backward" bfs in the
  // residual network starting at t

  num_updates[1]++;

  int n = G.number_of_nodes();

  for(int i=0; i<n; i++) count[i] = 0;

  dist.init(G,n);

  b_queue<node> Q(n);
  Q.append(t);

  dist[t] = 0;
  count[0] = 1;

  while ( !Q.empty() )
  { node v = Q.pop();
    int d = dist[v]+1;

    edge e;
    forall_out_edges(e,v)
    { if (flow[e] == 0) continue;
      node u = G.target(e,v); 
      int& du = dist[u];
      if (du >= n)
      { du = d;
        Q.append(u);
        if (excess[u] > 0) U.insert(u,d);
        count[d]++;
       }
     }

    forall_in_edges(e,v)
    { if (cap[e] == flow[e]) continue;
      node u = G.source(e,v); 
      int& du = dist[u];
      if (du >= n)
      { du = d;
        Q.append(u);
        if (excess[u] > 0) U.insert(u,d);
        count[d]++;
       }
    }
  } 

 dist[s] = n;
}



void compute_dist2(const graph& G, node s, node t,
                   const flow_array& flow,
                   const excess_array& excess,
                   dist_array& dist,
                   node_level_queue<graph_t>& U)
{ 
  // forward bfs starting in s

  num_updates[2]++;

  int n = G.number_of_nodes();

  dist.init(G,n);

  b_queue<node> Q(n);
  Q.append(s);
  dist[s] = 0;

  while ( !Q.empty() )
  { node v = Q.pop();
    int  d = dist[v];
    if (excess[v] > 0 && v != t) U.insert(v,d);
    d++;
    edge e;
    forall_out_edges(e,v)
    { if (flow[e] == 0) continue;
      node u = G.target(e,v); 
      if (dist[u] >= n)
      { dist[u] = d;
        Q.append(u);
       }
    }
  } 
}




void handle_gap(const graph& G, node v,
                const cap_array& cap,
                const flow_array& flow,
                dist_array& dist, 
                int* count)
{

  int gap_nodes = 0;

/*
  int dv = dist[v];

  node u;
  forall_nodes(u,G)
  { if (dist[u] >= n || dist[u] < dv) continue; 
    dist[u] = n;
    gap_nodes++;
  }

  num_gap_nodes[1] += gap_nodes;
*/

  int n = G.number_of_nodes();

  b_queue<node> Q(n);
  Q.append(v);

  dist[v] = n;

  while ( !Q.empty() )
  { edge e;
    node w = Q.pop(); 
    gap_nodes++;
    forall_out_edges(e,w)
    { node z = G.target(e,w);
      int& dz = dist[z];
      if (dz < n && flow[e] < cap[e])
      { Q.append(z);
        count[dz]--; 
        dz = n;
       }
    }
    forall_in_edges(e,w)
    { node z = G.source(e,w);
      int& dz = dist[z];
      if (dz < n && flow[e] > 0)
      { Q.append(z);
        count[dz]--; 
        dz = n;
       }
     }
   }

 // cout << string("gap: d = %d  sz = %d",dv,gap_nodes) << endl;

 num_gap_nodes[1] += gap_nodes;
}



void phase1(const graph& G, node s, node t,
            const cap_array& cap,
            flow_array& flow,
            excess_array& excess,
            dist_array& dist)
{ 
  int n = G.number_of_nodes();
  int m = G.number_of_edges();

  int heuristic  = (h > 0) ? int(h*m) : int(-h*m);
  int limit_heur = heuristic;

  int relabel_count = 0;
  int push_count = 0;
  int insp_count = 0;


  edge e;
  forall_out_edges(e,s) 
  { node u = G.target(e,s);
    NT c = cap[e];
    flow[e] = c;
    excess[s] -= c;
    excess[u] += c;
  }

  node_level_queue<graph_t>  U(G);
  int* count = new int[n];

  if (h < 0)
    compute_dist0(G,s,excess,dist,U,count);
  else
    compute_dist1(G,s,t,cap,flow,excess,dist,U,count);

  U.insert(t,-1); // we use t as a stopper (level -1)
  excess[t] += 1;  // prevents t from beeing inserted again

  node v;
  while ((v = U.del_max()) != t)
  {
    NT  ev = excess[v]; 
    int dv = dist[v]; 

    int dmin = n;    
  
    edge e;
    forall_out_edges(e,v)
    { insp_count++;
      NT& fe = flow[e];
      NT  rc = cap[e] - fe;
      if (rc == 0) continue;
      node w = G.target(e,v);
      int dw = dist[w];
      if (dw == dv-1)
       { push_count++;
         NT& ew = excess[w];
         if (ew == 0) U.insert_non_max(w,dw);
         if (ev <= rc) 
         { ew += ev;
           fe += ev;
           ev = 0;   // excess[v] exhausted (stop)
           break;
          }
         else
         { ew += rc;
           fe += rc;
           ev -= rc;
          }
        }
      else
        if (dw < dmin) dmin = dw;
    }
   
    excess[v] = ev;
    if (ev == 0) continue;

    forall_in_edges(e,v)
    { insp_count++;
      NT& fe = flow[e];
      if (fe == 0) continue;
      node w = G.source(e,v);
      int dw = dist[w];
      if (dw == dv-1)
       { push_count++;
         NT& ew = excess[w];
         if (ew == 0) U.insert_non_max(w,dw);
         if (ev <= fe) 
         { fe -= ev; 
           ew += ev;
           ev = 0;   // excess[v] exhausted
           break;
          }
         else
         { ew += fe;
           ev -= fe;
           fe = 0;
          }
        }
      else
        if (dw < dmin) dmin = dw;
    } 

    excess[v] = ev;
    if (ev == 0 ) continue;

    // remaining excess at v
    // relabel vertex v (i.e. update dist[v]) because all
    // admissible edges in the residual graph have been saturated 

    if (insp_count >= limit_heur)
     { U.clear();
       U.insert(t,-1);
       compute_dist1(G,s,t,cap,flow,excess,dist,U,count);
       limit_heur += heuristic;
      }
    else
     { relabel_count++;
       if (--count[dv] == 0) // gap
          handle_gap(G,v,cap,flow,dist,count);
       else
        { dv = dmin+1;
          dist[v] = dv;
          if (dv < n) 
          { count[dv]++;
            U.insert_max(v,dv);
           }
         }
      }
  }

  excess[t] -= 1;

  num_pushes[1] += push_count;
  num_relabels[1] += relabel_count;
  num_inspections[1] += insp_count;

  delete[] count;

  fval = excess[t];
}



void phase2(const graph& G, node s, node t,
            const cap_array& cap,
            flow_array& flow,
            excess_array& excess,
            dist_array& dist)
{ 
  int n = G.number_of_nodes();
  int m = G.number_of_edges();
  
  int heuristic  = (h > 0) ? int(h*m) : int(-h*m);
  int limit_heur = heuristic;

  int relabel_count = 0;
  int push_count = 0;
  int insp_count = 0;

  node_level_queue<graph_t>  U(G);

  U.insert(t,-1); // we use t as a stopper (level -1)

  compute_dist2(G,s,t,flow,excess,dist,U);

  node v;
  while ((v = U.del_max()) != t)
  {
    NT  ev = excess[v]; 
    int dv = dist[v]; 
    int dmin = n;

    edge e;
    forall_in_edges(e,v)
    { insp_count++;
      NT& fe = flow[e];
      if (fe == 0) continue;
      node w = G.source(e,v);
      int dw = dist[w];
      if (dw == dv-1)
       { push_count++;
         NT& ew = excess[w];
         if (ew == 0) U.insert_non_max(w,dw);
         if (ev <= fe) 
         { fe -= ev; 
           ew += ev;
           ev = 0; 
           break;
          }
         else
         { ew += fe;
           ev -= fe;
           fe = 0;
          }
        }
      else
        if (dw < dmin) dmin = dw;
    } 
    
    excess[v] = ev;
    if (ev == 0) continue;

    if (insp_count >= limit_heur)
      { U.clear();
        U.insert(t,-1);
        compute_dist2(G,s,t,flow,excess,dist,U);
        limit_heur += heuristic;
       }
    else
      { relabel_count++;
        dv = dmin+1;
        dist[v] = dv;
        U.insert_max(v,dv);
       }
  }

  num_pushes[2] += push_count;
  num_relabels[2] += relabel_count;
  num_inspections[2] += insp_count;
}




NT run(const graph& G, node s, node t, 
       const cap_array& cap, 
       flow_array& flow)
{ 
  if (s == t) error_handler(1,"MAXFLOW: source == sink");

  float T = used_time();

  g_p = &G;
  s_p = s;
  t_p = t;
  cap_p = &cap;
  flow_p = &flow;

  reset_counters();

  excess_array excess;
  assert(excess.use_node_data(G,0));

  dist_array dist;
  assert(dist.use_node_data(G,0));

  flow.init(G,0);

  phase1(G,s,t,cap,flow,excess,dist);
  cputime[1] = used_time(T);

  phase2(G,s,t,cap,flow,excess,dist);
  cputime[2] = used_time(T);

  num_pushes[0]      = num_pushes[1]      + num_pushes[2];
  num_relabels[0]    = num_relabels[1]    + num_relabels[2];
  num_updates[0]     = num_updates[1]     + num_updates[2];
  num_gap_nodes[0]   = num_gap_nodes[1]   + num_gap_nodes[2];
  num_inspections[0] = num_inspections[1] + num_inspections[2];

  cputime[0] = cputime[1] + cputime[2];

  return fval;
}


NT operator()(const graph& G, node s, node t, 
              const cap_array& cap, flow_array& flow) 
{ 
  return run(G,s,t,cap,flow); 
 }




void statistics(ostream& out)
{ 
  out << endl;

  for(int i=1; i<=2; i++)
  { out << 
    string("%8d pushes %7d relabels %3d updates %9d insp %5d gaps  %.2f s",
            pushes(i), relabels(i), updates(i), inspections(i), gap_nodes(i),
            cputime[i]) << endl;
  }

  out << endl;
}




bool check()
{
  const graph& G = *g_p;
  node s = s_p; 
  node t = t_p;
  const cap_array&  cap  = *cap_p;
  const flow_array& flow = *flow_p;

  int n = G.number_of_nodes();

  string loc = "max_flow::check: ";
  bool res = true;

  edge e;
  forall_edges(e,G) 
    res = res && leda_assert(flow[e] >= 0 && flow[e] <= cap[e], 
              loc + string("illegal flow value for edge %d",G.index(e)),1);
  
  node_array<NT,graph_t> excess(G,0);

  node v;
  forall_nodes(v,G) 
  { edge e;
    forall_out_edges(e,v) 
    { node w = G.target(e,v);
      excess[v] -= flow[e]; 
      excess[w] += flow[e];
     }
  }

  assert(fval == excess[t]);

  
  node_array<bool,graph_t> reached(G,false);
  reached[s] = true;

  b_queue<node> Q(n);
  Q.append(s); 

  while ( !Q.empty() )
  { node v = Q.pop(); 
    forall_out_edges(e,v) 
    { node w = G.target(e,v);
      if ( flow[e] < cap[e] && !reached[w] ) 
      { reached[w] = true; 
        Q.append(w); 
       }
    }
    forall_in_edges(e,v) 
    { node w = G.source(e,v);
      if ( flow[e] > 0 && !reached[w] ) 
      { reached[w] = true; 
        Q.append(w); 
       }
    }
  }
  res = res && leda_assert(!reached[t],"t is reachable in G_f",1);

  forall_nodes(v,G)
    res = res && leda_assert(v == s || v == t || excess[v] == 0,
                 loc + string("node %d has non-zero excess",G.index(v)),1);

  return res;
}
 
};

LEDA_END_NAMESPACE


#if LEDA_ROOT_INCL_ID == 430970
#undef LEDA_ROOT_INCL_ID
#include <LEDA/POSTAMBLE.h>
#endif


