#define LEDA_CHECKING_OFF

#include <LEDA/graph.h>
#include <LEDA/graph_alg.h>
#include <LEDA/dimacs.h>
#include <LEDA/node_list.h>
#include <LEDA/b_queue.h>
#include <fstream.h>
#include <assert.h>


void init_levels(graph& G, node s, node t, node_list*& level, 
                                           node_array<int>& dist)
{ 
  int n = G.number_of_nodes();

  level = new node_list[n];

  node_list* current_level = level;
  int level_num = 0;

  dist[t] = 0;
  current_level->push(t);

  while ( !current_level->empty() )
  { 
    node_list* next_level = current_level+1;

    level_num++;

    node v;
    forall(v,*current_level)
    { edge e;
      for(e = G.first_in_edge(v); e; e = G.in_succ(e))
      { node u = source(e); 
        if (u == s || dist[u] < n) continue;
        dist[u] = level_num;
        next_level->push(u);
       }
     }
    current_level = next_level;
  } 

}




template<class EdgeArray>
void compute_dist1(graph& G, node t, EdgeArray& flow, const EdgeArray& cap, 
                                                      node_array<int>& dist)
{ 
  // compute exact distance values by a "backward" bfs in the
  // residual network starting at t

  int n = G.number_of_nodes();

  b_queue<node> Q(n);
  Q.append(t);

  dist[t] = 0;

  while ( ! Q.empty() )
  { node v = Q.pop();
    int  d = dist[v] + 1;
    edge e;

    for(e = G.first_adj_edge(v); e; e = G.adj_succ(e) )
    { if (flow[e] == 0) continue;
      node u = target(e); 
      if (dist[u] >= n)
      { dist[u] = d;
        Q.append(u);
       }
     }

    for(e = G.first_in_edge(v); e; e = G.in_succ(e) )
    { if (cap[e] == flow[e]) continue;
      node u = source(e); 
      if (dist[u] >= n)
      { dist[u] = d;
        Q.append(u);
       }
     }
   } 

}


template<class EdgeArray>
void compute_dist2(const graph& G, node s, const EdgeArray& flow, 
                                           node_array<int>& dist)
{ // forward bfs starting in s
  // (when excess flows back to s) 

  int n = G.number_of_nodes();

  b_queue<node> Q(n);
  Q.append(s);
  dist[s] = n;

  while (! Q.empty() )
  { node v = Q.pop();
    int  d = dist[v] + 1;
    edge e;
    for(e = G.first_adj_edge(v); e; e = G.adj_succ(e) )
    { if (flow[e] == 0) continue;
      node u = target(e); 
      if (dist[u] == MAXINT)
      { dist[u] = d;
        Q.append(u);
       }
     }
   } 
}



//------------------------------------------------------------------------------
// level queue data structure
// used to implement (max) priority queue of nodes with positive excess 
// ordered by levels (dist values) 
//
// level_queue Q(max_level);
//
// void Q.insert(node v, int d)    insert v with level d (update max)
// void Q.insert0(node v, int d)   d is not greater than current max
// node Q.del()                    remove and return a node from max level
//                                 (nil if Q is empty)
// void Q.clear()                  
//------------------------------------------------------------------------------


// dummy graph for (mis)using node-info fields

static GRAPH<node,edge>* NEXT;


class level_queue {

const graph* gr;

node*  head;
node*  stop;
node*  max;


// array for saving and restoring original node infos
node* save;

public:


level_queue(const graph& G, int sz)
{ gr = &G;
  head = new node[sz];
  save = new node[G.number_of_nodes()];
  max  = head;
  stop = head + sz;
  node* p = head;
  while(p < stop) *p++ = nil; 
  node  v;
  p = save;
  forall_nodes(v,G) *p++ = (*NEXT)[v];
 }

~level_queue() 
 { node  v;
   node* p = save;
   forall_nodes(v,*gr) (*NEXT)[v] = *p++;
   delete[] save;
   delete[] head; 
  }


void insert0(node v, int i)
{ // insert without updating max pointer
   node* h = head + i;
  (*NEXT)[v] = *h;
  *h = v;
 }

void insert(node v, int i)
{ // insert and update max pointer
  node* h = head + i;
  if (h > max) max = h;
  (*NEXT)[v] = *h;
  *h = v;
 }

node del_max()
{ // remove and return maximum (nil if list is empty) 
  while (*max == nil && max > head) max--; 
  node v = *max;
  if (v != nil) *max = (*NEXT)[v];
  return v;
}

node find_max()
{ // return maximum (nil if list is empty) 
  while (*max == nil && max > head) max--; 
  return *max;
}

node del() { return del_max(); }


void clear(int i=0) 
{ // erase everything on and above level i
  node* p = head + i;
  while (max >= p) *max-- =  nil; 
  if (max < head) max = head;
}

int raise(int i, int n)
{ // raises all nodes in levels i to n-1 to level n
  // returns number of raised nodes
  int count = 0;
  node* pn = head + n;
  for(node* pi = head+i; pi < pn; pi++)
  { node q = *pi;
    node last = q;
    while (q) 
    { last = q;
      q = (*NEXT)[q];
      count++;
     }
    if (last)
    { (*NEXT)[last] = *pn;
      *pn = *pi;
     }
    *pi = 0;
   }
  if (*pn && max < pn) max = pn;
  return count;
}


};



template<class NT>
NT MAX_FLOWS1(graph& G, node s, node t, const edge_array<NT>& cap, 
                                        edge_array<NT>& flow, int H,
                                        int& num_pushes,
                                        int& num_relabels,
                                        int& num_updates,
                                        int& num_gaps,
                                        int& num_gap_nodes)
{

float T = used_time();

int n = G.number_of_nodes();
int m = G.number_of_edges(); 


// use free data slots for node_arrays if available

node_array<int> dist;
dist.use_node_data(G,n);

node_array<NT> excess;
excess.use_node_data(G,0);

//flow.init(G,0);
edge e;
forall_edges(e,G) flow[e] = 0;


// parameter for heuristic suggested by Goldberg to speed up algorithm 
// compute exact distance labels after every "heuristic" relabel steps

int heuristic = (H > 0) ? H : n;

int limit_heur   = heuristic;

num_pushes   = 0;
num_relabels = 0;
num_updates  = 0;
num_gaps     = 0;
num_gap_nodes= 0;


if (s == t) error_handler(1,"MAXFLOW: source == sink");

//cout << endl;
//cout << string("preparations: %.2f sec",used_time(T)) << endl;

level_queue  U(G,2*n);

node_list* level;
init_levels(G,s,t,level,dist);

//cout << string("compute_dist: %.2f sec",used_time(T)) << endl;

// saturate all edges leaving from s

forall_out_edges(e,s) 
{ NT c = cap[e];
  if (c == 0) continue;
  node v = target(e);
  if (excess[v] == 0) U.insert(v,dist[v]);
  flow[e] = c;
  excess[s] -= c;
  excess[v] += c;
 }


for(;;)  // main loop
{
 node v = U.del();

 if (v == nil) break;  // empty queue

 if (v == t) continue;

 NT  ev = excess[v]; // temporary excess of v 
 int dv = dist[v];   // level of v

 int dmin = MAXINT;  // will contains minimum level of nodes adjacent
                     // in residual network if ev > 0 after push step

 // push excess ev out of v (invariant: ev > 0)

 if (dv < n) // do not use outgoing arcs when excess flows back to source
 { // forall_adj_edges(e,v)
   edge e;
   for (e = G.first_adj_edge(v); e; e = G.adj_succ(e))
   { NT& fe = flow[e];
     NT  rc = cap[e] - fe;
     if (rc == 0) continue;
     node w = target(e);
     int dw = dist[w];
     if (dw == dv-1)
      { num_pushes++;
        NT& ew = excess[w];
        if (ew == 0) U.insert0(w,dw);
        if (ev <= rc) 
        { ew += ev;
          fe += ev;
          ev = 0;   // stop: excess[v] exhausted
          break;
         }
        else
        { ew += rc;
          fe += rc;
          ev -= rc;
         }
       }
     else
       if (dw < dmin) dmin = dw;
    }
  }



  if (ev > 0)  // stop if excess[v] == 0
  { // forall_in_edges(e,v)
    edge e;
    for (e = G.first_in_edge(v); e; e = G.in_succ(e))
    { NT& fe = flow[e];
      if (fe == 0) continue;
      node w = source(e);
      int dw = dist[w];
      if (dw == dv-1)
       { num_pushes++;
         NT& ew = excess[w];
         if (ew == 0) U.insert0(w,dw);
         if (ev <= fe) 
         { fe -= ev; 
           ew += ev;
           ev = 0;   // stop: excess[v] exhausted
           break;
          }
         else
         { ew += fe;
           ev -= fe;
           fe = 0;
          }
        }
      else
        if (dw < dmin) dmin = dw;
    } 
  }

  
  excess[v] = ev;

  if (ev > 0)  
  { 
    // remaining excess
    // relabel vertex v (i.e. update dist[v]) because all
    // admissible edges in the residual graph have been saturated 

    if (num_relabels <= limit_heur)
     { num_relabels++;
       dist[v] = ++dmin; 
       U.insert(v,dmin);

       if (dv < n) 
       { node_list* lv = level+dv;
         node_list* lstop = level+n;
         lv->del(v);
         if (dmin < n) level[dmin].push(v);

         if (lv->empty()) //gap
         { num_gaps++;
           U.raise(dv,n);
           while (lv < lstop)
           { node u;
             forall(u,*lv) 
             { dist[u] = n;
               num_gap_nodes++;
              }
             lv->clear();
             lv++;
            }
          }

        }
      }
    else
      { // heuristic suggested by Goldberg to reduce number of relabels:
        // periodically compute exact dist[] labels by a backward bfs 
        // starting at t
    
        limit_heur += heuristic;
    
        dist.use_node_data(G,MAXINT);
        compute_dist1(G,t,flow,cap,dist);
        compute_dist2(G,s,flow,dist);
        num_updates++;
    
        U.clear();

        for(int i=0; i<n; i++) level[i].clear();

        node v;
        forall_nodes(v,G)
        { int dv = dist[v]; 
          if (excess[v] > 0) U.insert(v,dv);  // excess[s] < 0
          if (dv < n) level[dv].push(v);
         }
      }
   } 
    
} // end of main loop
       

/*

// some checkings

node v;
forall_nodes(v,G) 
{ if (excess[v] > 0 && v != t) 
      error_handler(1,"MAX_FLOW: positive excess in node != t");
  if (excess[v] < 0 && v != s) 
      error_handler(1,"MAX_FLOW: negative excess in node != s");
}

if (-excess[s] != excess[t]) 
  error_handler(1,"MAX_FLOW: demand(s) != excess(t).");
*/


// cout << endl;
// cout << string("%5d pushes %5d relabels %5d updates %5d gaps %5d gap nodes",
//                num_pushes,num_relabels,num_updates,num_gaps,num_gap_nodes);
// cout << endl;
// cout << endl;

delete[] level;

return excess[t];  // value of maximum flow from s to t
  
}




template<class NT>
NT MAX_PREFLOW(graph& G, node s, node t, const edge_array<NT>& cap, 
                                         edge_array<NT>& flow, int H)
{

float T = used_time();

int n = G.number_of_nodes();
int m = G.number_of_edges(); 

// use free data slots for node_arrays if available

node_array<int> dist;
dist.use_node_data(G,n);

node_array<NT> excess;
excess.use_node_data(G,0);

flow.init(G,0);


// parameter for heuristic suggested by Goldberg to speed up algorithm 
// compute exact distance labels after every "heuristic" relabel steps

int heuristic = (H > 0) ? H : n;

int limit_heur   = heuristic;
int num_relabels = 0;
int num_pushes   = 0;
int num_updates  = 0;
int num_gaps     = 0;
int num_gap_nodes= 0;


if (s == t) error_handler(1,"MAXFLOW: source == sink");

if (G.outdeg(s) == 0) return 0;



cout << endl;
cout << string("preparations: %.2f sec",used_time(T)) << endl;

node_list* level;
init_levels(G,s,t,level,dist);

cout << string("compute_dist: %.2f sec",used_time(T)) << endl;

level_queue  U(G,n);

edge e;
forall_out_edges(e,s) 
{ NT c = cap[e];
  if (c == 0) continue;
  node v = target(e);
  if (excess[v] == 0 && dist[v] < n) U.insert(v,dist[v]);
  flow[e] = c;
  excess[s] -= c;
  excess[v] += c;
 }

cout << string("init queue:   %.2f sec",used_time(T)) << endl;

for(;;)  // main loop
{
 node v = U.del();

 if (v == nil) break;  // empty queue

 if (v == t) continue;

 NT  ev = excess[v]; // temporary excess of v 
 int dv = dist[v];   // level of v

 int dmin = MAXINT;  // will contains minimum level of nodes adjacent
                     // in residual network if ev > 0 after push step

 // push excess ev out of v (invariant: ev > 0)

 assert(ev > 0);
 assert(dv < n);

 // forall_adj_edges(e,v)
 edge e;
 for (e = G.first_adj_edge(v); e; e = G.adj_succ(e))
 { NT& fe = flow[e];
   NT  rc = cap[e] - fe;
   if (rc == 0) continue;
   node w = target(e);
   int dw = dist[w];
   if (dw == dv-1)
    { num_pushes++;
      NT& ew = excess[w];
      if (ew == 0) U.insert0(w,dw);
      if (ev <= rc) 
      { ew += ev;
        fe += ev;
        ev = 0;   // stop: excess[v] exhausted
        break;
       }
      else
      { ew += rc;
        fe += rc;
        ev -= rc;
       }
     }
   else
     if (dw < dmin) dmin = dw;
  }



  if (ev > 0)  // stop if excess[v] == 0
  { // forall_in_edges(e,v)
    edge e;
    for (e = G.first_in_edge(v); e; e = G.in_succ(e))
    { NT& fe = flow[e];
      if (fe == 0) continue;
      node w = source(e);
      int dw = dist[w];
      if (dw == dv-1)
       { num_pushes++;
         NT& ew = excess[w];
         if (ew == 0) U.insert0(w,dw);
         if (ev <= fe) 
         { fe -= ev; 
           ew += ev;
           ev = 0;   // stop: excess[v] exhausted
           break;
          }
         else
         { ew += fe;
           ev -= fe;
           fe = 0;
          }
        }
      else
        if (dw < dmin) dmin = dw;
    } 
  }

  
  excess[v] = ev;

  if (ev > 0)  
  { 
    // remaining excess
    // relabel vertex v (i.e. update dist[v]) because all
    // admissible edges in the residual graph have been saturated 

    if (num_relabels <= limit_heur)
     { num_relabels++;
       dist[v] = ++dmin; 
       if (dmin < n) U.insert(v,dmin);

       if (dv < n) 
       { node_list* lv = level+dv;
         lv->del(v);
         if (dmin < n) level[dmin].push(v);

         if (lv->empty()) // gap on level dv
         { num_gaps++;
           U.clear(dv);
           node_list* lstop = level+n;
           while (++lv < lstop)
           { node u;
             forall(u,*lv) 
             { dist[u] = n;
               num_gap_nodes++;
              }
             lv->clear();
            }
          }

        }
      }
    else
      { // heuristic suggested by Goldberg to reduce number of relabels:
        // periodically compute exact dist[] labels by a backward bfs 
        // starting at t
    
        limit_heur += heuristic;
    
        dist.use_node_data(G,n);
        compute_dist1(G,t,flow,cap,dist);
        num_updates++;
    
        U.clear();

        for(int i=0; i<n; i++) level[i].clear();

        node v;
        forall_nodes(v,G)
        { int dv = dist[v]; 
          if (dv >= n) continue;
          if (excess[v] > 0) U.insert(v,dv);  // excess[s] < 0
          level[dv].push(v);
         }
      }
   } 
    
} // end of main loop
       

/*

node v;
forall_nodes(v,G) 
{ if (excess[v] > 0 && v != t) 
      error_handler(1,"MAX_FLOW: positive excess in node != t");
  if (excess[v] < 0 && v != s) 
      error_handler(1,"MAX_FLOW: negative excess in node != s");
}

if (-excess[s] != excess[t]) 
  error_handler(1,"MAX_FLOW: demand(s) != excess(t).");
*/

// cout << endl;
// cout << string("%5d pushes %5d relabels %5d updates %5d gaps %5d gap nodes",
//                 num_pushes,num_relabels,num_updates,num_gaps,num_gap_nodes);
// cout << endl;
// cout << endl;

delete[] level;

return excess[t];  // value of maximum flow from s to t
  
}






int main(int argc, char** argv)
{
  int node_slots = 2;
  int edge_slots = 1;

  GRAPH<int,int> G(node_slots,edge_slots);


  if (argc > 1)  // file argument
      G.read(argv[1]);
  else
    { int n = read_int("|V| = ");
      int m = read_int("|E| = ");
      rand_int.set_seed(n*m);
      random_graph(G,n,m);
     }

  edge_array<int>& cap = G.edge_data();

  edge_array<int> flow;
  flow.use_edge_data(G,0);


int n = G.number_of_nodes();
int m = G.number_of_edges();

cout << endl;
cout << "nodes: " << n << endl;
cout << "edges: " << m << endl;
cout << endl;


node s = G.first_node();
node t = G.last_node();

if (argc < 2)
{ edge e;
  forall_edges(e,G) cap[e] = rand_int(1,1000);
  //G.write("flow_test.lgr");
 }

cout << endl;
cout << "Saving maxflow problem to flow.dimacs (DIMACS format) ... " << flush;
ofstream out("flow.dimacs");
Write_Dimacs_Maxflow(out,G,cap,s,t);
cout << endl;
cout << endl;

int pushes,relabels,updates,gaps,gap_nodes;

float T;
int f;


cout << "MAX_FLOW    " << flush;
T = used_time();
f = MAX_FLOW(G,s,t,cap,flow) ;
cout << string("time: %6.2f sec  f = %d",used_time(T),f) << endl;
cout << endl;

CHECK_MAX_FLOW(G,s,t,cap,flow);


cout << "MAX_FLOWS   " << flush;
T  = used_time();
f = MAX_FLOWS(G,s,t,cap,flow,0,pushes,relabels,updates,gaps,gap_nodes) ;
cout << string("time: %6.2f sec  f = %d",used_time(T),f) << endl;
cout  << endl;
cout << "            " << flush;
cout << string("pushes: %d relabels: %d updates: %d gaps: %d gap nodes: %d",
                pushes, relabels, updates, gaps, gap_nodes) << endl;
cout << endl;

CHECK_MAX_FLOW(G,s,t,cap,flow);



cout << "MAX_FLOWS1  " << flush;
T  = used_time();
f = MAX_FLOWS1(G,s,t,cap,flow,0,pushes,relabels,updates,gaps,gap_nodes) ;
cout << string("time: %6.2f sec  f = %d",used_time(T),f) << endl;
cout  << endl;
cout << "            " << flush;
cout << string("pushes: %d relabels: %d updates: %d gaps: %d gap nodes: %d",
                pushes, relabels, updates, gaps, gap_nodes) << endl;
cout << endl;

CHECK_MAX_FLOW(G,s,t,cap,flow);



cout << "PRE_FLOW   " << flush;
T = used_time();
f = MAX_PREFLOW(G,s,t,cap,flow,n) ;
cout << string("time: %6.2f sec  f = %d",used_time(T),f) << endl;
//CHECK_MAX_FLOW(G,s,t,cap,flow);
cout << endl;

int count1 = 0;
edge e;
forall_edges(e,G)
   if (flow[e] > 0) count1++;
cout << "edges with flow > 0 : " << count1 << endl;
cout << endl;

return 0;

}
