#include <LEDA/graph/graph.h>
#include <LEDA/core/map.h>

#include <LEDA/geo/d3_rat_point.h>
#include <LEDA/geo/d3_hull.h>
#include <LEDA/graphics/d3_window.h>


#include <assert.h>
#include <math.h>


using namespace leda;

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::cin;
using std::endl;
using std::flush;
using std::ofstream;
using std::ifstream;
#endif

static window W(550,650,"3D Convex Hull Demo");

static d3_window* pa;

static int  mode  = 0;
static bool trace = false;
static bool step  = false;
static bool check = false;
static bool filter = true;
static bool moving = false;

static int maxcoord = 1000;
static int anim_speed = 12;


static int vis_count = 1;


inline bool equal_xy(const d3_rat_point& p, const d3_rat_point& q)
{ return p.XD()*q.WD() == q.XD()*p.WD() && p.YD()*q.WD() == q.YD()*p.WD(); }

#define ORIENT(p,e) orientation_xy(p,G[source(e)], G[target(e)])


void TRIANGULATE(GRAPH<d3_rat_point,int>& G, list<node>& L, int orient)
{ 
  if (L.empty()) return;

  node          last_v = L.pop();
  d3_rat_point  last_p = G[last_v];

  while (!L.empty() && equal_xy(last_p,G[L.head()])) 
  { node v = L.pop();
    G.del_node(v);
   }

  if (!L.empty())
  { node v = L.pop();
    last_p = G[v];
/*
    edge x = G.new_edge(last_v,v,0,leda::after);
    edge y = G.new_edge(v,last_v,0,leda::after);
*/
    edge x = G.new_edge(last_v,v,0);
    edge y = G.new_edge(v,last_v,0);
    G[x] = G[y] = black;
    G.set_reversal(x,y);
    last_v = v;
   }

  // scan remaining points

  node v;
  forall(v,L) 
  { 
    d3_rat_point p = G[v];
    
    if (equal_xy(p,last_p)) 
    { G.del_node(v);
      continue; 
     }


    // walk up to upper tangent
    edge e = G.last_edge();
    do e = G.pred_face_edge(e); while (ORIENT(p,e) == orient);

    // walk down to lower tangent and triangulate
    do { edge succ_e = G.succ_face_edge(e);
         edge x = G.new_edge(succ_e,v,0,leda::after);
/*
         edge y = G.new_edge(v,source(succ_e),0,leda::after);
*/
         edge y = G.new_edge(v,source(succ_e),0);
         G[x] = G[y] = red;
         G.set_reversal(x,y);
         e = succ_e;
       } while (ORIENT(p,e) == orient);

    last_p = p;
   }

   edge e_hull = G.last_edge();
   edge e = e_hull;
   do { edge r = G.reversal(e);
        G[e] = G[r] = blue;
        e = G.face_cycle_succ(e);
   } while (e != e_hull);

}

void init_hull(list<d3_rat_point>& L, GRAPH<d3_rat_point,int>& H1,
                                      GRAPH<d3_rat_point,int>& H2)
{  L.sort();

   GRAPH<d3_rat_point,int> H;
   list<node> V;
   d3_rat_point p;
   forall(p,L) V.append(H.new_node(p));
   TRIANGULATE(H,V,+1);

   d3_rat_point min_z = H[H.first_node()];
   d3_rat_point max_z = min_z;

   list<d3_rat_point> L1;

   edge e_hull = H.last_edge();
   edge e = e_hull;
   do { d3_rat_point p = H[source(e)];
        if (p.ZD()*min_z.WD() < min_z.ZD()*p.WD()) min_z = p;
        if (p.ZD()*max_z.WD() > max_z.ZD()*p.WD()) max_z = p;
        e = H.face_cycle_succ(e);
   } while (e != e_hull);


   H1.clear();
   H2.clear();

   list<node> V1;
   list<node> V2;

   forall(p,L)
   { if (p.ZD()*max_z.WD() <= max_z.ZD()*p.WD()) V1.append(H1.new_node(p));
     if (p.ZD()*min_z.WD() >= min_z.ZD()*p.WD()) V2.push(H2.new_node(p));
    }

   TRIANGULATE(H1,V1,+1);
   TRIANGULATE(H2,V2,-1);
}




int convex(const GRAPH<d3_rat_point,int>& G, edge e) 
{ edge r = G.reversal(e);
  edge e1 = G.face_cycle_succ(r);
  edge e2 = G.face_cycle_succ(e);
  d3_rat_point a = G[source(e1)];
  d3_rat_point b = G[target(e1)];
  d3_rat_point c = G[source(e2)];
  d3_rat_point d = G[target(e2)];
  return orientation(a,b,c,d);
}

void check_colors(string s, const GRAPH<d3_rat_point,int>& G) 
{ edge e;
  forall_edges(e,G)
  { if (G[e] == blue) continue;
    int c = convex(G,e);
    if (c < 0 && G[e] == red)   cout << s << ": convex edge is red" << endl;
    if (c > 0 && G[e] == black) cout << s << ": non-convex edge is black" << endl;
  }
}
  
void compute_color(GRAPH<d3_rat_point,int>& G, edge e) 
{ if (G[e] == blue) return;
  edge r = G.reversal(e);
  int d = convex(G,e);
  G[e] =  G[r] = (d == 1) ? red : black;
 }

bool check_hull(GRAPH<d3_rat_point,int>& H, const list<d3_rat_point>& L)
{
   edge_array<edge> rev(H,nil);

   { node v;
     forall_nodes(v,H)
     { edge e;
       bool hull = false;
       if (H.outdeg(v) == 0) continue;
       forall_adj_edges(e,v)
          if (H[e] == blue) hull = true;
       if (!hull && H.outdeg(v) < 3)
       { error_handler(0,"non hull node with degree < 3");
         break;
        }
      }
    }

   if (!Is_Bidirected(H,rev))
   { edge x;
     forall_edges(x,H)
     { if (rev[x] == nil) 
         { cout << "no reversal: "; H.print_edge(x);
           cout << endl;
          }
      }
     error_handler(0,"check_hull: H is not bidirected");
    }

   if (!Is_Map(H))
   { edge x;
     forall_edges(x,H)
     { edge r = H.reversal(x);
       if (H.source(x) != H.target(r) || H.target(x) != H.source(r))
         { cout << "wrong reversal: "; H.print_edge(x);H.print_edge(r);
           cout << endl;
          }
      }
     error_handler(0,"check_hull: H is not a map");
    }


   if (!Is_Planar_Map(H))
   { H.print();
     cout << endl;
     cout << H.number_of_nodes() << endl;
     cout << endl;
     cout << "Genus: " << Genus(H) << endl;
     error_handler(0,"check_hull: H is not a planar map");
    }


   rat_vector c_vec(0,0,0,1);
   int n = 0;

   node v;
   forall_nodes(v,H) 
     if (H.degree(v) > 0) 
     { c_vec += H[v].to_vector();
       n++;
      }

   c_vec /= n;

   int orient_err = 0;
   int convex_err = 0;

   edge e;
   forall_edges(e,H) 
   { if (H[e] == blue) continue;
     H[e] = black;

     if (convex(H,e) > 0) 
     { H[e] = red;
       convex_err++;
      }

     }

   forall_edges(e,H)
      if (H.face_cycle_succ(H.face_cycle_succ(e)) == e)
      { cout << " face with 2 edges: ";
        H.print_edge(e);
        cout << endl;
        H[e] = H[H.reversal(e)] = green;
        edge x = H.face_cycle_succ(e);
        H[x] = H[H.reversal(x)] = green;
       }

   if (orient_err > 0 || convex_err > 0)
   { error_handler(0,string("%d orientation errors  %d convexity errors",
                                 orient_err,convex_err));
     return false;
   }

   forall_nodes(v,H)
   { d3_rat_point p = H[v];
     edge_array<bool> visited(H,false);
     edge e;
     forall_edges(e,H)
     { if (H[e]==blue) continue;
       if (visited[e]) continue;
       edge e1 = H.face_cycle_succ(e);
       edge e2 = H.face_cycle_succ(e1);
       visited[e] = visited[e1] = visited[e2] = true;
       d3_rat_point a = H[source(e)];
       d3_rat_point b = H[source(e1)];
       d3_rat_point c = H[source(e2)];
       if (orientation(a,b,c,p) > 0) 
       { cout << "point out of hull:  ";
         H.print_node(v);
         cout << endl;
         break;
        }
      }
    }

   return (orient_err == 0 && convex_err == 0);
}
     


int FLIPPING(GRAPH<d3_rat_point,int>& G,list<edge>& S)
{
  if (G.number_of_nodes() <= 3) return 0;

  int flip_count = 0;

  list<edge> S1;

  while ( !S.empty() )
  { 
    edge e = S.pop();
    edge r = G.reversal(e);

    if (G[e] == blue) continue;

    edge e1 = G.face_cycle_succ(r);
    edge e3 = G.face_cycle_succ(e);

    d3_rat_point a = G[source(e1)];
    d3_rat_point b = G[target(e1)];
    d3_rat_point c = G[source(e3)];
    d3_rat_point d = G[target(e3)];

    G[e] = G[r] = black;

    if (orientation(a,b,c,d) > 0) 
    { int orient_bda = orientation_xy(b,d,a);
      int orient_bdc = orientation_xy(b,d,c);
      if (orient_bda != orient_bdc
      && (orient_bda != 0 || G.outdeg(source(e1)) == 4 || G[e1] == blue)
      && (orient_bdc != 0 || G.outdeg(source(e3)) == 4 || G[e3] == blue))
        { 
          edge e2 = G.face_cycle_succ(e1);
          edge e4 = G.face_cycle_succ(e3);
    
          S.push(e1); 
          S.push(e2); 
          S.push(e3); 
          S.push(e4); 
      
          // flip
          G.move_edge(e,e2,source(e4));
          G.move_edge(r,e4,source(e2));
          if ((orient_bda==0 && G[e1]==blue) || (orient_bdc==0 && G[e3]==blue))
             G[e] = G[r] = blue;
          else
             G[e] = G[r] = black;
          flip_count++;
        }
      else 
       { S1.append(e);
         G[e] = G[r] = red;
        }
    }

  }

  //check_colors("FLIPPING",G);

  edge e;
  forall(e,S1)
     if (G[e] == red) S.append(e);

  return flip_count;
}



void simplify_face(GRAPH<d3_rat_point,int>& G, node v, list<edge>& E)
{ 
  // flip edges adjacent to v until no flips possible anymore or degree(v)
  // == 3 appends all flipped edges to E 

  if (G.outdeg(v) == 3) return;

  // e1,e2,e3 :  three consecutive edges incident to v
  edge e1 = G.first_adj_edge(v);
  edge e2 = G.cyclic_adj_succ(e1);
  edge e3 = G.cyclic_adj_succ(e2);

  // count : number of traversed edges since last flip
  int count = 0; 

  while (count++ < G.outdeg(v) && G.outdeg(v) > 3)
  { int orient1 = orientation_xy(G[target(e1)],G[target(e3)],G[source(e2)]);
    int orient2 = orientation_xy(G[target(e1)],G[target(e3)],G[target(e2)]);
    if (orient1 != orient2 && orient2 != 0)
      { edge r2 = G.reversal(e2);
        G.move_edge(e2,G.reversal(e1),target(e3),leda::before);
        G.move_edge(r2,G.reversal(e3),target(e1));
        E.append(e2);
        count = 0;
       }
    else 
       e1 = e2;

    e2 = e3;
    e3 = G.cyclic_adj_succ(e2);
  }
}


int CLIPPING(GRAPH<d3_rat_point,int>& G, const list<edge>& E) 
{ 
  node_list L;

  edge e;
  forall(e,E)
  { assert(G[e] == red);
    node v = source(e);
    node w = target(e);
    if (!L.member(v)) L.append(v);
    if (!L.member(w)) L.append(w);
   }

  int clip_count = 0;

  node v;
  forall(v,L)
  { int rdeg = 0;
    list<edge> E;
    edge e;
    forall_adj_edges(e,v)
    { if (G[e] == red)  rdeg++;
      E.append(G.face_cycle_succ(e));
     }
    if (rdeg == G.outdeg(v))
    { if (G.outdeg(v) > 3) simplify_face(G,v,E);
      L.del(v);
      G.del_node(v);
      clip_count++;
      forall(e,E) compute_color(G,e);
     }
   }

 
  forall(v,L)
  { 
    int deg  = G.outdeg(v);
    int rdeg = 0;
    int bdeg = 0;

    list<edge> E;
    edge e;
    forall_adj_edges(e,v)
    { if (G[e] == red)  rdeg++;
      if (G[e] == blue) bdeg++;
      E.append(G.face_cycle_succ(e));
     }

    if (rdeg < 3 || bdeg > 0) continue;

    if (G.outdeg(v) > 3) simplify_face(G,v,E);

    if (G.outdeg(v) != 3) cout << G.outdeg(v) << endl;

    assert(G.outdeg(v) == 3);

    if (rdeg == deg)
    { G.del_node(v);
      clip_count++;
     }
    else
    { edge e1 = G.first_adj_edge(v);
      edge e2 = G.cyclic_adj_succ(e1);
      edge e3 = G.cyclic_adj_succ(e2);
      int orient = orientation(G[target(e1)],G[target(e2)],G[target(e3)],G[v]);
//cout << "orient = " << orient << endl;
      if (orient <= 0)
        { G.del_node(v);
          clip_count++;
         }
      else
        { edge e;
          forall_adj_edges(e,v) E.append(e);
         }
     }
   
    forall(e,E) compute_color(G,e);
   }

  //check_colors("CLIPPING",G);

  return clip_count;
}




void generate_input(int n, int input, int maxc, list<d3_rat_point>& L)
{ L.clear();
  W.clear();

  switch (input) {

   case 0: random_points_in_cube(n,3*maxc/4,L);
           break;

   case 1: random_points_in_ball(n,maxc,L);
           break;

   case 2: random_points_in_square(n,maxc,L);
           break;

   case 3: random_points_on_paraboloid(n,maxc,L);
           break;

   case 4: lattice_points(n,3*maxc/4,L);
           break;

   case 5: random_points_on_sphere(n,maxc,L);
           break;

 //case 6: random_points_on_segment(n,maxc,L);
 //        break;

  }
}



void upper_lower_hull(GRAPH<d3_rat_point,int>& H, int D)
{ float T = used_time();
  for(;;)
  { edge e;
    list<edge> E;
    forall_edges(e,H) if (H[e] == red) E.append(e);
    if (E.empty()) break;
    FLIPPING(H,E);
    CLIPPING(H,E);
   }
  cout << string("flip/clip: %.2f sec",used_time(T)) << endl;
}




void mode_action(int new_mode)
{ filter = (new_mode & 1);
  trace = (new_mode & 2);
  check = (new_mode & 4);
  pa->set_elim(new_mode & 8);
  pa->set_solid(new_mode & 16);
  d3_rat_point::use_filter = (filter) ? 1 : 0;
 }

  

int main() {

 int n = 32;
 int t = 0;
 int input = 0;

  
 mode  = 1;

 filter = (mode & 1);
 trace = (mode & 2);
 check = (mode & 4);

 d3_rat_point::use_filter = (filter) ? 1 : 0;

 list<string> choices;
 choices.append("cube");
 choices.append("ball");
 choices.append("square");
 choices.append("para");
 choices.append("mesh");
 choices.append("sphere");
 //choices.append("leda");
 //choices.append("line");

 list<string> setup;
 setup.append("filter");
 setup.append("trace");
 setup.append("check");
 setup.append("elim");
 setup.append("solid");
 //setup.append("buffer");

 list<string> thousands;
 thousands.append("0");
 thousands.append("250");
 thousands.append("500");
 thousands.append("1000");
 thousands.append("2000");
 thousands.append("4000");
 thousands.append("8000");
 thousands.append("16000");

 W.choice_item("input",input,choices);
 W.choice_item("points",t,thousands);
 W.int_item("    +",n,4,200);
 W.choice_mult_item("setup",mode,setup,mode_action);

 W.button("gen",1);
 W.button("read",7);
//W.button("write",8);
 W.button("flip",2);
 W.button("clip",3);
 W.button("init",9);
 W.button("run1",4);
 W.button("run2",5);
 W.button("run12",13);
 W.button("d3h",11);
 W.button("move",10);
 W.button("check",6);
 W.button("exit",0);

 W.init(-maxcoord,maxcoord,-2*maxcoord/3);
 W.display();

 W.set_node_width(2);

 list<d3_rat_point> L;
 GRAPH<d3_rat_point,int> H;
 GRAPH<d3_rat_point,int> H1;

 d3_window anim(W,H);

 anim.set_elim(mode & 8);
 anim.set_solid(mode & 16);
 anim.set_speed(anim_speed);


 pa = &anim;

 int but = W.read_mouse();

 list<node> out;

 list<edge> E0;

 while (but != 0)
 {

   switch (but) {
  
   case 1: { int m = (1 << t)/2 * 250;
             rand_int.set_seed(m+n);
             generate_input(m+n,input,int(W.xmax()/2),L);
             W.clear();
             H.clear();
             d3_rat_point p;
             forall(p,L)
             { rat_point q(p.X(),p.Y(),p.W());
               W.draw_point(q.to_point()); 
              }
             break;
           }

   case 7: { ifstream in("d3hf.in");
             L.clear();
             int n;
             cout << read_line(in) << endl;
             in >> n;
             cout << n << endl;
             cout << endl;
             while (n--)
             { double x,y,z;
               in >> x >> y >> z;
               L.append(d3_rat_point(int(1000000000*x),int(1000000000*y),int(1000000000*z),1000000));
              }
             W.clear();
             H.clear();
             d3_rat_point p;
             forall(p,L)
             { rat_point q(p.X(),p.Y(),p.W());
               W.draw_point(q.to_point()); 
              }
             break;
           }

   case 8: { ofstream out("d3hf.out");
             node v;
             forall_nodes(v,H) out << H[v] << endl;
             break;
            }
         

   case 2: { E0 = H.all_edges();
             FLIPPING(H,E0);
             edge e;
             forall_edges(e,H) anim.set_color(e,H[e]);
             anim.draw();
             break;
            }

   case 3: { CLIPPING(H,E0);
             edge e;
             forall_edges(e,H) anim.set_color(e,H[e]);
             anim.draw();
             break;
            }


   case 9: { GRAPH<d3_rat_point,int> H1;
             GRAPH<d3_rat_point,int> H2;
             init_hull(L,H1,H2);
             H.join(H1);
             node_array<rat_vector> pos(H);
             node v;
             forall_nodes(v,H) pos[v] = H[v].to_vector();
             anim.init(pos);
             forall_nodes(v,H) anim.set_color(v,blue);
             W.clear();
             anim.draw();
             break;
            }

   case 4:
   case 5: { int D;
             if (but == 4) 
                D = +1;
             else
                D = -1;

             float T = used_time();

             GRAPH<d3_rat_point,int> H1;
             GRAPH<d3_rat_point,int> H2;
             init_hull(L,H1,H2);

             H.clear();

             if (D == 1)
               H.join(H1);
             else
               H.join(H2);

             cout << string("init_hull: %.2f sec",used_time(T)) << endl;

             node_array<rat_vector> pos(H);
             node v;
             forall_nodes(v,H) pos[v] = H[v].to_vector();
             anim.init(pos);
             forall_nodes(v,H) anim.set_color(v,blue);
             W.clear();
             anim.draw();

//          int count;

             T = used_time();
             int i = 0;

             for(;;)
             { edge e;
               list<edge> E;
               forall_edges(e,H) if (H[e] == red) E.append(e);
               if (E.empty()) break;

               cout << "flipping: " << flush;
               int f = FLIPPING(H,E);
               cout << string("%4d",f) << flush;

               cout << "    clipping: " << flush;
               int c = CLIPPING(H,E);
               cout << string("%4d",c) << endl;

               if (i++ > 500) 
               { error_handler(0,"flip/clip loop.");
                 break;
                }
              }

              cout << string("flip/clip: %.2f sec", used_time(T)) << endl;
              cout << endl;

              edge e;
              forall_edges(e,H) anim.set_color(e,H[e]);
              anim.draw();
              break;
            }

    case 13: { node v;
               edge e;

               float T = used_time();
               GRAPH<d3_rat_point,int> H1;
               GRAPH<d3_rat_point,int> H2;
               init_hull(L,H1,H2);
               cout << string("init_hull: %.2f sec",used_time(T)) << endl;

               upper_lower_hull(H1,+1);
               forall_edges(e,H1) H1[e] = blue2;

               list<edge> B1;

               edge e_hull = H1.last_edge();
               e = e_hull;
               do { edge r = H1.reversal(e);
                    H1[e] = H1[r] = blue;
                    B1.append(e);
                    e = H1.face_cycle_succ(e);
               } while (e != e_hull);

               upper_lower_hull(H2,-1);
               forall_edges(e,H2) H2[e] = orange;

               list<edge> B2;

               list_item it0 = 0;

               e_hull = H2.last_edge();
               e = e_hull;
               do { edge r = H2.reversal(e);
                    H2[e] = H2[r] = red;
                    list_item it = B2.push(e);
                    if (it0 == 0 &&
                        equal_xy(H2[source(e)],H[source(B1.head())])) it0 = it;
                    e = H2.face_cycle_succ(e);
               } while (e != e_hull);


               forall_nodes(v,H2) H[v] = H[v].translate(0,0,40,1);

               H.join(H1);
               H.join(H2);

               forall(e,B1)
               { edge e2 = B2[it0];
                 it0 = B2.cyclic_succ(it0);
                 edge x = H.new_edge(e,source(e2),0,leda::after);
                 edge y = H.new_edge(e2,source(e),0,leda::after);
                 H[x] = H[y] = black;
                 H.set_reversal(x,y);
                }

               cout << endl;

               node_array<rat_vector> pos(H);
               forall_nodes(v,H) pos[v] = H[v].to_vector();
               anim.init(pos);
               forall_nodes(v,H) anim.set_color(v,blue);
               forall_edges(e,H) anim.set_color(e,H[e]);
               anim.draw();
               break;
             }

    case 11: { float T = used_time();
               D3_HULL(L,H,0);
              cout << string("D3_HULL: n = %d  %.2f sec",H1.number_of_nodes(), used_time(T)) << endl;
              node_array<rat_vector> pos(H);
              node v;
              forall_nodes(v,H) pos[v] = H[v].to_vector();
              anim.init(pos);
              forall_nodes(v,H) anim.set_color(v,blue);
              W.clear();
              anim.draw();
              break;
             }

   case 10: moving = !moving;
           break;

   case 6: { check_hull(H,L); 
             edge e;
             forall_edges(e,H) anim.set_color(e,H[e]);
             anim.draw();
             break;
            }

/*
   case 7: { panel P;
             P.int_item("max xcoord",maxcoord,1,500);
             P.int_item("anim speed",anim_speed,1,32);
             P.open(W);
             W.init(-maxcoord,maxcoord,-2*maxcoord/3);
             anim.set_speed(anim_speed);
             break;
           }
*/


   }


  if (moving)
     but = anim.move();
  else
   { if (H.number_of_nodes() > 0)
     { //anim.init(W,H);
       node_array<rat_vector> pos(H);
       node v;
       forall_nodes(v,H) pos[v] = H[v].to_vector();
       anim.init(pos);
       forall_nodes(v,H) anim.set_color(v,blue);
       forall(v,out) anim.set_color(v,red);
       edge e;
       forall_edges(e,H) anim.set_color(e,H[e]);
       anim.draw();
      }
     but = W.read_mouse();
    }
 }


 return 0;

}

