#include <LEDA/graph/graph.h>
#include <LEDA/graph/dimacs_graph.h>

#include <LEDA/system/multi_thread.h>
#include <LEDA/system/thread.h>
#include <LEDA/system/assert.h>

#include <LEDA/core/queue.h>
#include <LEDA/core/b_queue.h>
#include <LEDA/core/array.h>
#include <LEDA/core/list.h>

#include <iostream>
#include <fstream>


using namespace leda;
using std::cout;
using std::cerr;
using std::cin;
using std::endl;


class fifo_set{

  list<node> L;

public:

  fifo_set(){}

  node del() { if (!L.empty()) return L.pop(); else return nil; }
 
  void push(node v)  { L.push(v); }

  void insert(node v, int)  { L.append(v); }
  void insert0(node v, int) { L.append(v); }                            

  void split(int& thr_num, array<fifo_set>& F, node_array<int>& dist)
  { int g = 0;
    node v;
    forall(v,L)
    { int number = g % thr_num;
      F[number].insert(v, dist[v]);
      g++;
     }
    L.clear();
  } 


  int size() { return L.size(); }

  bool empty() { return L.empty(); }

  void clear() { L.clear(); }

  ~fifo_set(){}
};



struct thread_data
{
  graph* G;
  node s;
  node t;

  edge_array<int>* cap;
  edge_array<int>* flow;
  node_array<int>* dist;
  node_array<int>* excess;

  fifo_set* liste;

  int* num_pushes; 
  int* num_edge_inspections; 
  int* num_relabels; 
  int* num_global_relabels; 

  float h; 
  int limit_heur;

  node_array<leda_mutex*>* lock;

  bool local_relabel;
};



void compute_dist_t(const graph& G, node t, 
                    const edge_array<int>& flow, 
                    const edge_array<int>& cap, 
                    const node_array<int>& excess, 
                    node_array<int>& dist,
                    fifo_set& U)
{ 
  int n = G.number_of_nodes();
  b_queue<node> Q(n);
  Q.append(t);

  dist.init(G,n);
  dist[t] = 0;  

  while ( !Q.empty() )
  { node v = Q.pop();
    int  d = dist[v] + 1;
    if ( d >= n) break;

    edge e;
    forall_out_edges(e,v)
    { if ( flow[e] == 0 ) continue;
      node u = target(e); 
      int& du = dist[u];
      if ( du >= n ) 
      { du = d;
        Q.append(u); 
        //if ( excess[u] > 0 ) U.insert(u,d); 
        if ( excess[u] > 0 ) U.push(u); 
       }
     }

    forall_in_edges(e,v)
    { if ( cap[e] == flow[e] ) continue;
      node u = source(e); 
      int& du = dist[u];
      if ( du >= n )  
      { du = d;  
        Q.append(u); 
        //if (excess[u] > 0) U.insert(u,d); 
        if ( excess[u] > 0 ) U.push(u); 
      }
    }
  }
}



void maxflow(const graph& G, node s, node t, 
             const edge_array<int>& cap, 
             edge_array<int>& flow, 
             node_array<int>& dist, 
             node_array<int>& excess,
             fifo_set& U, 
             int& num_pushes, 
             int& num_edge_inspections,
             int& num_relabels, 
             int& num_global_relabels, 
             float h, 
             int limit_heur, 
             node_array<leda_mutex*>& lock, 
             bool local_relabel)
{   
  int n = G.number_of_nodes();  

  int local_num_pushes = 0;
  int local_num_relabels = 0;
  int local_num_inspections = 0;

  for(;;) 
  {
    node v = U.del();

    if (v == nil) break;
    if (v == t) continue;
    if (dist[v] >= n) continue;

    assert(excess[v] > 0);

    //if (excess[v] <= 0) continue;

    if (lock[v]->trylock() != 0) // wenn Knoten v gelockt, fuege v in Liste ein 
    { U.insert(v, dist[v]);      // und betrachte ihn spaeter
      continue;
     }

    int ev   = excess[v];
    int dv   = dist[v];   
    int dmin = MAXINT;


    assert(ev > 0);

    edge e; 
    forall_out_edges(e,v)
    { local_num_inspections++;
      int& fe = flow[e];
      int  rc = cap[e] - fe;
      if (rc == 0) continue;

      node w = target(e);

      if (lock[w]->trylock() != 0)  continue;

      int dw = dist[w];
      if ( dw == dv - 1 ) 
      { local_num_pushes++;
        int& ew = excess[w];
        if (ew == 0) U.insert0(w,dw); 
        if (ev <= rc) 
        { ew += ev; 
          fe += ev;
          ev = 0; 
          lock[w]->unlock(); 
          break;
         }
         else
         { ew += rc; 
           fe += rc;
           ev -= rc;
           }
        }
       else 
         if (dw < dmin) dmin = dw;

       lock[w]->unlock();
     }

    if ( ev > 0 )
    { 
      forall_in_edges(e,v)
      { local_num_inspections++;
        int& fe = flow[e];
        if (fe == 0) continue;
 
        node w = source(e);
 
        if (lock[w]->trylock() != 0) continue;
 
        int dw = dist[w];
        if ( dw == dv - 1 ) 
        { Fetch_And_Increment(&num_pushes);
          local_num_pushes++;
          int& ew = excess[w];
          if (ew == 0) U.insert0(w,dw); 
          if (ev <= fe) 
           { fe -= ev; 
             ew += ev;
             ev = 0; 
             lock[w]->unlock();
             break;
            }
           else
            { ew += fe; 
              ev -= fe;
              fe = 0;
             }
          }
          else 
            if (dw < dmin) dmin = dw;
 
          lock[w]->unlock();
        }
    }
        
    excess[v] = ev;
        
    if (ev > 0 && local_relabel) 
    { 
      if (local_num_inspections > limit_heur)
      { lock[v]->unlock();
        break;
       }

      dmin++; 

      if (dmin < n)
      { local_num_relabels++;
        dist[v] = dmin;
        U.insert(v,dmin);
       }
    }

   lock[v]->unlock();

  }

  Fetch_And_Add(&num_relabels,local_num_relabels);
  Fetch_And_Add(&num_pushes,local_num_pushes);
  Fetch_And_Add(&num_edge_inspections,local_num_inspections);
}


void* maxflow_thread(void *threadarg)
{ 
  thread_data* my_data = (thread_data *) threadarg;       
  graph& G = *(my_data->G);       

  node s = my_data->s;        
  node t = my_data->t;                

  edge_array<int>& cap = *(my_data->cap);
  edge_array<int>& flow = *(my_data->flow);
  node_array<int>& dist = *(my_data->dist);
  node_array<int>& excess = *(my_data->excess);

  //array<fifo_set>& U = *(my_data->liste);         
  fifo_set& U = *(my_data->liste);         

  int& num_pushes = *(my_data->num_pushes);
  int& num_edge_inspections = *(my_data->num_edge_inspections);
  int& num_relabels = *(my_data->num_relabels);
  int& num_global_relabels = *(my_data->num_global_relabels);

  float h = my_data->h;
  int limit_heur = my_data->limit_heur;

  node_array<leda_mutex*>& lock = *(my_data->lock);

  bool local_relabel = my_data->local_relabel;

  maxflow(G,s,t,cap,flow,dist,excess,U,
          num_pushes,num_edge_inspections,num_relabels,num_global_relabels, 
          h, limit_heur, lock, local_relabel);

  return 0;
}


int main(int argc, char** argv)
{
  if (argc < 3)
  { cerr << "usage: mfc thrnum h" << endl;
    return 1;
   }

  int   thr_num = atoi(argv[1]);
  float h = atof(argv[2]);

  graph G;
  node s;
  node t;
  edge_array<int> cap(G);
  edge_array<int> flow(G);
  int num_pushes, num_edge_inspections, num_relabels, num_global_relabels;

  //int thr_num = 4;

  // Graph einlesen
  read_dimacs_mf(cin,G,s,t,cap);
  if (s == t) LEDA_EXCEPTION(1,"MAXFLOW: source == sink");

/**************** Bitte hier waehlen ! ***************************************/

  bool local_relabel = true; 

/*****************************************************************************/

  //for(thr_num=1; thr_num <= 4; thr_num *= 2)
  {
  float t3 = used_time(); 
  float t4 = real_time();

  int n = G.number_of_nodes();
  int m = G.number_of_edges(); 

  cout << endl;
  cout << string("n = %d  m = %d  threads = %d  h = %.1f",n,m,thr_num,h);
  cout << endl;

  fifo_set U; 

  flow.init(G,0);
  node_array<int>  excess(G,0);

  if (G.outdeg(s) == 0) return 0;
 
  node_array<leda_mutex*> lock(G);
  node v;
  forall_nodes(v,G) lock[v] = new leda_mutex;

  // saturate all edges leaving s

  edge e;
  forall_out_edges(e,s) 
  { int c = cap[e];
    if (c == 0) continue;
    node v = target(e);
    flow[e] = c;
    excess[s] -= c;
    excess[v] += c;
   } 

 
  int limit_heur  = (int) (h*m)/thr_num;
  
  node_array<int> dist(G);

  compute_dist_t(G,t,flow,cap,excess,dist,U);

  num_relabels = num_pushes = num_edge_inspections = 0;
  num_global_relabels  = 0;


  while(1)
  {

    if(U.size() <= thr_num)
    {
      maxflow(G, s, t, cap, flow, dist, excess, U, num_pushes, 
              num_edge_inspections, num_relabels, num_global_relabels, h, 
              limit_heur, lock, local_relabel);
      U.clear();
     }
     else
     {
       array<fifo_set> F(thr_num);
       U.split(thr_num, F, dist);

       // Threads
       // thread_data thread_data_array[thr_num];

       array<thread_data> thread_data_array(thr_num);
       thread_t* thr = new thread_t[thr_num];


       // Threadargumente
       for(int i=0; i<thr_num; i++) {
        thread_data_array[i].G = &G;
        thread_data_array[i].s = s;
        thread_data_array[i].t = t;

        thread_data_array[i].cap = &cap;
        thread_data_array[i].flow = &flow;
        thread_data_array[i].dist = &dist;
        thread_data_array[i].excess = &excess;
        //thread_data_array[i].liste = &F;
        thread_data_array[i].liste = &F[i];

        thread_data_array[i].num_pushes = &num_pushes;
        thread_data_array[i].num_edge_inspections = &num_edge_inspections;
        thread_data_array[i].num_relabels = &num_relabels;
        thread_data_array[i].num_global_relabels = &num_global_relabels;

        thread_data_array[i].h = h;
        thread_data_array[i].limit_heur = limit_heur;

        thread_data_array[i].lock = &lock;
        thread_data_array[i].local_relabel = local_relabel;
       }


      for(int i=0; i<thr_num; i++) 
          thr[i] = thread_create(maxflow_thread,(void*)&thread_data_array[i]);

      for(int i=0; i<thr_num; i++) thread_join(thr[i]);
    }


    num_global_relabels++;
    assert(U.empty());
    compute_dist_t(G,t,flow,cap,excess,dist,U);

    if ( U.empty() ) break;
  }


   float time_complete = real_time(t4);

   cout << string("flow: %6d ",excess[t]);
   cout << string ("time: %5.2f  cpu: %5.2f sec", time_complete, used_time(t3));
   cout << endl;
   
   int ops = num_relabels + num_pushes + num_edge_inspections;
   
   cout << string("pushes: %7d ",  num_pushes);
   cout << string("relabels: %4d ", num_relabels);
   cout << string("inspects: %7d ", num_edge_inspections);
   cout << string("updates: %2d ",  num_global_relabels);
   cout << string("ops/sec: %6d ",  int(ops/time_complete)) << endl;
   
   
   forall_nodes(v,G) delete lock[v]; 
  }

  return 0;
}
