#include <LEDA/graph/graph.h>
#include <LEDA/graph/node_level_queue.h>
#include <LEDA/graph/dimacs_graph.h>

#include <LEDA/system/multi_thread.h>
#include <LEDA/system/thread.h>
#include <LEDA/system/assert.h>

#include <LEDA/core/queue.h>
#include <LEDA/core/b_queue.h>
#include <LEDA/core/array.h>
#include <LEDA/core/list.h>

#include <iostream>
#include <fstream>


#define residual_cap(e)  (e.out() ? (cap[e] - flow[e]) : flow[e])
#define push_flow(e,x)   { flow[e] += (e.out() ? x : -(x)); }



using namespace leda;
using std::cout;
using std::cerr;
using std::cin;
using std::endl;


const bool local_relabel = true; 



void split(const list<node>& L, int& thr_num, array<list<node> >& F)
{ for(int i=0; i<thr_num; i++) F[i].clear();
  node v;
  int j = 0;
  forall(v,L) F[j++ % thr_num].append(v);
} 


struct thread_data
{
  graph* G;
  node s;
  node t;

  edge_array<int>* cap;
  edge_array<int>* flow;
  node_array<int>* dist;
  node_array<int>* excess;
  node_array<node>* succ;

  node_array<leda_mutex*>* lock;

  list<node>* liste;

  int* num_pushes; 
  int* num_edge_inspections; 
  int* num_relabels; 
  int* num_global_relabels; 

  float h; 
  int  limit_heur;
};



void compute_dist_t(const graph& G, node s, node t, 
                    const edge_array<int>& flow, 
                    const edge_array<int>& cap, 
                    const node_array<int>& excess, 
                    node_array<int>& dist,
                    list<node>& L)
{
  L.clear();

  int n = G.number_of_nodes();
  b_queue<node> Q(n);
  Q.append(t);

  dist.init(G,n);
  dist[t] = 0;  

  int d = 0;

  while (!Q.empty())
  { node v = Q.pop();
    d = dist[v] + 1;

    assert(d < n);

    edge e;
    forall_out_edges(e,v)
    { if ( flow[e] == 0 ) continue;
      node u = G.target(e); 
      int& du = dist[u];
      if ( du >= n ) 
      { du = d;
        Q.append(u); 
        if (excess[u] > 0) L.push(u); 
       }
     }

    forall_in_edges(e,v)
    { if ( cap[e] == flow[e] ) continue;
      node u = G.source(e); 
      int& du = dist[u];
      if ( du >= n )  
      { du = d;  
        Q.append(u); 
        if (excess[u] > 0) L.push(u); 
      }
    }
  }

}



void maxflow(const graph& G, node s, node t, 
             const edge_array<int>& cap, 
             edge_array<int>& flow, 
             node_array<int>& dist, 
             node_array<int>& excess,
             node_array<node>& succ,
             node_array<leda_mutex*>& lock,
             const list<node>& L, 
             int& num_pushes, 
             int& num_edge_inspections,
             int& num_relabels, 
             int& num_global_relabels, 
             float h, 
             int limit_heur )
{   
  int n = G.number_of_nodes();  

  int local_num_pushes = 0;
  int local_num_relabels = 0;
  int local_num_inspections = 0;

  node_level_queue<graph, node_array<node> > U(G,succ);

  node v;
  forall(v,L) U.insert(v,dist[v]);

  U.insert(t,0);
  excess[t] += 1; // now t will never be inserted again

  while ( (v = U.del_max()) != t)
  {
    assert(dist[v] < n);
    assert(excess[v] >= 0);
    if (excess[v] == 0) continue;

/*
    if (lock[v]->trylock() != 0) // wenn Knoten v gelockt, fuege v in Liste ein 
    { U.insert(v, dist[v]);      // und betrachte ihn spaeter
      continue;
     }
*/
    //lock[v]->lock();

    int ev   = excess[v];
    int dv   = dist[v];   

    while (ev > 0 && dv < n) 
    {
      int dmin = MAXINT;

      graph::incident_edge e; 
      forall_incident_edges(e,v)
      { if (ev == 0) break;
   
        local_num_inspections++;
  
        int x = residual_cap(e);
        if (x == 0) continue;
  
        node w = G.opposite(e,v);
  
        //if (lock[w]->trylock() != 0)  continue; // sn: could be a problem ?
  
        int dw = dist[w];
  
        if (dw == dv-1) 
        { local_num_pushes++;
          int& ew = excess[w];
          if (ew == 0) U.insert(w,dw); 
          if (ev < x) x = ev; 
          push_flow(e,x);
          ew += x; 
          ev -= x; 
         }
         else 
           if (dw < dmin) dmin = dw;
  
         //lock[w]->unlock();
       }
          
      if (ev > 0 && local_relabel) 
      { 
        // remaining excess: relabel vertex v
  
        if (local_num_inspections > limit_heur)
        { excess[v] = ev;
          dist[v] = dv;
          //lock[v]->unlock();
          break;
         }
  
        local_num_relabels++;
        dv = dmin + 1; 
      }
  
    }
  
    excess[v] = ev;
    dist[v] = dv;
    //lock[v]->unlock();
  }

  excess[t] -= 1;

  Fetch_And_Add(&num_relabels,local_num_relabels);
  Fetch_And_Add(&num_pushes,local_num_pushes);
  Fetch_And_Add(&num_edge_inspections,local_num_inspections);
}


void* maxflow_thread(void *threadarg)
{ 
  thread_data* my_data = (thread_data *) threadarg;       
  graph& G = *(my_data->G);       

  node s = my_data->s;        
  node t = my_data->t;                

  edge_array<int>& cap = *(my_data->cap);
  edge_array<int>& flow = *(my_data->flow);
  node_array<int>& dist = *(my_data->dist);
  node_array<int>& excess = *(my_data->excess);
  node_array<node>& succ = *(my_data->succ);

  list<node>& U = *(my_data->liste);         

  int& num_pushes = *(my_data->num_pushes);
  int& num_edge_inspections = *(my_data->num_edge_inspections);
  int& num_relabels = *(my_data->num_relabels);
  int& num_global_relabels = *(my_data->num_global_relabels);

  float h = my_data->h;
  int limit_heur = my_data->limit_heur;

  node_array<leda_mutex*>& lock = *(my_data->lock);

  maxflow(G,s,t,cap,flow,dist,excess,succ,lock,U,
          num_pushes,num_edge_inspections,num_relabels,num_global_relabels, 
          h, limit_heur);

  return 0;
}


int main(int argc, char** argv)
{
  if (argc < 3)
  { cerr << "usage: mfc thrnum h" << endl;
    return 1;
   }

  int   thr_num = atoi(argv[1]);
  float h = atof(argv[2]);

  graph G(4,2);  // 2 data slots for nodes and edges

  node s;
  node t;

  edge_array<int> cap;
  assert(cap.use_edge_data(G));

  edge_array<int> flow;
  assert(flow.use_edge_data(G));

  node_array<int>  excess;
  assert(excess.use_node_data(G));

  node_array<int> dist;
  assert(dist.use_node_data(G));

  node_array<node> succ;
  assert(succ.use_node_data(G));

  node_array<leda_mutex*> lock;
  assert(lock.use_node_data(G));


  int num_pushes, num_edge_inspections, num_relabels, num_global_relabels;


  cout << endl;
  cout << "Reading Dimacs Network ... " << endl; 

  read_dimacs_mf(cin,G,s,t,cap);
  assert (s != t);


  int n = G.number_of_nodes();
  int m = G.number_of_edges(); 

  cout << string("n = %d  m = %d  threads = %d  h = %.1f",n,m,thr_num,h);
  cout << endl;

  if (G.outdeg(s) == 0) return 0;


  float t3 = used_time(); 
  float t4 = real_time();
 
  node v;
  forall_nodes(v,G) 
  { lock[v] = new leda_mutex;
    excess[v] = 0;
   }

  edge e;
  forall_edges(e,G) flow[e] = 0;

  // saturate all edges leaving s
  forall_out_edges(e,s) 
  { int c = cap[e];
    if (c == 0) continue;
    node v = target(e);
    flow[e] = c;
    excess[s] -= c;
    excess[v] += c;
   } 

 
  int limit_heur  = (int) (h*m)/thr_num;
  

  num_relabels = num_pushes = num_edge_inspections = 0;
  num_global_relabels  = 0;


  // Threadargumente
  array<thread_data> thread_data_array(thr_num);

  array<list<node> > F(thr_num);

  for(int i=0; i<thr_num; i++)
  {
    thread_data_array[i].G = &G;
    thread_data_array[i].s = s;
    thread_data_array[i].t = t;

    thread_data_array[i].cap = &cap;
    thread_data_array[i].flow = &flow;
    thread_data_array[i].dist = &dist;
    thread_data_array[i].excess = &excess;
    thread_data_array[i].succ = &succ;

    thread_data_array[i].liste = &F[i];

    thread_data_array[i].num_pushes = &num_pushes;
    thread_data_array[i].num_edge_inspections = &num_edge_inspections;
    thread_data_array[i].num_relabels = &num_relabels;
    thread_data_array[i].num_global_relabels = &num_global_relabels;

    thread_data_array[i].h = h;
    thread_data_array[i].limit_heur = limit_heur;

    thread_data_array[i].lock = &lock;
  }



  while(1)
  {
    list<node> L; 

    num_global_relabels++;
    compute_dist_t(G,s,t,flow,cap,excess,dist,L);

    if ( L.empty() ) break;

    cout << L.size() << endl;

    //if (L.size() <= thr_num)
    if (L.size() <= 10*thr_num)
    {
      maxflow(G, s, t, cap, flow, dist, excess, succ, lock, L, 
              num_pushes,num_edge_inspections,num_relabels,num_global_relabels,
              h, limit_heur);
      L.clear();
     }
     else
     { thread_t* thr = new thread_t[thr_num];

       split(L, thr_num, F);

       for(int i=0; i<thr_num; i++) 
          thr[i] = thread_create(maxflow_thread,(void*)&thread_data_array[i]);

       for(int i=0; i<thr_num; i++) thread_join(thr[i]);
     }

  }


  float time_complete = real_time(t4);

  cout << string("flow: %6d ",excess[t]);
  cout << string ("time: %5.2f  cpu: %5.2f sec", time_complete, used_time(t3));
  cout << endl;
   
  int ops = num_relabels + num_pushes + num_edge_inspections;
   
  cout << string("pushes: %7d ",  num_pushes);
  cout << string("relabels: %4d ", num_relabels);
  cout << string("inspects: %7d ", num_edge_inspections);
  cout << string("updates: %2d ",  num_global_relabels);
  cout << string("ops/sec: %6d ",  int(ops/time_complete)) << endl;
   
   
  forall_nodes(v,G) delete lock[v]; 

  return 0;
}
