
//------------------------------------------------------------------------------

template <class graph_t>
class  STRONG_COMPONENTS_TARJAN0 {

typedef typename graph_t::node node;
typedef typename graph_t::edge edge;

int dfs_count;
int scc_count;

float cputime;

template <class compnum_array>
void dfs(const graph_t& G, node v, compnum_array& compnum,
                                   node_array<int,graph_t>& incurrent,
                                   node_array<int,graph_t>& dfsnum,
                                   node_array<int,graph_t>& lowpnt,
                                   b_stack<node>& current)
{
  dfs_count++;
  dfsnum[v] = dfs_count;
  lowpnt[v] = dfs_count;

  incurrent[v] = true;
  current.push(v);

  edge e;
  forall_out_edges(e,v)
  { node w = G.target(e);
    if (dfsnum[w] == 0)
    { dfs(G,w,compnum,incurrent,dfsnum,lowpnt,current);
      if (lowpnt[w] < lowpnt[v]) lowpnt[v] = lowpnt[w];
     }
    else
      if (incurrent[w] && dfsnum[w] < lowpnt[v]) lowpnt[v] = dfsnum[w];
  }


  if (lowpnt[v] == dfsnum[v])
  { node u;
    do { u = current.pop();
         incurrent[u] = false;
         compnum[u] = scc_count;
     } while (u != v);
    scc_count++;
   }
}


public:

template <class compnum_array>
int run(const graph_t& G, compnum_array& compnum)
{
  float t = used_time();

  int n = G.number_of_nodes();

  node_array<int,graph_t> incurrent(G);
  b_stack<node>          current(n);
  node_array<int,graph_t>  dfsnum(G);
  node_array<int,graph_t>  lowpnt(G);

  scc_count = 0;
  dfs_count = 0;

  node v;
  forall_nodes(v,G)   
  { dfsnum[v] = 0;
    incurrent[v] = false;
   }

  forall_nodes(v,G)   
    if (dfsnum[v] == 0)
       dfs(G,v,compnum,incurrent,dfsnum,lowpnt,current);

 cputime = used_time(t);
 return scc_count;
}

template <class compnum_array>
int operator()(const graph_t& G, compnum_array& compnum)
{ return run(G,compnum); }

float cpu_time() const { return cputime; }

};



//---------------------------------------------------------------------------


template <class graph_t>
class  STRONG_COMPONENTS_TARJAN1 {

typedef typename graph_t::node node;
typedef typename graph_t::edge edge;

int dfs_count;
int scc_count;

float cputime;


/*
   As in the implementation of the Cheriyan/Mehlhorn algorithm (suggested 
   by Peter Sanders) we use only one node_array "compnum" for both dfs 
   and component numbering. For every node v there are 3 cases: 

   unvisited:   compnum[v] == 0
   completed:   compnum[v]  > 0  is the scc-number of v 
   unfinished:  compnum[v]  < 0  is the dfs-number of v
   
   For the computation of lowpoints it is important that dfs-numbers
   are generated increasingly. Therefore we start dfs numbering at -n,
   i.e. the range of dfs numbers is [-n ... -1].

   The most important improvement compared wit the text book version
   of the algorithm is the implicit representation of lowpoints:
   We do not store low points in a node_array ! In the recursive dfs
   function it suffices to store the lowpoint of the current node v in 
   a local variable and to make dfs return this value.

   See the code below for more details.
*/


template <class compnum_array>
int dfs(const graph_t& G, node v, compnum_array& compnum,
                                  b_stack<node>& unfinished)
{ int dfsnum = dfs_count++;
  int lowpnt = dfsnum;
  compnum[v] = dfsnum; 
  unfinished.push(v);

  edge e;
  forall_out_edges(e,v)
  { node w = G.target(e);
    int d = compnum[w];
    if (d == -1) d = dfs(G,w,compnum,unfinished);
    if (d < lowpnt) lowpnt = d;
   }

  if (dfsnum == lowpnt)
  { node u; 
    do { u = unfinished.pop();
         compnum[u] = scc_count;
       } while (u != v);
    scc_count++;
   }

  return lowpnt;
}


public:

template <class compnum_array>
int run(const graph_t& G, compnum_array& compnum)
{
  float t = used_time();

  int n = G.number_of_nodes();

  b_stack<node> unfinished(n);

  dfs_count = -(n+1);
  scc_count = 0;

  node v;
  forall_nodes(v,G) compnum[v] = -1;

  forall_nodes(v,G)   
    if (compnum[v] == -1) dfs(G,v,compnum,unfinished);

 cputime = used_time(t);
 return scc_count;
}

template <class compnum_array>
int operator()(const graph_t& G, compnum_array& compnum)
{ return run(G,compnum); }

float cpu_time() const { return cputime; }

};


//------------------------------------------------------------------------------


template <class graph_t>
class  STRONG_COMPONENTS_TARJAN2 {

typedef typename graph_t::node node;
typedef typename graph_t::edge edge;

int dfs_count;
int scc_count;

float cputime;

template <class compnum_array>
int dfs(const graph_t& G, node v, compnum_array& compnum,
                                  b_stack<node>& edgeStack,
                                  b_stack<node>& unfinished)
{ int dfsnum = dfs_count++;
  int lowpnt = dfsnum;
  compnum[v] = dfsnum; 
  unfinished.push(v);

  int sz = edgeStack.size();
  edge e;
  forall_rev_out_edges(e,v) edgeStack.push(G.target(e));

  while (edgeStack.size() > sz)
  { node w = edgeStack.pop();
    int d = compnum[w];
    if (d == -1) d = dfs(G,w,compnum,edgeStack,unfinished);
    if (d < lowpnt) lowpnt = d;
   }

  if (dfsnum == lowpnt)
  { node u; 
    do { u = unfinished.pop();
         compnum[u] = scc_count;
       } while (u != v);
    scc_count++;
   }

  return lowpnt;
}


public:

template <class compnum_array>
int run(const graph_t& G, compnum_array& compnum)
{
  float t = used_time();

  int n = G.number_of_nodes();
  int m = G.number_of_edges();

  b_stack<node> edgeStack(m);
  b_stack<node> unfinished(n);

  dfs_count = -(n+1);
  scc_count = 0;

  node v;
  forall_nodes(v,G) compnum[v] = -1;

  forall_nodes(v,G)   
  { if (compnum[v] != -1) continue;
    dfs(G,v,compnum,edgeStack,unfinished);
   }

 cputime = used_time(t);
 return scc_count;
}

template <class compnum_array>
int operator()(const graph_t& G, compnum_array& compnum)
{ return run(G,compnum); }

float cpu_time() const { return cputime; }

};
 


//------------------------------------------------------------------------------



template <class graph_t>
class  STRONG_COMPONENTS_TARJAN_NR1 {

typedef typename graph_t::node node;
typedef typename graph_t::edge edge;

float cputime;

public:


template<class compnum_array>
int run(const graph_t& G, compnum_array& compnum)
{ 
  float t = used_time();

  int n = G.number_of_nodes();
  int m = G.number_of_edges();

  b_stack<int>  dfsnumStack(n);
  b_stack<int>  lowpntStack(n);
  b_stack<node> edgeStack(n+m);

  b_stack<node> unfinished(n);

  int count = -(n+1);   // dfs numbers:  -(n+1), -n, ... 
  int scc_count =  0;   // scc numbers:   0,  1, ...

  node v;
  forall_nodes(v,G) compnum[v] = -1;  // mark unvisited

  forall_nodes(v,G)
  { 
    if (compnum[v] != -1) continue;

    int dfsnum = count++;
    int lowpnt = dfsnum;

    compnum[v] = dfsnum;
    unfinished.push(v);

    edgeStack.push(G.stop_node());
    edge e;
    forall_rev_out_edges(e,v) edgeStack.push(G.target(e));

    for(;;)
    { 
      node w = edgeStack.pop();

      if (w == G.stop_node())   // all edges scanned: finish current node
      { if (dfsnum == lowpnt) 
        { int d = 0;
          while (d != dfsnum)
          { node u = unfinished.pop(); 
            d = compnum[u];
            compnum[u] = scc_count;
          } 
          scc_count++;
         }
   
        if (dfsnumStack.empty()) break;

        dfsnum = dfsnumStack.pop();
        int lp = lowpntStack.pop();
        if (lp < lowpnt) lowpnt = lp; 

        continue;
       }

      int d = compnum[w];

      if (d == -1)  // tree edge (v,w)
      { dfsnumStack.push(dfsnum);
        lowpntStack.push(lowpnt);
        edgeStack.push(G.stop_node());
        edge e;
        forall_rev_out_edges(e,w) edgeStack.push(G.target(e));
        dfsnum = count++;
        lowpnt = dfsnum;
        compnum[w] = dfsnum;
        unfinished.push(w);
        continue;
       }

      if (d < lowpnt)  lowpnt = d;
    } 
  }

  cputime = used_time(t);
  return scc_count;
}


template <class compnum_array>
int operator()(const graph_t& G, compnum_array& compnum)
{ return run(G,compnum); }

float cpu_time() const { return cputime; }

};


//------------------------------------------------------------------------------

template <class graph_t>
class  STRONG_COMPONENTS_TARJAN_NR2 {

typedef typename graph_t::node node;
typedef typename graph_t::edge edge;

float cputime;

public:


template<class compnum_array>
int run(const graph_t& G, compnum_array& compnum)
{ 
  float t = used_time();

  int n = G.number_of_nodes();
  int m = G.number_of_edges();

  b_stack<int>  nodeStack(2*n); // for dfsnum + lowpoint
  b_stack<node> edgeStack(n+m);

  b_stack<node> unfinished(n);

  int count = -(n+1);   // dfs numbers:  -(n+1), -n, ... 
  int scc_count =  0;   // scc numbers:   0,  1, ...

  node v;
  forall_nodes(v,G) compnum[v] = -1;  // mark unvisited

  forall_nodes(v,G)
  { 
    if (compnum[v] != -1) continue;

    int dfsnum = count++;
    int lowpnt = dfsnum;

    compnum[v] = dfsnum;
    unfinished.push(v);

    edgeStack.push(G.stop_node());
    edge e;
    forall_rev_out_edges(e,v) edgeStack.push(G.target(e));

    for(;;)
    { 
      node w = edgeStack.pop();

      if (w == G.stop_node())   // all edges scanned: finish current node
      { if (dfsnum == lowpnt) 
        { int d = 0;
          while (d != dfsnum)
          { node u = unfinished.pop(); 
            d = compnum[u];
            compnum[u] = scc_count;
          } 
          scc_count++;
         }
   
        if (nodeStack.empty()) break;

        dfsnum = nodeStack.pop();
        int lp = nodeStack.pop();
        if (lp < lowpnt) lowpnt = lp; 

        continue;
       }

      int d = compnum[w];

      if (d == -1)  // tree edge (v,w)
      { nodeStack.push(lowpnt);
        nodeStack.push(dfsnum);
        edgeStack.push(G.stop_node());
        edge e;
        forall_rev_out_edges(e,w) edgeStack.push(G.target(e));
        dfsnum = count++;
        lowpnt = dfsnum;
        compnum[w] = dfsnum;
        unfinished.push(w);
        continue;
       }

      if (d < lowpnt)  lowpnt = d;
    } 
  }

  cputime = used_time(t);
  return scc_count;
}


template <class compnum_array>
int operator()(const graph_t& G, compnum_array& compnum)
{ return run(G,compnum); }

float cpu_time() const { return cputime; }

};


